import pandas as pd

from lhcsmapi.analysis.CircuitAnalysis import CircuitAnalysis
from lhcsmapi.analysis.pc.PcAnalysis import PcAnalysis
from lhcsmapi.analysis.decorators import check_dataframe_empty
from lhcsmapi.reference.Reference import Reference


class R60_80_120ACircuitAnalysis(PcAnalysis, CircuitAnalysis):
    """Class for analysis of 60A, 80-120A circuits

    """
    @staticmethod
    @check_dataframe_empty(mode='any', warning='I_A_REF, I_A signals are empty, t_quench calculation skipped!')
    def find_time_of_quench(i_ref_df: pd.DataFrame, i_a_df: pd.DataFrame) -> float:
        """ Function calculating the time at which the reference and actual power converter currents start to diverge.
        The algorithm firstly takes an absolute value of both signals and re-interpolates them to the same sampling
        time.
        Then, the maximum difference for time range [-1, t=i_max] s is obtained.
        The time of deviation is equal to the first index for which the difference exceeds the maximum one before an
        event.

        :param i_ref_df: reference power converter current
        :type i_ref_df: pd.DataFrame
        :param i_a_df: actual power converter current (non-filtered, high sampling rate)
        :type i_a_df: pd.DataFrame

        :return: time at which I_A and I_REF signals start to diverge
        """

        # take absolute value of each signal
        i_ref_abs_df = abs(i_ref_df)
        i_a_abs_df = abs(i_a_df)
        # re-interpolate both signals
        idx = i_ref_abs_df.index.union(i_a_abs_df.index)
        signal_comparison_df = i_ref_abs_df.reindex(idx).interpolate('index')
        signal_comparison_df = signal_comparison_df.join(i_a_abs_df).dropna()

        # find maximum difference between the time of maximum I_REF value and 0.1 second before/after
        t_i_ref_max = i_ref_abs_df[(i_ref_abs_df.index > -0.1) & (i_ref_abs_df.index < 0.1)].idxmax().values[0]
        diff_df = abs(signal_comparison_df['STATUS.I_REF'] - signal_comparison_df['IAB.I_A'])
        max_diff = diff_df[(diff_df.index > (t_i_ref_max - 1)) & (diff_df.index < t_i_ref_max)].max()

        # find time of start of deviation from maximum difference between maximum value of I_REF and 0 (FGC timestamp)
        mask = (diff_df.index > t_i_ref_max) & (diff_df.index < 0) & (diff_df.values > max_diff)

        return diff_df[mask].idxmin()

    def create_mp3_results_table(self):
        """ Method creating an MP3-compatible quench database row for a circuit

        :return: pd.DataFrame with the results of analysis calculation
        """
        columns_mp3 = Reference.read_quench_database_columns(circuit_type='60_80_120A', table_type='MP3')
        mp3_table = pd.DataFrame(columns=columns_mp3, index=self.results_table.index)

        mp3_table[columns_mp3] = self.results_table[columns_mp3]

        return mp3_table

