import pandas as pd

import lhcsmapi
from lhcsmapi.Time import Time
from lhcsmapi.analysis.CircuitQuery import CircuitQuery, get_quench_current
from lhcsmapi.analysis.busbar.BusbarResistanceQuery import BusbarResistanceQuery
from lhcsmapi.analysis.dfb.DfbQuery import DfbQuery
from lhcsmapi.analysis.diode.DiodeLeadResistanceQuery import DiodeLeadResistanceQuery
from lhcsmapi.analysis.ee.EeQuery import Ee13kAQuery
from lhcsmapi.analysis.nqps.NqpsQuery import NqpsRqQuery
from lhcsmapi.analysis.pc.PcQuery import Pc13kAQuery
from lhcsmapi.analysis.pic.PicQuery import PicQuery
from lhcsmapi.analysis.qds.QdsQuery import QdsQuery
from lhcsmapi.analysis.qh.QuenchHeaterQuery import QuenchHeaterQuery
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.reference.Reference import Reference


class RqCircuitQuery(PicQuery, Pc13kAQuery, Ee13kAQuery, NqpsRqQuery, QuenchHeaterQuery, BusbarResistanceQuery,
                     DiodeLeadResistanceQuery,  QdsQuery, DfbQuery, CircuitQuery):
    """Class  extending the CircuitQuery class with methods for creation of templates of analysis tables and query
    of voltage across magnets in main quadrupole circuits.

    """

    def create_report_analysis_template(self, source_timestamp_qds_df: pd.DataFrame,
                                        source_timestamp_nqps_df: pd.DataFrame, timestamp_fgc: int,
                                        timestamp_pic: int, init_file_path: str, i_meas_rqd_df: pd.DataFrame,
                                        i_meas_rqf_df:pd.DataFrame, author: str) -> pd.DataFrame:
        """ Method creating a report analysis template for main quadrupole circuits

        :param source_timestamp_qds_df: table with QDS PM source and timestamp
        :type source_timestamp_qds_df: pd.DataFrame
        :param source_timestamp_nqps_df: table with QDS PM source and timestamp
        :type source_timestamp_nqps_df: pd.DataFrame
        :param timestamp_fgc: unix timestamp of the first FGC event in RQ circuit in nanoseconds
        :type timestamp_fgc: int
        :param timestamp_pic: unix timestamp of the first PIC event in RQ circuit in nanoseconds
        :type timestamp_pic: int
        :param init_file_path: path to an init file to read LHCSM-HWC notebooks version
        :type init_file_path: str
        :param i_meas_rqd_df: measured power converter current in RQD circuit
        :type i_meas_rqd_df: pd.DataFrame
        :param i_meas_rqf_df: measured power converter current in RQF circuit
        :type i_meas_rqf_df: pd.DataFrame
        :param author: NICE account name of the analysis author
        :type author: str
        :return: pre-filled pd.DataFrame with a template to be filled further with data from analysis
        of a powering event in main quadrupole (RQ) circuits
        """
        columns = Reference.read_quench_database_columns(circuit_type='RQ', table_type='MP3')
        results_table = pd.DataFrame(columns=columns)

        results_table['Position'] = source_timestamp_qds_df['source']
        results_table['nQPS RQD crate name'] = results_table['Position']. \
            apply(lambda col: MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, col))
        results_table['nQPS RQF crate name'] = results_table['Position']. \
            apply(lambda col: MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, col))

        results_table['Delta_t(iQPS-PIC)'] = 1e-6 * (source_timestamp_qds_df['timestamp'] - timestamp_pic)
        results_table['I_Q_MQD'] = results_table['Delta_t(iQPS-PIC)'].apply(
            lambda col: get_quench_current(i_meas_rqd_df, 1e-3 * col))
        results_table['I_Q_MQF'] = results_table['Delta_t(iQPS-PIC)'].apply(
            lambda col: get_quench_current(i_meas_rqf_df, 1e-3 * col))

        results_table['Circuit Name'] = '%s.%s' % (self.circuit_type, self.circuit_name.split('.')[1])
        results_table['Circuit Family'] = 'RQ'
        results_table['Period'] = Reference.get_mp3_period(timestamp_fgc)
        results_table['Date (FGC)'] = Time.to_string_short(timestamp_fgc).split(' ')[0]
        results_table['Time (FGC)'] = Time.to_string_short(timestamp_fgc).split(' ')[1]
        results_table['Analysis performed by'] = author
        results_table['lhcsmapi version'] = lhcsmapi.__version__
        results_table['lhcsm notebook version'] = lhcsmapi.parse_lhcsmhwc_version(init_file_path)

        if not source_timestamp_nqps_df.empty:
            results_table['Delta_t(nQPS_RQD-PIC)'] = 1e-6 * (source_timestamp_nqps_df['timestamp'] - timestamp_pic)
            results_table['Delta_t(nQPS_RQF-PIC)'] = 1e-6 * (source_timestamp_nqps_df['timestamp'] - timestamp_pic)

        return results_table
