import warnings
from copy import deepcopy
from typing import Dict, List, Tuple

from sklearn.linear_model import LinearRegression
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from lhcsmapi.analysis.CircuitAnalysis import CircuitAnalysis
from lhcsmapi.analysis.busbar.LinReg import LinReg
from lhcsmapi.pyedsl.dbsignal.SignalAnalysis import SignalAnalysis
from lhcsmapi.pyedsl.AssertionBuilder import AssertionBuilder
from lhcsmapi.pyedsl.PlotBuilder import PlotBuilder
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.pyedsl.dbsignal.SignalIndexConversion import SignalIndexConversion
from lhcsmapi.pyedsl.ResistanceBuilder import ResistanceBuilder


def generate_busbar_status(beam_mode_with_pattern_series: pd.DataFrame, pattern: Dict[str, float]) -> pd.DataFrame:
    """ Function searching for provided beam mode patterns. It returns a DataFrame with beam mode indices and
    corresponding timestamps.

    :param beam_mode_with_pattern_series: a series containing all beam modes
    :param pattern: a particular beam mode pattern to be found
    :return: A DataFrame with timestamps corresponding to beam modes in the pattern
    """
    # find patterns long-enough
    beam_mode_spfr_ramp_pattern_df, _ = SignalAnalysis.find_rows_with_exact_pattern_in_series(
        beam_mode_with_pattern_series, pattern=list(pattern.values()))
    beam_mode_spfr_ramp_pattern_df['is_sb_long_enough'] = (beam_mode_spfr_ramp_pattern_df[14.0] -
                                                           beam_mode_spfr_ramp_pattern_df[8.0]) > 1800 * 1e9
    beam_mode_spfr_ramp_pattern_long_enough_df = beam_mode_spfr_ramp_pattern_df.query('is_sb_long_enough == True')

    t_setups = beam_mode_spfr_ramp_pattern_long_enough_df[2].values
    t_preramps = beam_mode_spfr_ramp_pattern_long_enough_df[6].values
    t_flattops = beam_mode_spfr_ramp_pattern_long_enough_df[8].values
    t_start_injs = np.array([max(t_s.item(), t_p.item() - int(1800 * 1e9)) for t_s, t_p in zip(t_setups, t_preramps)])
    t_end_injs = np.array([t_preramp.item() for t_preramp in t_preramps])
    t_start_sbs = np.array([t_flattop.item() for t_flattop in t_flattops])
    t_end_sbs = np.array([t_flattop.item() + int(1800 * 1e9) for t_flattop in t_flattops])
    busbar_df = pd.DataFrame({'t_start_injs': t_start_injs,
                              't_end_injs': t_end_injs,
                              't_start_sbs': t_start_sbs,
                              't_end_sbs': t_end_sbs})
    busbar_df.sort_values(by='t_start_injs', inplace=True)

    return busbar_df


def get_complete_busbar_df(t_start: int, t_end: int, pattern: Dict[str, float], spark) -> pd.DataFrame:
    """ Function querying a list of beam mode indices corresponding to the input pattern.

    :param t_start: start time of a search for a beam mode
    :param t_end: end time of a search for a beam mode
    :param pattern: pattern mapping from string to float
    :param spark: spark connector
    :return: DataFrame with results from NXCALS
    """
    # {'SETUP': 2.0, 'PRERAMP': 6.0, 'FLATTOP': 8.0, 'RAMPDOWN': 14.0}

    # query beam mode
    beam_mode_with_pattern_df = QueryBuilder().with_nxcals(spark) \
        .with_duration(t_start=t_start, t_end=t_end) \
        .with_query_parameters(nxcals_system='CMW', signal='HX:BMODE') \
        .signal_query() \
        .filter_values(list(pattern.keys())) \
        .map_values(pattern) \
        .dfs[0]

    # generate busbar status
    return generate_busbar_status(beam_mode_with_pattern_df["HX:BMODE"], pattern)


def extract_intersecting_plateaus(plateau_starts: List[List[int]], plateau_ends: List[List[int]]
                                  ) -> Tuple[List[int], List[int]]:
    """ Function extracting intersecting plateaus. It considers lists of list of plateau start and end. It returns a tuple
    with a list of start and end times of intersecting plateaus.

    :param plateau_starts: list of list of plateau starts (outer list per current; inner per plateau)
    :param plateau_ends: list of list of plateau ends (outer list per current; inner per plateau)
    :return: a tuple with a list of start and end times of intersecting plateaus
    """

    class Plateau(object):
        def __init__(self, plateau_starts, plateau_ends):
            self.plateau_starts = plateau_starts
            self.plateau_ends = plateau_ends

        def count_overlap(self, t_start, t_end):
            count = 0
            for plateau_start, plateau_end in zip(self.plateau_starts, self.plateau_ends):
                for plateau_start_el, plateau_end_el in zip(plateau_start, plateau_end):
                    if min(t_end, plateau_end_el) > max(t_start, plateau_start_el):
                        count += 1

            return count

    p = Plateau(plateau_starts, plateau_ends)
    plateau_starts_shared = []
    plateau_ends_shared = []

    # Find overlapping intervals

    # For each of plateau list find overlapping intervals
    # take the intersection of the overlapping intervals (max t_start and min t_end)
    # take only an interval that overlaps with all intervals, the overlap count should be equal to the number of lists
    for plateau_start, plateau_end in zip(plateau_starts, plateau_ends):
        plateau_start_shared = []
        plateau_end_shared = []
        for plateau_start_el, plateau_end_el in zip(plateau_start, plateau_end):
            if p.count_overlap(plateau_start_el, plateau_end_el) == len(plateau_starts):
                plateau_start_shared.append(plateau_start_el)
                plateau_end_shared.append(plateau_end_el)
        plateau_starts_shared.append(plateau_start_shared)
        plateau_ends_shared.append(plateau_end_shared)

    # Take only the latest plateau start and the earliest plateau end
    return list(np.array(plateau_starts_shared).max(axis=0)), list(np.array(plateau_ends_shared).min(axis=0))


def find_start_end_of_the_longest_ramp_up(plateau_start: List[int], plateau_end: List[int]) -> Tuple[int, int]:
    """ Function finding the start and end of the longest ramp up

    :param plateau_start: list of current plateau starts (in ns)
    :param plateau_end: list of current plateau ends (in ns)
    :return: start and end of the longest timestamp
    """
    # calculate differences between plateau_start[i+1] and plateau_end[i] for i=0...len(plateau_start)-2
    ramp_up_duration = []
    for i in range(len(plateau_start)-1):
        ramp_up_duration.append(plateau_start[i+1] - plateau_end[i])

    # find the index of the maximum ramp up
    index_max_duration = ramp_up_duration.index(max(ramp_up_duration))

    return plateau_end[index_max_duration], plateau_start[index_max_duration+1]


class BusbarResistanceAnalysis(CircuitAnalysis):
    """Base class for busbar resistance analysis

    """

    @staticmethod
    def convert_to_col(res_row_df: pd.DataFrame, signal_name: str) -> pd.DataFrame:
        """ Method converting a resistance value as a row into a column

        :param res_row_df: tow with resistance values
        :param signal_name: voltage signal name (e.g., U_RES)
        :return: DataFrame with resistances as a row
        """
        res_row_df = res_row_df.dropna(axis=1)
        res_row_df.index = [signal_name.replace('U', 'R')]
        res_row_df = res_row_df.T
        res_row_df.index = [index.split('_')[0] + ':' + signal_name for index in res_row_df.index]
        res_row_df['nxcals_variable_name'] = res_row_df.index
        res_row_df.reset_index(inplace=True, drop=True)

        return res_row_df

    @staticmethod
    def calculate_current_plateau_start_end(i_meas_raw_dfs: List[pd.DataFrame], *, i_meas_threshold=50,
                                            min_duration_in_sec=1, time_shift_in_sec=(0,0)
                                            ) -> Tuple[List[int], List[int]]:
        """ Method calculating a list of intersecting plateaus across a list of power converter currents.

        :param i_meas_raw_dfs: list of power converter currents for plateau calculation
        :param i_meas_threshold: minimum power converter current
        :param min_duration_in_sec: minimum plateau duration in seconds
        :param time_shift_in_sec: tuple with the time subtracted from the start (first element) and end (second element)
        of a plateau
        :return: a tuple with a list of start and end times of intersecting plateaus
        """

        # Find plateau start and end based on the current profile
        plateau_starts = []
        plateau_ends = []
        for i_meas_raw_df in i_meas_raw_dfs:
            if i_meas_raw_df.max().values[0] < 100 and len(i_meas_raw_dfs) > 3: # For IT it is OK
                print('There is at least one signal with the peak current below 100 A.')
                return [], []

        for i_meas_raw_df in i_meas_raw_dfs:
            if i_meas_raw_df.max().values[0] > i_meas_threshold:
                plateau_start, plateau_end = BusbarResistanceAnalysis.\
                    find_plateau_start_and_end(i_meas_raw_df, i_meas_threshold=i_meas_threshold,
                                               min_duration_in_sec=min_duration_in_sec,
                                               time_shift_in_sec=time_shift_in_sec)
                if plateau_start is None:
                    print('\tThe signal does not contain ramp-up!')
                    return [], []

                plateau_starts.append(plateau_start)
                plateau_ends.append(plateau_end)

        max_len_plateau_starts = max(len(plateau_start) for plateau_start in plateau_starts)
        max_len_plateau_ends = max(len(plateau_end) for plateau_end in plateau_ends)
        if max_len_plateau_starts != max_len_plateau_ends:
            return [], []

        return extract_intersecting_plateaus(plateau_starts, plateau_ends)

    @staticmethod
    def find_plateau_start_and_end(i_meas_raw_df: pd.DataFrame, i_meas_threshold=500, min_duration_in_sec=1,
                                   time_shift_in_sec=(0,0)) -> Tuple[List[int], List[int]]:
        """ Method finding a tuple of two lists with starts and ends of current plateaus

        :param i_meas_raw_dfs: list of power converter currents for plateau calculation
        :param i_meas_threshold: minimum power converter current
        :param min_duration_in_sec: minimum plateau duration in seconds
        :param time_shift_in_sec: tuple with the time subtracted from the start (first element) and end (second element)
        of a plateau
        :return: a tuple with a list of start and end times of intersecting plateaus
        """
        if sum(time_shift_in_sec) > min_duration_in_sec:
            raise ValueError('The sum of time shift {} is greater than the minimum duration in sec {} s'
                             .format(time_shift_in_sec, min_duration_in_sec))

        i_meas_df = deepcopy(i_meas_raw_df)
        I_MEAS = i_meas_df.columns[0]

        i_meas_df['acqStamp'] = i_meas_df.index
        i_meas_df = SignalIndexConversion.synchronize_df(i_meas_df)
        i_meas_df = SignalIndexConversion.convert_indices_to_sec(i_meas_df)
        i_meas_df = SignalAnalysis.resample(i_meas_df, ts=1)
        i_meas_df['acqStamp'] = i_meas_df['acqStamp'].apply(lambda row: int(row))

        # take the current only above current threshold in [A]
        i_meas_df = i_meas_df[i_meas_df[I_MEAS] >= i_meas_threshold]

        # calculate time derivative
        i_meas_df.loc[:, 'time'] = i_meas_df.index
        i_meas_df['dI_MEAS'] = i_meas_df[I_MEAS].diff()
        i_meas_df['dtime'] = i_meas_df['time'].diff()
        i_meas_df = i_meas_df.dropna()
        i_meas_df['dI_MEAS_dt'] = i_meas_df.apply(lambda col: col['dI_MEAS'] / col['dtime'], axis=1)

        # round to the nearest integer
        i_meas_df['round_dI_MEAS_dt'] = i_meas_df.apply(lambda row: np.around(row['dI_MEAS_dt'], decimals=1), axis=1)

        # get sign of the round value
        i_meas_df['sign_dI_MEAS_dt'] = i_meas_df.apply(lambda row: np.sign(row['round_dI_MEAS_dt']), axis=1)

        # to account for ramp-up and ramp-down, replace -1 with 1 (now 1 denotes a current change, and 0 denotes plateau)
        i_meas_df['sign_dI_MEAS_dt'] = i_meas_df['sign_dI_MEAS_dt'].apply(lambda col: 1 if col == -1 else col)

        # Find change of derivative
        _, plateau_start_pattern_idx = SignalAnalysis.find_rows_with_exact_pattern_in_series(
            i_meas_df['sign_dI_MEAS_dt'], [1, 0])

        _, plateau_end_pattern_idx = SignalAnalysis.find_rows_with_exact_pattern_in_series(
            i_meas_df['sign_dI_MEAS_dt'], [0, 1])

        # plateau start, plateau_end
        plateau_start = list(i_meas_df.loc[i_meas_df.index[plateau_start_pattern_idx[1::2]], 'acqStamp'].values)
        plateau_end = list(i_meas_df.loc[i_meas_df.index[plateau_end_pattern_idx[1::2]], 'acqStamp'].values)

        if plateau_end[0] < plateau_start[0]:
            plateau_start = [i_meas_df['acqStamp'].values[0]] + plateau_start

        if len(plateau_start) > len(plateau_end):
            plateau_end.append(i_meas_raw_df.index[-1])

        if len(plateau_start) < len(plateau_end):
            plateau_start = i_meas_raw_df.index[0].values + plateau_start

        plateau_start_duration = []
        plateau_end_duration = []

        for plateau_start_el, plateau_end_el in zip(plateau_start, plateau_end):
            if (plateau_end_el - plateau_start_el) / 1e9 > min_duration_in_sec:
                plateau_start_duration.append(plateau_start_el + int(time_shift_in_sec[0]*1e9))
                plateau_end_duration.append(plateau_end_el - int(time_shift_in_sec[1]*1e9))

        return plateau_start_duration, plateau_end_duration

    @staticmethod
    def analyze_busbar_magnet_resistance(res_df: pd.DataFrame, signal_name: str, value_max: float) -> pd.DataFrame:
        """ Method analyzing busbar or magnet resistances by comparing them to a maximum allowed value

        :param res_df: DataFrame with resistances (column format)
        :param signal_name: name of resistance signal
        :param value_max: maximum allowed value
        :return: a subset of the input DataFrame with resistances outside of the allowed range
        """
        features_outside_range_df = AssertionBuilder().with_feature(res_df) \
            .has_min_max_value(feature=signal_name, value_min=0, value_max=value_max) \
            .show_plot(xlabel='Busbar, [-]', ylabel=('%s (Calculated), [Ohm]' % signal_name)) \
            .get_features_outside_range()

        if not features_outside_range_df.empty:
            warnings.warn('Due to the limited accuracy of the measurement it cannot be excluded that '
                          'one or more joint resistances in the circuit are larger than 1 nOhm.', stacklevel=1)

        return features_outside_range_df

    @staticmethod
    def analyze_busbar_magnet_voltage(u_res_df: pd.DataFrame, value_max=5e-3, ylabel='', title='') -> pd.DataFrame:
        """ Method analyzing busbar or magnet voltages by comparing to an allowed maximum

        :param u_res_df: DataFrame with voltages
        :param value_max: maximum allowed value
        :param ylabel: label for the y- axis
        :param title: plot title
        :return: DataFrame with voltages above the allowed maximum
        """
        return AssertionBuilder().with_feature(u_res_df) \
            .has_min_max_value(feature='max_abs', value_min=0, value_max=value_max) \
            .show_plot(xlabel='Busbar, [-]', ylabel=ylabel, title=title) \
            .get_features_outside_range()

    @staticmethod
    def check_signal_count(u_df: pd.DataFrame, expected_signal_count: int, signal: str) -> None:
        """ Method checking the signal count,i.e., if there are all signals logged in the database.

        :param u_df: DataFrame with an signals to count
        :param expected_signal_count: expected count of signals
        :param signal: signal name (for print out)
        :return: None
        """
        non_nan_signal_count = len(u_df.dropna())
        if non_nan_signal_count < expected_signal_count:
            warnings.warn('There were missing %s signals. Found %d signals, expected %d.' %
                          (signal, non_nan_signal_count, 156))
        else:
            print('All (%d) %s signals were found.' % (expected_signal_count, signal))

    @staticmethod
    def plot_i_meas_u_res_current_plateau(i_meas_df: pd.DataFrame, u_res_df: pd.DataFrame, t0: int,
                                          plateau_start: List[int], plateau_end: List[int], title='') -> None:
        """ Method plotting current and voltage measurement along with current plateaus. Used for analyzing cases with
        resistances outside of range.

        :param i_meas_df: main power converter current signal
        :param u_res_df: magnet or busbar voltage
        :param t0: start time for plateau synchronization (in ns)
        :param plateau_start: list with starts of current plateaus (in ns)
        :param plateau_end: list with ends of current plateaus (in ns)
        :param title: plot title
        :return: None
        """
        ax = PlotBuilder().with_signal(i_meas_df, title=title, grid=True) \
            .with_ylabel(ylabel='I_MEAS, [A]') \
            .with_signal(u_res_df) \
            .with_ylabel(ylabel='U_RES, [V]') \
            .plot(show_plot=False).get_axes()[0]

        for ps, pe in zip(plateau_start, plateau_end):
            ax.axvspan(xmin=(ps - t0) / 1e9,
                       xmax=(pe - t0) / 1e9, facecolor='xkcd:orange')
        plt.show()


class ItBusbarResistanceAnalysis(BusbarResistanceAnalysis):
    """ Class with busbar calculation for IT circuit

    """
    @staticmethod
    def calculate_resistance(i_meas_feature_df: pd.DataFrame, u_mag_feature_df: pd.DataFrame) -> pd.DataFrame:
        """ Method for calculating resistance based on current and voltage features

        :param i_meas_feature_df: current features
        :param u_mag_feature_df: voltage features
        :return: DataFrame with resistances
        """
        pd.set_option('mode.chained_assignment', None)
        i_meas_feature_df = i_meas_feature_df[['device', 'class', 'mean', 'std']]

        i_rqx_df = i_meas_feature_df[i_meas_feature_df['device'].str.contains('RQX')]
        i_rtqx1_df = i_meas_feature_df[i_meas_feature_df['device'].str.contains('RTQX1')]
        i_rtqx2_df = i_meas_feature_df[i_meas_feature_df['device'].str.contains('RTQX2')]

        # Calculate I_Q1
        I_Q1_1 = i_rqx_df[i_rqx_df['class'] == 1]['mean'].values[0] + \
                 i_rtqx1_df[i_rtqx1_df['class'] == 1]['mean'].values[0]
        i_meas_feature_df.loc[len(i_meas_feature_df)] = ['I_Q1', 1, I_Q1_1, 0]

        I_Q1_2 = i_rqx_df[i_rqx_df['class'] == 2]['mean'].values[0] + \
                 i_rtqx1_df[i_rtqx1_df['class'] == 2]['mean'].values[0]
        i_meas_feature_df.loc[len(i_meas_feature_df)] = ['I_Q1', 2, I_Q1_2, 0]

        # Calculate I_Q2
        I_Q2_1 = i_rqx_df[i_rqx_df['class'] == 1]['mean'].values[0] + \
                 i_rtqx2_df[i_rtqx2_df['class'] == 1]['mean'].values[0]
        i_meas_feature_df.loc[len(i_meas_feature_df)] = ['I_Q2', 1, I_Q2_1, 0]

        I_Q2_2 = i_rqx_df[i_rqx_df['class'] == 2]['mean'].values[0] + \
                 i_rtqx2_df[i_rtqx2_df['class'] == 2]['mean'].values[0]
        i_meas_feature_df.loc[len(i_meas_feature_df)] = ['I_Q2', 1, I_Q2_2, 0]

        # Calculate I_Q3
        I_Q3_1 = i_rqx_df[i_rqx_df['class'] == 1]['mean'].values[0]
        i_meas_feature_df.loc[len(i_meas_feature_df)] = ['I_Q3', 1, I_Q3_1, 0]

        I_Q3_2 = i_rqx_df[i_rqx_df['class'] == 2]['mean'].values[0]
        i_meas_feature_df.loc[len(i_meas_feature_df)] = ['I_Q3', 2, I_Q3_2, 0]

        i_q_feature_df = i_meas_feature_df[i_meas_feature_df['device'].str.contains('I_Q')]
        pd.set_option('mode.chained_assignment', 'warn')

        res = {}
        u_mag_feature_df = u_mag_feature_df[(u_mag_feature_df['class'] > 0) & (u_mag_feature_df['class'] < 3)]
        for magnet_name in ['Q1', 'Q2', 'Q3']:
            u_mean = u_mag_feature_df[u_mag_feature_df['nxcals_variable_name'].str.contains(magnet_name)].sort_values(
                by='class')['mean'].values
            i_mean = i_q_feature_df[i_q_feature_df['device'] == 'I_%s' % magnet_name].sort_values(by='class')[
                'mean'].values
            try:
                res['R_%s' % magnet_name] = abs(LinearRegression().fit(i_mean.reshape(-1, 1), u_mean).coef_[0])
            except ValueError as ve:
                warnings.warn(str(ve))
                res['R_%s' % magnet_name] = float('nan')

        return pd.DataFrame({'R_RES': res})


class SingleBusbarResistanceAnalysis(BusbarResistanceAnalysis):
    """ Class with a method for calculating a single busbar based on linear regression and error estimation

    """
    @staticmethod
    def calculate_splice_resistance_linreg_with_errors(u_res_df: pd.DataFrame, i_meas_df: pd.DataFrame,
                                                       plateau_start: List[int], plateau_end: List[int]
                                                       ) -> float:
        """ Method calculating splice resistance based on voltage and current signals during plateaus. In addition,
        estimation errors are displayed.

        :param u_res_df: voltage signal
        :param i_meas_df: current signal
        :param plateau_start: list of plateau starts
        :param plateau_end: list of plateau ends
        :return: resistance
        """
        # Merge both signals
        voltage_col = u_res_df.columns[0]
        current_col = i_meas_df.columns[0]
        idx = u_res_df.index.union(i_meas_df.index)
        u_i_merged = u_res_df.reindex(idx).interpolate('index')
        u_i_merged = u_i_merged.join(i_meas_df.reindex(idx).interpolate('index')).dropna()

        # Take only constant current part
        mask = np.zeros((len(u_i_merged.index), 1))
        for ps, pe in zip(plateau_start, plateau_end):
            mask_index = (u_i_merged.index >= ps) & (u_i_merged.index <= pe)
            mask = mask + mask_index.reshape((len(u_i_merged.index), 1)).astype(int)

        u_i_plateau_df = u_i_merged[mask.squeeze().astype(bool)]

        # Calculate resistance as a linear fit
        x = u_i_plateau_df[current_col].values
        y = u_i_plateau_df[voltage_col].values
        x_intercept = LinReg.add_intercept_to_x(x)
        linreg = LinReg().fit(x_intercept, y)
        y_hat = linreg.predict(x_intercept)
        theta_error = LinReg.estimate_linear_fit_error(x_intercept, y, y_hat)

        u_i_plateau_df['fit'] = u_i_plateau_df[current_col] * linreg.theta[1] + linreg.theta[0]

        ax = u_i_plateau_df.plot.scatter(x=current_col, y=voltage_col, figsize=(15, 7))
        u_i_plateau_df.plot.line(x=current_col, y='fit', ax=ax, grid=True, color='red')
        ax.legend(['fit', voltage_col])
        ax.set_xlabel('I_MEAS, [A]')
        ax.set_ylabel('U_RES, [V]')

        print('The resistance estimate from the linear regression is %e Ohm +/- %e Ohm.' %
              (linreg.theta[1], theta_error[1]))

        return linreg.theta[1]

    @staticmethod
    def calculate_splice_resistance_linreg(u_res_df: pd.DataFrame, i_meas_df: pd.DataFrame, plateau_start: List[int],
                                           plateau_end: List[int]) -> float:
        """ Method calculating splice resistance based on voltage and current signals during plateaus

        :param u_res_df: voltage signal
        :param i_meas_df: current signal
        :param plateau_start: list of plateau starts
        :param plateau_end: list of plateau ends
        :return: resistance
        """
        # Merge both signals
        voltage_col = u_res_df.columns[0]
        current_col = i_meas_df.columns[0]
        idx = u_res_df.index.union(i_meas_df.index)
        u_i_merged = u_res_df.reindex(idx).interpolate('index')
        u_i_merged = u_i_merged.join(i_meas_df.reindex(idx).interpolate('index')).dropna()

        # Take only constant current part
        mask = np.zeros((len(u_i_merged.index), 1))
        for ps, pe in zip(plateau_start, plateau_end):
            mask_index = (u_i_merged.index >= ps) & (u_i_merged.index <= pe)
            mask = mask + mask_index.reshape((len(u_i_merged.index), 1)).astype(int)

        u_i_plateau_df = u_i_merged[mask.squeeze().astype(bool)]

        # Calculate resistance as a linear fit
        linreg = LinearRegression().fit(np.array(u_i_plateau_df[current_col]).reshape((-1, 1)),
                                        u_i_plateau_df[voltage_col])

        u_i_plateau_df['fit'] = u_i_plateau_df[current_col] * linreg.coef_[0] + linreg.intercept_
        ax = u_i_plateau_df.plot.scatter(x=current_col, y=voltage_col, figsize=(15, 7))
        u_i_plateau_df.plot.line(x=current_col, y='fit', ax=ax, grid=True, color='red')
        ax.legend(['fit', voltage_col])
        ax.set_xlabel('I_MEAS, [A]')
        ax.set_ylabel('U_RES, [V]')

        return linreg.coef_[0]


class MultipleBusbarResistanceAnalysis(BusbarResistanceAnalysis):
    """ Method for calculating multiple busbars (e.g., RB, RQ)

    """
    def calculate_resistance(self, i_meas_feature_df: pd.DataFrame, u_res_feature_df: pd.DataFrame, signal_name: str,
                             t_start: int, circuit_name: str) -> pd.DataFrame:
        """ Method calculating multiple busbars with mean current and voltage per current plateau.

        :param i_meas_feature_df: mean power converter current
        :param u_res_feature_df: mean busbar/magnet voltage
        :param signal_name: signal name
        :param t_start: start time for an output resistance index
        :param circuit_name: name of the circuit to sort resistances according to their tunnel location
        :return: a column DataFrame with resistance values
        """
        return ResistanceBuilder().with_busbar_voltage(u_res_feature_df) \
            .with_busbar_current(i_meas_feature_df) \
            .calculate_mean_resistance(self.circuit_type) \
            .convert_to_row(index=t_start) \
            .fill_missing_columns(signal_name=signal_name.replace('U', 'R'), circuit_type=self.circuit_type) \
            .sort_columns_busbar_location(signal_name=signal_name.replace('U', 'R'), circuit_type=self.circuit_type,
                                          circuit_name=circuit_name, split='_') \
            .df
