from typing import Tuple, List

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from pyspark.sql.types import IntegerType
import pyspark.sql.functions as F
from IPython.display import display, HTML

from lhcsmapi.Time import Time
from lhcsmapi.analysis.CircuitQuery import CircuitQuery, execution_count
from lhcsmapi.analysis.busbar.BusbarResistanceAnalysis import BusbarResistanceAnalysis, extract_intersecting_plateaus, \
    find_start_end_of_the_longest_ramp_up
from lhcsmapi.analysis.expert_input import check_show_next
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder


class BusbarResistanceQuery(CircuitQuery):
    """ Class for query of busbar voltage and current to calculate resistance

    """

    @execution_count()
    def query_busbar_features_nxcals(self, t_start, t_end=None, duration=None, *, signal='U_RES',
                                     features=('mean', 'std', 'max', 'min', 'count'), spark) -> pd.DataFrame:
        """ Method querying NXCALS to compute features of a busbar signal

        :param t_start: start time of a query
        :param t_end: end time of a query (optional)
        :param duration: duration w.r.t. t_start (optional)
        :param t0: synchronization time for a signal
        :param signal: signal name to query (scalar)
        :param features: list of features to query
        :param spark: spar connector
        :return: DataFrame with columns corresponding to the features of the selected busbar signal
        """
        t_start, t_end = Time.get_query_period_in_unix_time(t_start, t_end, duration, '')
        return QueryBuilder() \
            .with_nxcals(spark) \
            .with_duration(t_start=t_start, t_end=t_end) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='BUSBAR', signal=signal, wildcard={'BUSBAR': '*'}) \
            .feature_query(features) \
            .calculate_max_abs() \
            .sort_busbar_location(self.circuit_type, self.circuit_name) \
            .df

    @execution_count()
    def calculate_current_plateau_start_end(self, t_start, t_end, *, i_meas_threshold=50, min_duration_in_sec=1,
                                            time_shift_in_sec=(0,0), spark) -> Tuple[List[int], List[int]]:
        """ Method querying and calculating current plateau start and end intervals

        :param t_start: start time for the query
        :param t_end: end tim for the query
        :param i_meas_threshold: minimum current to be considered for plateau calculation
        :param min_duration_in_sec: minimum duration of a plateau
        :param time_shift_in_sec: time shift in seconds from the start and the end of a plateau
        :param spark: spark connector
        :return: tuple of start and end plateau timestamps (each as a list)
        """

        i_meas_raw_dfs = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=t_start, t_end=int(t_end)) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PC', signal='I_MEAS') \
            .signal_query() \
            .dfs

        return BusbarResistanceAnalysis\
            .calculate_current_plateau_start_end(i_meas_raw_dfs, i_meas_threshold=i_meas_threshold,
                                                 min_duration_in_sec=min_duration_in_sec,
                                                 time_shift_in_sec=time_shift_in_sec)

    @execution_count()
    def get_busbar_resistances(self, t_start, t_end, plateau_start: List[int], plateau_end: List[int],
                               *, signal_name='U_RES', system='BUSBAR', spark) -> pd.DataFrame:
        """ Method calculating, with the NXCALS cluster, busbar resistances between start and end time as well sa
        current plateaus.

        :param t_start: query start time (in unix time)
        :param t_end: query end time (in unix time)
        :param plateau_start: list of starts of plateaus
        :param plateau_end: list of ends of plateaus
        :param signal_name: name of a voltage signal to query
        :param system: system to query
        :param spark: spark connector
        :return: DataFrame with resistances
        """

        plateau_start = np.array(plateau_start)
        plateau_end = np.array(plateau_end)
        ramp_up_start, ramp_up_end = find_start_end_of_the_longest_ramp_up(plateau_start, plateau_end)

        def translate(timestamp):
            mask_plateau = (timestamp >= plateau_start) & (timestamp <= plateau_end)
            index_plateau = np.where(mask_plateau == True)[0]

            if len(index_plateau) > 0:
                return int((index_plateau[0] + 1))

            if (timestamp > ramp_up_start) and (timestamp < ramp_up_end):
                return 0

            return -1

        plateau_translate_udf = F.udf(translate, IntegerType())

        # query current
        i_meas_feature_df = QueryBuilder() \
            .with_nxcals(spark) \
            .with_duration(t_start=int(t_start), t_end=int(t_end)) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PC', signal='I_MEAS') \
            .feature_query(['mean', 'std'], function=plateau_translate_udf) \
            .sort_values(by='class') \
            .df

        # query voltage
        u_res_feature_df = QueryBuilder() \
            .with_nxcals(spark) \
            .with_duration(t_start=int(t_start), t_end=int(t_end)) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system,
                           signal=signal_name, wildcard={'BUSBAR': '*'}) \
            .feature_query(['mean', 'std'], function=plateau_translate_udf) \
            .sort_busbar_location(self.circuit_type, circuit_name=self.circuit_name) \
            .correct_voltage_sign() \
            .df

        return u_res_feature_df, i_meas_feature_df

    @staticmethod
    def query_and_plot_outlier_voltage(res_outliers_df: pd.DataFrame, t_start, t_end, t_0: int,
                                       plateau_start: List[int], plateau_end: List[int],
                                       *, is_automatic=False, spark) -> None:
        """ Method querying and plotting outlier voltages. Outliers are provided from the resistance assertion.

        :param res_outliers_df: DataFrame with outlier resistances
        :param t_start: start time of a query
        :param t_end: end time of a query
        :param t_0: synchronization time for current plateaus
        :param plateau_start: start of current plateaus (for plotting)
        :param plateau_end: end of current plateaus (for plotting)
        :param is_automatic: flag for looping through the outliers
        :param spark: spark connector
        :return: None
        """
        if not res_outliers_df.empty:
            print('Resistance(s) outside of range:')
            display(HTML(res_outliers_df.to_html()))
        else:
            print('All resistances within the range.')

        for index, row in res_outliers_df.iterrows():

            u_df = QueryBuilder().with_nxcals(spark) \
                .with_duration(t_start=t_start, t_end=t_end) \
                .with_query_parameters(nxcals_system='CMW', signal=row['nxcals_variable_name']) \
                .signal_query() \
                .synchronize_time() \
                .convert_index_to_sec().dfs[0]

            if not u_df.empty:
                ax = u_df.plot(figsize=(15, 7), grid=True)
                # Add ranges of detected plateaus in orange
                for ps, pe in zip(plateau_start, plateau_end):
                    ax.axvspan(xmin=(ps - Time.to_unix_timestamp(t_0)) / 1e9,
                               xmax=(pe - Time.to_unix_timestamp(t_0)) / 1e9,
                               facecolor='xkcd:goldenrod')
                ax.set_xlabel("time, [s]", fontsize=15)
                ax.set_ylabel("U, [V]", fontsize=15)
                ax.tick_params(labelsize=15)
                plt.show()

            if check_show_next(index, len(res_outliers_df), is_automatic=is_automatic):
                break
