from typing import Union, List, Tuple

import pandas as pd
from tqdm.notebook import tqdm

from lhcsmapi.analysis.CircuitQuery import CircuitQuery, execution_count
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.Time import Time
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder


def get_magnet_type_i_meas_u_diode(magnet: str, circuit_name: str) -> Tuple[str, str, str]:
    """ Method returning a tuple of magnet type (MB or MQ), I_MEAS and U_DIODE name for a given magnet and circuit name.

    :param magnet: magnet name
    :param circuit_name: circuit name
    :return: tuple of manget_type, I_MEAS and U_DIODE names per circuit tyoe (RB, RQ)
    """

    circuit_type = circuit_name[0:2]
    magnet_type = "M" + circuit_type[1:2]

    i_meas = SignalMetadata.get_signal_name(circuit_type, circuit_name, system="PC", database="CALS", signal="I_MEAS")
    u_diode = SignalMetadata.get_signal_name(circuit_type, circuit_name, database='CALS',
                                             system="DIODE_{}".format(circuit_name.split('.')[0]),
                                             signal="U_DIODE_{}".format(circuit_name.split('.')[0]))
    u_diode = u_diode.replace("%MAGNET%", "{}.{}".format(magnet_type, magnet))

    return magnet_type, i_meas, u_diode


class DiodeLeadResistanceQuery(CircuitQuery):
    """Class containing methods for query of events and signals for diode lead resistance calculation.

    """

    def find_source_timestamp_nqps_pm(self, source_qds: str, timestamp_qds: int, duration=[(600, 's'), (600, 's')],
                                       ) -> pd.DataFrame:
        """ Method searching PM nQPS events.
        If the timestamp is NaN, then an empty dataframe is returned.

        :param source_qds: PM QDS source according to metadata
        :type source_qds: str
        :param timestamp_qds: PM QDS timestamp in ns precision
        :type timestamp_qds: int
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: pd.DataFrame with source and timestamp of the PM nQPS events
        """
        qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, source_qds)
        system = 'DIODE_{}'.format(self.circuit_name.split('.')[0])

        return QueryBuilder().with_pm() \
            .with_duration(t_start=timestamp_qds, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, source=qps_crate) \
            .event_query(verbose=self.verbose).df

    @execution_count()
    def find_source_timestamp_nqps(self, t_start: Union[int, str, float], duration=[(10, 's'), (400, 's')],
                                   ) -> pd.DataFrame:
        """ Method searching PM nQPS events for diode query
        If the start is NaN, then an empty dataframe is returned.

        :param t_start: start time of a search
        :type t_start: Union[str, int, float]
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: pd.DataFrame with source and timestamp of the PM nQPS events
        """
        system = 'DIODE_{}'.format(self.circuit_name.split('.')[0])
        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, source='*') \
            .event_query(verbose=self.verbose) \
            .filter_source(self.circuit_name, 'QDS') \
            .drop_duplicate_source() \
            .sort_values(by='timestamp').df

    def _query_i_u_diode_pm(self, circuit_name: str, source_qds: str, timestamp_fgc: int, timestamp_qds_a: int,
                            timestamp_qds_b: int) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """ Method querying PM for current and voltage of a diode
        Signals are **not** synchronized to the input timestamp and index is **not** converted to seconds.

        :param circuit_name: name of a circuit to query
        :type circuit_name: str
        :param source_qds: PM QDS source according to metadata
        :type source_qds: str
        :param timestamp_fgc: FGC PM event timestamp for I_A query (ns precision)
        :type timestamp_fgc: int
        :param timestamp_qds_a: QDS Board A PM event timestamp (ns precision)
        :type timestamp_qds_a: int
        :param timestamp_qds_b: QDS Board B PM event timestamp (ns precision)
        :type timestamp_qds_b: int
        :return: tuple of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        i_a_rqd_df: pd.DataFrame = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_fgc) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system='PC', signal='I_A') \
            .signal_query().dfs[0]

        qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, source_qds)
        system = 'DIODE_{}'.format(circuit_name.split('.')[0])
        signal = 'U_DIODE_{}'.format(circuit_name.split('.')[0])
        wildcard = 'MB.{}'.format(source_qds) if self.circuit_type == 'RB' else 'MQ.{}'.format(source_qds)

        u_diode_a_rqd_df: pd.DataFrame = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_qds_a) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system=system, signal=signal, source=qps_crate,
                           wildcard={'MAGNET': wildcard}) \
            .signal_query().dfs[0]

        u_diode_b_rqd_df: pd.DataFrame = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_qds_b) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system=system, signal=signal, source=qps_crate,
                           wildcard={'MAGNET': wildcard}) \
            .signal_query().dfs[0]

        u_ref_rqd_df: pd.DataFrame = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_qds_a) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system=system, signal='U_REF_N1', source=qps_crate,
                           wildcard={'QPS_CRATE': qps_crate}) \
            .signal_query().dfs[0]

        return i_a_rqd_df, u_diode_a_rqd_df, u_diode_b_rqd_df, u_ref_rqd_df

    def _query_i_u_diode_nxcals(self, circuit_name: str, source_qds: str, timestamp_qds: int, *, spark,
                                duration) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """ Method querying NXCALS for current and voltage of a diode
        Signals are synchronized to the input timestamp and index is converted to seconds.

        :param circuit_name: name of a circuit to query
        :type circuit_name: str
        :param source_qds: PM QDS source according to metadata
        :type source_qds: str
        :param timestamp_qds: QDS Board A PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_qds: int
        :param spark: spark session variable needed to perform an NXCALS query
        :type spark: spark session
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: tuple of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        i_meas_diode_df: pd.DataFrame = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=timestamp_qds, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system='PC', signal='I_MEAS') \
            .signal_query() \
            .synchronize_time(timestamp_qds) \
            .convert_index_to_sec().dfs[0]

        system = 'DIODE_{}'.format(circuit_name.split('.')[0])
        signal = 'U_DIODE_{}'.format(circuit_name.split('.')[0])
        wildcard = 'MB.{}'.format(source_qds) if self.circuit_type == 'RB' else 'MQ.{}'.format(source_qds)

        u_diode_rqd_df: pd.DataFrame = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=timestamp_qds, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system=system, signal=signal, wildcard={'MAGNET': wildcard}) \
            .signal_query() \
            .synchronize_time(timestamp_qds) \
            .convert_index_to_sec().dfs[0]

        return i_meas_diode_df, u_diode_rqd_df

    @execution_count()
    def query_current_voltage_diode_leads_pm(self, timestamp_fgc: int, source_timestamp_qds_df: pd.DataFrame) \
            -> List[Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
        """ Method querying PM for current and voltage of a diode

        :param timestamp_fgc: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_fgc: int
        :param source_timestamp_qds_df: source timestamp qds table
        :type source_timestamp_qds_df: pd.DataFrame
        :return: List of tuples of current and voltage pd.DataFrames
        """
        i_meas_u_diode_u_ref_pm_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying U, I Diode PM'):
            source_qds, timestamp_qds = row[['source', 'timestamp']]
            u_diode_db = 'PM' if 1e-9 * (timestamp_qds - timestamp_fgc) < 2 else 'NXCALS'

            if u_diode_db == 'PM':
                source_timestamp_nqps_df = self.find_source_timestamp_nqps_pm(source_qds, timestamp_qds)

                if not source_timestamp_nqps_df.empty:
                    timestamp_nqps_board_a = source_timestamp_nqps_df.loc[0, 'timestamp']
                    timestamp_nqps_board_b = source_timestamp_nqps_df.loc[1, 'timestamp']
                    i_meas_u_diode_u_ref_pm_df = self._query_i_u_diode_pm(self.circuit_name, source_qds, timestamp_fgc,
                                                                          timestamp_nqps_board_a,
                                                                          timestamp_nqps_board_b)
                else:
                    i_meas_u_diode_u_ref_pm_df = (pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame)
            else:
                i_meas_u_diode_u_ref_pm_df = (pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame)

            i_meas_u_diode_u_ref_pm_dfs.append(i_meas_u_diode_u_ref_pm_df)

        return i_meas_u_diode_u_ref_pm_dfs

    @execution_count()
    def query_current_voltage_diode_leads_nxcals(self, source_timestamp_qds_df: pd.DataFrame, *, spark,
                                                 duration=[(50, 's'), (150, 's')]) -> List[
        Tuple[pd.DataFrame, pd.DataFrame]]:
        """ Method querying NXCALS for current and voltage of a diode

        :param source_timestamp_qds_df: analysis results table
        :type source_timestamp_qds_df: pd.DataFrame
        :param spark: spark session variable needed to perform an NXCALS query
        :type spark: spark session
        :return: List of tuples of current and voltage pd.DataFrames
        """
        i_meas_u_diode_nxcals_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying U, I Diode leads NXCALS'):
            source_qds, timestamp_qds = row[['source', 'timestamp']]

            timestamp_qds = Time.to_unix_timestamp(timestamp_qds)
            i_meas_u_diode_nxcals_df = self._query_i_u_diode_nxcals(self.circuit_name, source_qds,
                                                                    timestamp_qds, spark=spark, duration=duration)

            i_meas_u_diode_nxcals_dfs.append(i_meas_u_diode_nxcals_df)

        return i_meas_u_diode_nxcals_dfs
