import warnings
from typing import List

import numpy as np
import pandas as pd
from IPython.display import display, HTML
import matplotlib.pyplot as plt

from lhcsmapi.Time import Time
from lhcsmapi.analysis.CircuitAnalysis import CircuitAnalysis
from lhcsmapi.analysis.decorators import check_dataframe_empty
from lhcsmapi.pyedsl import SignalTransformationBuilder
from lhcsmapi.pyedsl.AssertionBuilder import AssertionBuilder
from lhcsmapi.pyedsl.FeatureBuilder import FeatureBuilder
from lhcsmapi.pyedsl.PlotBuilder import create_title, convert_plot_features_map_to_text, PlotBuilder
from lhcsmapi.reference.Reference import Reference


class PcAnalysis(CircuitAnalysis):
    """ Base class with methods for PC analysis in all circuits

    """
    @staticmethod
    @check_dataframe_empty(mode='any', warning='PC signal is empty, plot is skipped!')
    def plot_i_meas(i_meas_df: pd.DataFrame, xlim=(np.nan, np.nan), ylim=(np.nan, np.nan), title='') -> None:
        """ Method plotting a power converter current for a given x and y limits as well as a title.

        :param i_meas_df: main power converter current
        :type i_meas_df: pd.DataFrame
        :param xlim: limits of the x-axis
        :type xlim: Tuple[float]
        :param ylim: limits of the y-axis
        :type ylim: Tuple[float]
        :param title: plot title
        :return: None
        """
        PlotBuilder().with_signal(i_meas_df, title=title, grid=True) \
            .with_ylabel(ylabel='I, [A]') \
            .with_xlim(lim=xlim) \
            .with_ylim(lim=ylim) \
            .plot()

    @staticmethod
    @check_dataframe_empty(mode='any', warning='PC signals are empty, plot is skipped!')
    def plot_i_meas_with_current_plateau(i_meas_df, t0, plateau_start, plateau_end, title=''):
        """ Method plotting power converter current along with a boxes for constant current part. It is assumed that
        the start and current plateau lists are of the same length.

        :param i_meas_df: power converter current
        :param t0: initial timestamp of a current profile to synchronize current plateau duration
        :param plateau_start: a list of starts of current plateaus
        :param plateau_end: a list of ends of current plateaus
        :param title: plot title
        :return: None
        """
        ax = PlotBuilder().with_signal(i_meas_df, title=title, grid=True) \
            .with_ylabel(ylabel='I_MEAS, [A]') \
            .plot(show_plot=False).get_axes()[0]

        for ps, pe in zip(plateau_start, plateau_end):
            ax.axvspan(xmin=(ps - Time.to_unix_timestamp(t0)) / 1e9,
                       xmax=(pe - Time.to_unix_timestamp(t0)) / 1e9, facecolor='xkcd:goldenrod')
        plt.show()

    @staticmethod
    @check_dataframe_empty(mode='any', warning='PC signal is empty, assertion is skipped!')
    def assert_v_meas(timestamp_ee: int, timestamp_pic: int, t_after_ee: int, v_meas_df: pd.DataFrame,
                      value_min: float, value_max: float, title='', xmax=None) -> None:
        """ Method asserting whether a power converter voltage is between minimum and maximum value for a given amount
        of time from a PIC timestamp. Plot of voltage along with an acceptance bar is displayed

        :param timestamp_ee: EE PM timestamp (second EE)
        :param timestamp_pic: PIC NXCALS timestamp
        :param t_after_ee: time after an EE trigger
        :param v_meas_df: power converter voltage
        :param value_min: minimum allowed voltage value
        :param value_max: maximum allowed voltage value
        :param title: plot title
        :param xmax: maximum x axis
        :return:
        """
        t_assertion = (timestamp_ee - timestamp_pic) * 1e-9 + t_after_ee
        v_meas_at_assertion = v_meas_df.reindex([t_assertion], method='nearest', tolerance=.1).values[0]
        xmax = v_meas_df.index[-1] if xmax is None else xmax
        ax = PlotBuilder().with_signal(v_meas_df, title=title, grid=True) \
            .with_ylabel(ylabel='V_MEAS, [V]') \
            .with_xlim(lim=(0, xmax)) \
            .with_ylim(lim=(value_min - 0.5, value_max + 0.5)) \
            .plot(show_plot=False).get_axes()[0]

        y_err = (value_max + value_min) / 2
        error_range = value_max - y_err
        ax.errorbar(t_assertion, y_err, yerr=[[error_range], [error_range]], fmt='o', color='red')
        plt.show()
        print('Analysis result')
        if value_min < v_meas_at_assertion < value_max:
            print('%s signal value at t = %f s  (%.1f V) is within of the acceptance range [%.1f, %.1f] V.' %
                  (v_meas_df.columns[0], t_assertion, v_meas_at_assertion, value_min, value_max))
        else:
            warnings.warn('%s signal value at t = %f s  (%.1f V) is outside of the acceptance range [%.1f, %.1f] V.' %
                  (v_meas_df.columns[0], t_assertion, v_meas_at_assertion, value_min, value_max))

    @check_dataframe_empty(mode='any', warning='PC signals are empty, plot is skipped!')
    def plot_i_meas_pc(self, circuit_name: str, timestamp_fgc: int, i_dfs, xlim=(np.nan, np.nan),
                       ylim=(np.nan, np.nan)) -> None:
        """ Method plotting power converter currents.
        If any input signal is missing, then the analysis is skipped.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_fgc: timestamp of the FGC PM event (for the title)
        :type timestamp_fgc: int
        :param i_dfs: list of power converter currents
        :type i_dfs: List[pd.DataFrame]
        :param i_ref_dfs: list of reference power converter currents
        :type i_ref_dfs: List[pd.DataFrame]
        :param xlim: limits of the x-axis
        :type xlim: Tuple[float]
        :param ylim: limits of the y-axis
        :type ylim: Tuple[float]
        """
        title = create_title(circuit_name, timestamp_fgc, 'I_MEAS, I_A, I_REF')

        i_meas_dfs = [i_df for i_df in i_dfs if ('I_MEAS' in i_df.columns[0])]
        i_a_dfs = [i_df for i_df in i_dfs if ('I_A' in i_df.columns[0])]
        i_ref_dfs = [i_df for i_df in i_dfs if ('I_REF' in i_df.columns[0])]

        len_i_meas, len_i_ref, len_i_a = len(i_meas_dfs), len(i_ref_dfs), len(i_a_dfs)

        color_i_meas = ['C%d' % i for i in range(len_i_meas)]
        color_i_ref = ['C%d' % i for i in range(len_i_ref)]
        color_i_a = ['C%d' % i for i in range(max(len_i_meas, len_i_ref), max(len_i_meas, len_i_ref) + len_i_a)]

        color = color_i_meas + color_i_ref + color_i_a
        marker = [None] * len_i_meas + ['x'] * len_i_ref + [None] * len_i_a

        # Global view
        PlotBuilder().with_signal(i_meas_dfs + i_ref_dfs + i_a_dfs, title=title, color=color, marker=marker,
                                  grid=True) \
            .with_ylabel(ylabel='I, [A]') \
            .with_xlim(lim=xlim) \
            .with_ylim(lim=ylim) \
            .plot()

    @staticmethod
    @check_dataframe_empty(mode='any', warning='I_MEAS signals are empty, tau calculation is skipped!')
    def calculate_i_meas_tau(i_meas_dfs: List[pd.DataFrame], duration_decay=(2, 100)):
        """ Method calculating time constant based on power converter I_MEAS current

        :param i_meas_dfs: list of power converter I_MEAS currents
        :type i_meas_dfs: List[pd.DataFrame]
        :param duration_decay: duration of decay calculation
        :type duration_decay: Tuple[float]
        :return: list of time constant calculated for a set of input currents
        """
        return FeatureBuilder().with_signal(i_meas_dfs) \
            .subtract_min_value() \
            .with_duration(t_start=duration_decay[0], t_end=duration_decay[1]) \
            .calculate_features('tau_energy') \
            .values[0]

    @check_dataframe_empty(mode='any', warning='PC signals are empty, plot is skipped!')
    def plot_i_meas_tau_pc(self, circuit_name: str, timestamp_fgc: int, i_dfs, xlim=(np.nan, np.nan),
                           ylim=(np.nan, np.nan)) -> None:
        """ Method plotting power converter currents.
        If any input signal is missing, then the analysis is skipped.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_fgc: timestamp of the FGC PM event (for the title)
        :type timestamp_fgc: int
        :param i_dfs: list of power converter currents
        :type i_dfs: List[pd.DataFrame]
        :param i_ref_dfs: list of reference power converter currents
        :type i_ref_dfs: List[pd.DataFrame]
        :param xlim: limits of the x-axis
        :type xlim: Tuple[float]
        :param ylim: limits of the y-axis
        :type ylim: Tuple[float]
        """
        i_meas_dfs = [i_df for i_df in i_dfs if ('I_MEAS' in i_df.columns[0])]
        i_a_dfs = [i_df for i_df in i_dfs if ('I_A' in i_df.columns[0])]
        i_ref_dfs = [i_df for i_df in i_dfs if ('I_REF' in i_df.columns[0])]

        title = create_title(circuit_name, timestamp_fgc, 'I_MEAS, I_A, I_REF')

        len_i_meas, len_i_ref, len_i_a = len(i_meas_dfs), len(i_ref_dfs), len(i_a_dfs)

        color_i_meas = ['C%d' % i for i in range(len_i_meas)]
        color_i_ref = ['C%d' % i for i in range(len_i_ref)]
        color_i_a = ['C%d' % i for i in range(max(len_i_meas, len_i_ref), max(len_i_meas, len_i_ref) + len_i_a)]

        color = color_i_meas + color_i_ref + color_i_a
        marker = [None] * len_i_meas + ['x'] * len_i_ref + [None] * len_i_a

        # Global view
        PlotBuilder().with_signal(i_meas_dfs + i_ref_dfs + i_a_dfs, title=title, color=color, marker=marker, grid=True)\
            .with_ylabel(ylabel='I, [A]') \
            .with_xlim(lim=xlim) \
            .with_ylim(lim=ylim) \
            .plot()

    @check_dataframe_empty(mode='any', warning='PC signals are empty, zoomed plot is skipped!')
    def plot_i_meas_pc_zoom(self, circuit_name: str, timestamp_fgc: int, t_quench: float, i_dfs,
                            xlim=(np.nan, np.nan), ylim=(np.nan, np.nan)) -> None:
        """ Method plotting power converter currents in zoom view around the start of a quench or
        If any input signal is missing, then the analysis is skipped.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_fgc: timestamp of the FGC PM event (for the title)
        :type timestamp_fgc: int
        :param t_quench: time of quench start (in seconds)
        :type t_quench: pd.DataFrame
        :param i_dfs: list power converter currents
        :type i_dfs: List[pd.DataFrame]
        :param i_ref_dfs: list power converter currents
        :type i_ref_dfs: List[pd.DataFrame]
        :param xlim: limits of the x-axis
        :type xlim: Tuple[float]
        :param ylim: limits of the y-axis
        :type ylim: Tuple[float]
        """
        title = create_title(circuit_name, timestamp_fgc, 'I_MEAS, I_A, I_REF')

        # Zoomed view
        # # in x
        if np.isnan(xlim[0]) and np.isnan(xlim[1]):
            xlim = [t_quench - 0.01, 0.01]

        # # in y
        if np.isnan(ylim[0]) and np.isnan(ylim[1]):
            mid = 0
            for i_df in i_dfs:
                idx = i_df.index.get_loc(t_quench, 'nearest')
                mid = max(mid, i_df[i_df.index == i_df.index[idx]].values[0][0])
            ylim = [mid - 10, mid + 10]

        i_meas_dfs = [i_df for i_df in i_dfs if ('I_MEAS' in i_df.columns[0])]
        i_a_dfs = [i_df for i_df in i_dfs if ('I_A' in i_df.columns[0])]
        i_ref_dfs = [i_df for i_df in i_dfs if ('I_REF' in i_df.columns[0])]

        len_i_meas, len_i_ref, len_i_a = len(i_meas_dfs), len(i_ref_dfs), len(i_a_dfs)

        color_i_meas = ['C%d' % i for i in range(len_i_meas)]
        color_i_ref = ['C%d' % i for i in range(len_i_ref)]
        color_i_a = ['C%d' % i for i in range(max(len_i_meas, len_i_ref), max(len_i_meas, len_i_ref) + len_i_a)]

        color = color_i_meas + color_i_ref + color_i_a
        marker = [None] * len_i_meas + ['x'] * len_i_ref + [None] * len_i_a

        # Global view
        PlotBuilder().with_signal(i_meas_dfs + i_ref_dfs + i_a_dfs, title=title, color=tuple(color),
                                  marker=tuple(marker),
                                  grid=True) \
            .with_ylabel(ylabel='I, [A]') \
            .with_axvline(x=t_quench, ls='--') \
            .with_xlim(lim=xlim) \
            .with_ylim(lim=ylim) \
            .plot()

    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, PC trigger analysis skipped!')
    def analyze_i_meas_pc_trigger(self, timestamp_fgc: int, events_action_df: pd.DataFrame,
                                  events_symbol_df: pd.DataFrame) -> None:
        """ Method finding and displaying the source of a PC trigger.
        If any input signal is missing, then the analysis is aborted.

        :param timestamp_fgc: FGC event timestamp
        :type timestamp_fgc: int
        :param events_action_df: PC events action
        :type events_action_df: pd.DataFrame
        :param events_symbol_df: PC events symbol
        :type events_symbol_df: pd.DataFrame
        """
        event_pc_df = pd.concat([events_action_df, events_symbol_df], axis=1)
        fault_event_pc_df = event_pc_df[event_pc_df['EVENTS.ACTION'] == 'FAULT']
        if not fault_event_pc_df.empty:
            array = np.asarray(fault_event_pc_df.index)
            idx = (np.abs(array - timestamp_fgc)).argmin()

            print('The reason for PC PM event is:')
            display(HTML(fault_event_pc_df[fault_event_pc_df.index == fault_event_pc_df.index[idx]].to_html()))

    @check_dataframe_empty(mode='all', warning='I_MEAS signal is empty, dI_MEAS/dt analys skipped!')
    def calculate_current_slope(self, i_meas_df: pd.DataFrame, col_name=('Ramp rate', 'Plateau duration')) -> None:
        """ Method analyzing the power converter current derivative.
        If the input signal is missing, then the analysis is skipped.
        In case the current derivative prior to an event was zero (less than 0.1 A/s), then
        - the value of the current plateau and its duration are reported in the results table
        otherwise only the value of the derivative (ramp-rate) is reported in the results table

        :param i_meas_df: measured power converter current
        :type i_meas_df: pd.DataFrame
        :param col_name: name of a column to set in results table
        :type col_name: Tuple[str, str]
        """
        # Ramp Rate and Plateau Duration
        pd.options.mode.chained_assignment = None
        idx = max([i_meas_df.idxmin().values[0], i_meas_df.idxmax().values[0]])
        i_meas_df_der = i_meas_df[i_meas_df.index < idx]

        last_di_dt, last_di_dt_duration = SignalTransformationBuilder.calculate_last_const_di_dt(i_meas_df_der)
        print('Last di/dt prior to an FPA is %f A/s.' % last_di_dt)
        print('Duration of the last di/dt prior to an FPA is %f s (as seen in the PM buffer).' % last_di_dt_duration)
        # Store analysis result
        if self.results_table is not None:
            if abs(last_di_dt) > 0.1:
                self.results_table[col_name[0]] = last_di_dt
                self.results_table[col_name[1]] = 0
            else:
                self.results_table[col_name[0]] = 0
                self.results_table[col_name[1]] = last_di_dt_duration

        pd.options.mode.chained_assignment = 'warn'

    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, PC trigger analysis skipped!')
    def calculate_current_miits(self, i_meas_df: pd.DataFrame, t_quench: float, col_name='I_MEAS MIITs') -> None:
        """ Method calculating quench load (MIITs) of the power converter current.
        If any input signal is missing, then the analysis is aborted.
        The calculated MIITs is set in the results table.

        :param i_meas_df: measured and filtered power converter current
        :type i_meas_df: pd.DataFrame
        :param t_quench: time of quench start (in seconds)
        :type t_quench: pd.DataFrame
        :param col_name: name of a column to set the output results table
        :type col_name: str
        """
        i_meas_pos_df = i_meas_df[i_meas_df.index >= t_quench]
        miits_i_meas = FeatureBuilder().with_signal(i_meas_pos_df).calculate_features('miits').loc[0, :].values[0]
        print('%s is %s MA^2s from t_1 = %f s (quench detection) to t_2 = %f s (end of PM buffer).' %
              (col_name, miits_i_meas, i_meas_pos_df.index[0], i_meas_pos_df.index[-1]))

        # Store analysis result
        if self.results_table is not None:
            self.results_table[col_name] = miits_i_meas

    @check_dataframe_empty(mode='any', warning='I_EARTH signal is empty, analysis skipped!')
    def calculate_max_i_earth_pc(self, i_earth_df: pd.DataFrame, col_name='Earth Current') -> None:
        """ Method calculating the absolute maximum value of earth current in mA.
        If the input signal is missing, then the calculation is skipped.
        The calculated absolute maximum value is set in the results table.

        :param i_earth_df: current of the earth circuit in the power converter
        :type i_earth_df: pd.DataFrame
        :param col_name: name of a column to set in results table
        :type col_name: str
        """

        i_earth_median = 1000 * i_earth_df.dropna().median().values[0]
        i_earth_max_unbiased = FeatureBuilder().with_signal(1000 * i_earth_df - i_earth_median) \
                                   .calculate_features('max_abs_min').loc[0, :].values[0]

        print('The maximum unbiased abs(I_EARTH) is %s mA.' % (round(i_earth_max_unbiased, 2)))

        # Store analysis result
        if self.results_table is not None:
            self.results_table[col_name] = round(i_earth_max_unbiased, 6)

    @check_dataframe_empty(mode='any', warning='I_EARTH signal is empty, analysis skipped!')
    def plot_i_earth_pc(self, circuit_name: str, timestamp_fgc: int, i_earth_df: pd.DataFrame) -> None:
        """ Method plotting the earth current in mA.
        If the input signal is missing, then the plot is skipped.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_fgc: timestamp of the FGC PM event (for the title)
        :type timestamp_fgc: int
        :param i_earth_df: current of the earth circuit in the power converter
        :type i_earth_df: pd.DataFrame
        """
        title = create_title(circuit_name, timestamp_fgc, 'I_EARTH')
        PlotBuilder().with_signal(1000 * i_earth_df, title=title, grid=True) \
            .with_ylabel(ylabel='IEARTH, [mA]') \
            .plot()

    @check_dataframe_empty(mode='any', warning='I_EARTH_PCNT signal is empty, analysis skipped!')
    def plot_i_earth_pcnt_pc(self, circuit_name: str, timestamp_fgc: int, i_earth_pcnt_df: pd.DataFrame) -> None:
        """ Method plotting the percentage earth current.
        If the input signal is missing, then the plot is skipped.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_fgc: timestamp of the FGC PM event (for the title)
        :type timestamp_fgc: int
        :param i_earth_pcnt_df: current of the earth circuit in the power converter
        :type i_earth_pcnt_df: pd.DataFrame
        """
        title = create_title(circuit_name, timestamp_fgc, 'I_EARTH_PCNT')
        PlotBuilder().with_signal(i_earth_pcnt_df, title=title, grid=True) \
            .with_ylabel(ylabel='I_EARTH_PCNT, [%]') \
            .plot()

    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, I_MEAS quench analysis skipped!')
    def calculate_quench_current(self, i_meas_df: pd.DataFrame, t_quench: float,
                                 col_name='Quench Current') -> None:
        """ Method calculating the quench current, i.e., the current at the moment when I_A and I_REF start to
        deviate from one another.
        If any input signal is missing, then the analysis is skipped.
        The calculated quench current is set in the results table.

        :param i_meas_df: measured power converter current
        :type i_meas_df: pd.DataFrame
        :param t_quench: time of quench start (in seconds)
        :type t_quench: pd.DataFrame
        :param col_name: name of a column to set in results table
        :type col_name: str
        """
        i_meas_pos_df = i_meas_df[i_meas_df.index >= t_quench]
        print('%s is %s A.' % (col_name, round(i_meas_pos_df.iloc[0].values[0], 1)))

        # Store analysis result
        if self.results_table is not None:
            self.results_table[col_name] = round(i_meas_pos_df.iloc[0].values[0], 1)


class Pc13kAAnalysis(PcAnalysis):
    """ Class with methods for 13 kA PC analysis in RB and RQ

    """
    @check_dataframe_empty(mode='any', warning='PC signals are empty, I_MEAS plot is skipped!')
    def analyze_i_meas_pc(self, circuit_name: str, timestamp_fgc: int, timestamp_fgc_ref: int, timestamp_pic: int,
                          i_meas_df: pd.DataFrame, i_meas_ref_df: pd.DataFrame, duration_decay=(2, 100)) -> None:
        """ Method analyzing and plotting main power converter current and compares to its reference.
        It computes characteristic time constant for a given time window. The actual characteristic decay time has to
        fall within +/- 20 % of the reference characteristic decay time.

        :param circuit_name: circuit name (either RB or RQ - 13 kA power converter)
        :param timestamp_fgc: FGC PM timestamp
        :param timestamp_fgc_ref: reference FGC PM timestamp
        :param timestamp_pic: PIC NXCALS timestamp
        :param i_meas_df: main power converter current
        :param i_meas_ref_df: reference power converter current
        :param duration_decay: time window over which the characteristic decay time is calculated
        :return: None
        """
        # Calculate Features
        # # characteristic time constant
        tau_i_meas, tau_i_meas_ref = FeatureBuilder().with_signal(i_meas_df, i_meas_ref_df) \
            .with_duration(t_start=duration_decay[0], t_end=duration_decay[1]) \
            .calculate_features('tau_energy') \
            .values[0]

        # Plot
        title = '{} {}: I_MEAS(t) and I_MEAS(t)/dI_MEAS(t) (Reference)\n'. \
            format(Time.to_string_short(timestamp_fgc), circuit_name)

        tau_i_meas_df = SignalTransformationBuilder.calculate_char_time_profile(i_meas_df)
        tau_i_meas_ref_df = SignalTransformationBuilder.calculate_char_time_profile(i_meas_ref_df)

        # Plot
        ylim = (50, 110) if self.circuit_type == 'RB' else (0, 35)
        PlotBuilder().with_signal([i_meas_df, i_meas_ref_df], title=title, grid=True, figsize=(13, 6.5),
                                  style=['-', '--']) \
            .with_ylabel(ylabel='I_MEAS, [A]') \
            .with_legend(labels=['I_MEAS', 'I_MEAS (Reference)']) \
            .with_signal([tau_i_meas_df, tau_i_meas_ref_df], style=['-', '--']) \
            .with_ylabel(ylabel='-I_MEAS/dI_MEAS/dt [s]') \
            .with_legend(labels=['-I_MEAS/dI_MEAS_dt', '-I_MEAS/dI_MEAS_dt (Reference)']) \
            .with_ylim(ylim).with_xlim((-50, np.nan)) \
            .plot()

        # Compare features to the reference
        pc_act_df = pd.DataFrame({'tau_i_meas': tau_i_meas}, index=['act'])

        tau_i_meas_comp_df = Reference.compare_features_to_reference(pc_act_df, self.circuit_type, 'PC')
        tau_i_meas_comp_ref_df = pd.DataFrame(index=['tau_i_meas_ref'],
                                              data={'min': 0.8 * tau_i_meas_ref,
                                                    'max': 1.2 * tau_i_meas_ref,
                                                    'act': tau_i_meas,
                                                    'result': 0.8 * tau_i_meas_ref < tau_i_meas < 1.2 * tau_i_meas_ref})
        tau_i_meas_comp_df = pd.concat((tau_i_meas_comp_df, tau_i_meas_comp_ref_df), axis=0)

        # Display analysis results
        print('Analysis results')
        print('----------------')
        plot_features = {'Circuit name': [circuit_name, ''],
                         'PIC Abort': [Time.to_string_short(timestamp_pic), ''],
                         'PC Off': [Time.to_string_short(timestamp_fgc), ''],
                         'Reference': [Time.to_string_short(timestamp_fgc_ref), ''],
                         'max(I_MEAS)': [round(i_meas_df.max().values[0]), 'A'],
                         'max(I_MEAS, Reference)': [round(i_meas_ref_df.max().values[0]), 'A']}
        print(convert_plot_features_map_to_text(plot_features))
        Reference.check_feature_comparison(tau_i_meas_comp_df)

    @staticmethod
    @check_dataframe_empty(mode='any', warning='PC signals are empty, I_MEAS plot is skipped!')
    def plot_i_meas_smoothness(i_meas_df: pd.DataFrame, title='', value_min=-10, value_max=10, window=10) -> None:
        """ Method calculating and displaying power converter current smoothness (second time derivative).
        The derivative is calculated as a rolling finite difference with a given window length.

        :param i_meas_df: main power converter current
        :param title: plot title
        :param value_min: minimum value for the current smoothness
        :param value_max: maximum value for the current smoothness
        :param window: size of a window for rolling derivative calculation
        :return: None
        """
        dt = i_meas_df.index[window - 1] - i_meas_df.index[0]
        i_meas_discharge_df = i_meas_df[i_meas_df.index >= 1]
        i_meas_discharge_dt_df = i_meas_discharge_df.rolling(window)\
            .apply(lambda x: (x[window - 1] - x[0]) / dt, raw=True)
        i_meas_discharge_dt_dt_df = i_meas_discharge_dt_df.rolling(window)\
            .apply(lambda x: (x[window - 1] - x[0]) / dt, raw=True)
        signal_name = i_meas_discharge_df.columns[0]
        i_meas_discharge_dt_dt_df.rename(columns={signal_name: 'd%s/dt^2' % signal_name}, inplace=True)

        AssertionBuilder().with_signal([i_meas_discharge_dt_dt_df]) \
            .has_min_max_value(value_min=value_min, value_max=value_max) \
            .show_plot(title=title, ylabel='dI_MEAS/dt^2, [A/s^2]', data_sec=i_meas_discharge_df,
                       ylabel_sec='I_MEAS, [A]')

    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, I_EARTH analysis skipped!')
    def analyze_i_earth_pc(self, circuit_name: str, timestamp_fgc: int, i_a_df: pd.DataFrame, i_earth_df: pd.DataFrame,
                           i_earth_ref_df: pd.DataFrame, xlim=(-0.1, 0.3)) -> None:
        """ Method plotting main power converter current and earth current. Earth current plotted with its reference.

        :param circuit_name: RQ circuit name
        :param timestamp_fgc: RQ FGC timestamp (for title)
        :param i_a_df: main power converter current (not-filtered, high sampling-rate)
        :param i_earth_df: earth current
        :param i_earth_ref_df: reference earth current
        :param xlim: limits for x-axis
        :return: None
        """

        title = '{} {}: I_A and I_EARTH (Reference)'.format(Time.to_string_short(timestamp_fgc), circuit_name)

        PlotBuilder().with_signal(i_a_df, title=title, grid=True) \
            .with_ylabel(ylabel='I_A, [A]') \
            .with_signal([1000 * i_earth_df, 1000 * i_earth_ref_df], style=['-', '--']) \
            .with_ylabel(ylabel='I_EARTH, [mA]') \
            .with_legend(labels=['I_EARTH', 'I_EARTH (Reference)']) \
            .with_xlim(xlim) \
            .plot()

    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, I_EARTH_PCNT analysis skipped!')
    def analyze_i_earth_pcnt_pc(self, circuit_type: str, circuit_name: str, timestamp_fgc: int, i_meas_df: pd.DataFrame,
                                i_meas_ref_df: pd.DataFrame, i_earth_pcnt_df: pd.DataFrame,
                                i_earth_pcnt_ref_df: pd.DataFrame, xlim=(5, 120)) -> None:
        """ Method plotting main power converter current and percentage earth current. Earth current is shown with its
            reference. Reference percentage earth current is scaled according to the main currents. A comment related to
            the percentage earth current is updated. In addition, similarity between the percentage earth currents is
            calculated.

            :param circuit_name: RQ circuit name
            :param timestamp_fgc: RQ FGC timestamp (for title)
            :param i_meas_df: main power converter current (for scaling)
            :param i_meas_ref_df: reference  main power converter current (for scaling)
            :param i_earth_pcnt_df: percentage earth current
            :param i_earth_pcnt_ref_df: reference percentage earth current
            :return: None
            """

        title = '{} {}: I_MEAS and I_EARTH_PCNT (Reference)'.format(Time.to_string_short(timestamp_fgc), circuit_name)
        scaling = (i_meas_df.max() / i_meas_ref_df.max()).values[0]
        PlotBuilder().with_signal(i_meas_df, title=title, grid=True) \
            .with_ylabel(ylabel='I_MEAS, [A]') \
            .with_signal([i_earth_pcnt_df[i_earth_pcnt_df.index > 3],
                          i_earth_pcnt_ref_df[i_earth_pcnt_ref_df.index > 3] * scaling], yerr=[None, 3]) \
            .with_ylabel(ylabel='I_EARTH_PCNT, [%]') \
            .with_legend(labels=['I_EARTH_PCNT', 'I_EARTH_PCNT (Reference)']) \
            .with_xlim(xlim) \
            .plot()

        AssertionBuilder().with_signal(i_earth_pcnt_df[i_earth_pcnt_ref_df.index > 3]) \
            .is_between_two_references(first_ref=i_earth_pcnt_ref_df[i_earth_pcnt_ref_df.index > 3],
                                       first_scaling=scaling, first_offset=3,
                                       second_ref=i_earth_pcnt_ref_df[i_earth_pcnt_ref_df.index > 3],
                                       second_scaling=scaling, second_offset=3, is_plot=False) \
            .get_assertion_result()


class Pc13kARqAnalysis(Pc13kAAnalysis):
    """ Class with methods for RQ 13 kA PC analysis

    """
    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, I_EARTH_PCNT analysis skipped!')
    def analyze_i_earth_pcnt_pc(self, circuit_name: str, timestamp_fgc: int, i_meas_df: pd.DataFrame,
                                i_meas_ref_df: pd.DataFrame, i_earth_pcnt_df: pd.DataFrame,
                                i_earth_pcnt_ref_df: pd.DataFrame) -> None:
        """ Method plotting main power converter current and percentage earth current. Earth current is shown with its
        reference. Reference percentage earth current is scaled according to the main currents. A comment related to the
        percentage earth current is updated.

        :param circuit_name: RQ circuit name
        :param timestamp_fgc: RQ timestamp (for title)
        :param i_meas_df: main power converter current (for scaling)
        :param i_meas_ref_df: reference  main power converter current (for scaling)
        :param i_earth_pcnt_df: percentage earth current
        :param i_earth_pcnt_ref_df: reference percentage earth current
        :return: None
        """
        title = '{} {}: I_MEAS and I_EARTH_PCNT (Reference)'.format(Time.to_string_short(timestamp_fgc), circuit_name)
        scaling = (i_meas_df.max() / i_meas_ref_df.max()).values[0]
        PlotBuilder().with_signal(i_meas_df, title=title, grid=True) \
            .with_ylabel(ylabel='I_MEAS, [A]') \
            .with_signal([i_earth_pcnt_df[i_earth_pcnt_df.index > 3],
                          i_earth_pcnt_ref_df[i_earth_pcnt_ref_df.index > 3] * scaling]) \
            .with_ylabel(ylabel='I_EARTH_PCNT, [%]') \
            .with_legend(labels=['I_EARTH_PCNT', 'I_EARTH_PCNT (Reference)']) \
            .with_xlim((5, 120)) \
            .plot()
