from typing import List, Union
import requests

import pandas as pd
from tqdm.notebook import tqdm

from lhcsmapi.analysis.CircuitQuery import CircuitQuery, execution_count
from lhcsmapi.analysis.decorators import check_nan_timestamp_signals, check_nan_timestamp
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder


class QdsQuery(CircuitQuery):
    """Base class for quench detection system (QDS) query

    """
    @execution_count()
    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_iqps_nxcals(self, timestamp_fgc: int, *, signal_names: List[str], spark, duration=[(300, 's'), (10, 's')],
                          ) -> List[pd.DataFrame]:
        """ Method querying NXCALS for iQPS signals

        :param timestamp_fgc: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_fgc: int
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param spark: spark session variable needed to perform an NXCALS query
        :type spark: spark session
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        return QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=timestamp_fgc, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='QDS', signal=signal_names) \
            .signal_query(verbose=self.verbose) \
            .synchronize_time(timestamp_fgc) \
            .convert_index_to_sec() \
            .filter_median().dfs

    @execution_count()
    @check_nan_timestamp(return_type=pd.DataFrame())
    def find_source_timestamp_qds(self, t_start: Union[int, str, float], duration=[(10, 's'), (400, 's')]
                                  ) -> pd.DataFrame:
        """ Method searching PM QDS events
        If the start is NaN, then an empty dataframe is returned.

        :param t_start: start time of a search
        :type t_start: Union[str, int, float]
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: pd.DataFrame with source and timestamp of the PM QDS events
        """
        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='QDS', source='*') \
            .event_query(verbose=self.verbose) \
            .filter_source(self.circuit_name, 'QDS') \
            .drop_duplicate_source() \
            .sort_values(by='timestamp').df

    @execution_count()
    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_qds_pm(self, timestamp_qds: int, timestamp_qds_sync: int, *, signal_names: List[str]
                     ) -> List[pd.DataFrame]:
        """ Method querying PM for QDS signals
        If the timestamp is NaN, then a list of empty dataframes of the size of the signal names is returned.
        The signals are synchronized in time and converted from ns to s.

        :param timestamp_qds: PM QDS timestamp in ns precision
        :type timestamp_qds: int
        :param timestamp_qds: PM QDS timestamp in ns precision (for signal synchronization)
        :type timestamp_qds: int
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_qds) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='QDS', signal=signal_names) \
            .signal_query(verbose=self.verbose) \
            .synchronize_time(timestamp_qds_sync) \
            .convert_index_to_sec() \
            .filter_median().dfs

    @execution_count()
    def query_iqps_analog_pm(self, source_timestamp_qds_df: pd.DataFrame,
                             signal_names: List[str]) -> List[List[pd.DataFrame]]:
        """ Method querying PM for iQPS analog signals
        Signals are synchronized to the iQPS PM event and index is converted to seconds.

        :param source_timestamp_qds_df: dataframe with PM source and timestamp for QDS events
        :type source_timestamp_qds_df: pd.DataFrame
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :return: List of list of iQPS signals. The outer list has as many elements as iQPS events and the lenght
            of the inner is equal to the number of iQPS signals.
        """
        iqps_analog_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying iQPS PM'):
            timestamp_qds = row['timestamp']
            source_qds = row['source']

            iqps_analog_df = QueryBuilder().with_pm() \
                .with_timestamp(timestamp_qds) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='QDS', signal=signal_names, source=source_qds,
                               wildcard={'CELL': source_qds}) \
                .signal_query() \
                .synchronize_time(timestamp_qds) \
                .convert_index_to_sec() \
                .filter_median().dfs

            iqps_analog_dfs.append(iqps_analog_df)

        return iqps_analog_dfs

    @execution_count()
    def query_iqps_digital_pm(self, source_timestamp_qds_df: pd.DataFrame,
                              signal_names: List[str]) -> List[List[pd.DataFrame]]:
        """ Method querying PM for iQPS digital signals
        Signals are synchronized to the iQPS PM event and index is converted to seconds.

        :param source_timestamp_qds_df: dataframe with PM source and timestamp for QDS events
        :type source_timestamp_qds_df: pd.DataFrame
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :return: List of list of iQPS signals. The outer list has as many elements as iQPS events and the lenght
            of the inner is equal to the number of iQPS signals.
        """
        iqps_digital_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying IQPS PM'):
            timestamp_qds = row['timestamp']
            source_qds = row['source']

            iqps_digital_df = QueryBuilder().with_pm() \
                .with_timestamp(timestamp_qds) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='QDS', signal=signal_names, source=source_qds) \
                .signal_query() \
                .synchronize_time(timestamp_qds) \
                .convert_index_to_sec().dfs
            iqps_digital_dfs.append(iqps_digital_df)

        return iqps_digital_dfs


class QdsIpdQuery(QdsQuery):
    """ Class for IPD quench detection system (QDS) query

    """
    @execution_count()
    @check_nan_timestamp(return_type=pd.DataFrame())
    def find_source_timestamp_qds(self, t_start: Union[int, str, float], duration=[(10, 's'), (400, 's')],
                                  ) -> pd.DataFrame:
        """ Method searching PM QDS events
        If the start is NaN, then an empty dataframe is returned.

        :param t_start: start time of a search
        :type t_start: Union[str, int, float]
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: pd.DataFrame with source and timestamp of the PM QDS events
        """
        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='QDS', source='*') \
            .event_query(verbose=self.verbose) \
            .sort_values(by='timestamp').df


class QdsRbQuery(QdsQuery):
    """ Class for RB quench detection system (QDS) query

    """

    # iQPS
    @execution_count()
    def query_pm_iqps_board_type(self, source_timestamp_qds_df: pd.DataFrame) -> pd.DataFrame:
        """ Method querying PM for qps board type

        :param source_timestamp_qds_df: table with QDS PM source and timestamp
        :type source_timestamp_qds_df: pd.DataFrame

        :return: table with iQPS board types (1, 0)
        """
        iqps_board_type_df = pd.DataFrame(columns=['iqps_board_type'])
        metadata_qds = SignalMetadata.get_circuit_signal_database_metadata(self.circuit_type, self.circuit_name, 'QDS',
                                                                           'PM')
        board_types = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying PM'):
            source_qds, timestamp_qds = row[['source', 'timestamp']]

            # ToDo: It would be nice to add this query to the library
            pm_request = "http://pm-api-pro/v2/pmdata?system=QPS&className=" + metadata_qds['className'] + "&source=" \
                         + source_qds + "&TimestampInNanos=" + str(timestamp_qds) + "&pageSize=1&pageIndex=0"
            resp = requests.get(pm_request)
            board_type = resp.json()["content"][0]["namesAndValues"][1]["value"]

            board_types.append(board_type)

        iqps_board_type_df['iqps_board_type'] = board_types
        return iqps_board_type_df

    def find_source_timestamp_qds_board_ab(self, t_start: Union[int, str, float], duration=[(10, 's'), (400, 's')],
                                           ) -> pd.DataFrame:
        """ Method searching PM QDS events
        If the start is NaN, then an empty dataframe is returned.

        :param t_start: start time of a search
        :type t_start: Union[str, int, float]
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: pd.DataFrame with source and timestamp of the PM QDS events
        """
        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='QDS', source='*') \
            .event_query(verbose=self.verbose) \
            .filter_source(self.circuit_name, 'QDS') \
            .sort_values(by=['source', 'timestamp']).df \
            .reset_index(drop=True)

    @execution_count()
    def query_voltage_logic_iqps(self, source_timestamp_qds_df, *, signal_names, filter_window=1):
        """ Method querying voltage and logic signals of iQPS.
        The signals are synchronized in time and converted from ns to s.

        :param source_timestamp_qds_df: table with source and timestamp for iQPS
        :type source_timestamp_qds_df: pd.DataFrame
        :param signal_names: list of signal names to query
        :type signal_names: List[str]
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        u_qds_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying iQPS logic PM'):
            source_qds, timestamp_qds = row[['source', 'timestamp']]

            u_qds_df = QueryBuilder().with_pm() \
                .with_timestamp(timestamp_qds) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='QDS', source=source_qds,
                               wildcard={'CELL': source_qds},
                               signal=signal_names) \
                .signal_query() \
                .synchronize_time(timestamp_qds) \
                .convert_index_to_sec()\
                .filter_median(filter_window).dfs

            u_qds_dfs.append(u_qds_df)

        return u_qds_dfs
