import warnings
from typing import List, Tuple

from IPython.display import display_html
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

from lhcsmapi.Time import Time
from lhcsmapi.analysis.CircuitAnalysis import CircuitAnalysis
from lhcsmapi.pyedsl.dbsignal.SignalAnalysis import SignalAnalysis
from lhcsmapi.pyedsl.FeatureBuilder import FeatureBuilder
from lhcsmapi.pyedsl.PlotBuilder import create_title, PlotBuilder
from lhcsmapi.reference.Reference import Reference
from lhcsmapi.analysis.expert_input import check_show_next
from lhcsmapi.pyedsl.dbsignal.SignalIndexConversion import SignalIndexConversion
from lhcsmapi.analysis.warnings import warning_on_one_line

warnings.formatwarning = warning_on_one_line


class QuenchHeaterAnalysis(CircuitAnalysis):
    """ Base class for quench heater analysis in circuits for which only the voltage is stored in PM buffers:
    - IT
    - IPQ
    - IPD
    Note that although MQ magnets (RQ circuit) also fall into this category, due to some differences they are analysed
    with a dedicated module.
    """

    def analyze_qh(self, circuit_name: str, timestamp_qds: int, u_hds_dfs: List[pd.DataFrame],
                   u_hds_ref_dfs: List[pd.DataFrame]) -> None:
        """ Method analyzing QH voltage signals in IT, IPD, IPQ circuits. It is plotting all signals and calculating:
            - initial voltage
            - final voltage
            - time constant based on the charge approach

        :param circuit_name: circuit name to analyze ('RQX.R1')
        :type circuit_name: str
        :param timestamp_qds: unix timestamp of the currently analyzed QH discharge
        :type timestamp_qds: int
        :param u_hds_dfs: list of lists of QH voltage signals. There are as many outer lists as quenched
            magnets and as many elements in the inner list as there are QH voltage signals to analyze
        :type u_hds_dfs: List[List[pd.DataFrame]]
        :param u_hds_ref_dfs: list of lists of reference QH voltage signals. There are as many outer lists as
            quenched magnets and as many elements in the inner list as there are reference QH voltage signals to analyze
        :type u_hds_ref_dfs: List[pd.DataFrame]
        """

        if not any([not u_hds_df.empty for u_hds_df in u_hds_dfs]):
            warnings.warn('All discharge curves are empty, please check the PM data (%d). Analysis skipped.' % timestamp_qds)
            return None

        # Plot
        title = create_title(circuit_name, timestamp_qds, 'U_HDS')
        PlotBuilder().with_signal(u_hds_dfs, title=title, figsize=(15, 7), grid=True) \
            .with_xlim((-0.5, 1.5)) \
            .with_signal(u_hds_ref_dfs, is_twinx=False, grid=True, style='--') \
            .with_ylabel(ylabel='U_HDS, [V]', color='C0') \
            .plot()

        PlotBuilder().with_signal(u_hds_dfs, title=title, figsize=(15, 7), grid=True) \
            .with_xlim((-0.5, 1.5)) \
            .with_signal(u_hds_ref_dfs, is_twinx=False, grid=True, style='--') \
            .with_ylabel(ylabel='U_HDS, [V]', color='C0') \
            .plot(yscale='log')

        # Take decay only
        u_hds_decay_dfs = []
        for u_hds_df in u_hds_dfs:
            if u_hds_df.max().values[0] > u_hds_df.min().values[0]:
                u_hds_decay_dfs.append(u_hds_df[(u_hds_df.index >= 0) & (u_hds_df.index < u_hds_df.idxmin().values[0])])
            else:
                u_hds_decay_dfs.append(u_hds_df[(u_hds_df.index >= 0)])

        u_hds_ref_decay_dfs = [
            u_hds_ref_df[(u_hds_ref_df.index >= 0) & (u_hds_ref_df.index < u_hds_ref_df.idxmin().values[0])] for
            u_hds_ref_df in u_hds_ref_dfs]

        # Calculate features
        # # compare to reference values
        first_last_u_df = FeatureBuilder().with_signal(u_hds_decay_dfs) \
            .calculate_features(features=['first', 'last20mean'], index=timestamp_qds)

        first_last_u_comp_df = Reference.compare_features_to_reference(first_last_u_df, self.circuit_type, 'QH',
                                                                       wildcard={'CIRCUIT': circuit_name})
        is_first_last_u_ok = first_last_u_comp_df['result'].all()

        # # compare to a reference discharge
        timestamp_ref = Reference.get_quench_heater_reference_discharge(self.circuit_type, circuit_name)

        tau_df = FeatureBuilder().with_signal(u_hds_decay_dfs) \
            .calculate_features(features=['tau_charge'], index=timestamp_qds)

        tau_ref_df = FeatureBuilder().with_signal(u_hds_ref_decay_dfs) \
            .calculate_features(features=['tau_charge'], index=timestamp_ref)

        tau_comp_df = Reference.compare_difference_of_features_to_reference(tau_df, tau_ref_df, self.circuit_type,
                                                                            'QH', wildcard={'CIRCUIT': circuit_name})
        is_tau_correct = tau_comp_df['result'].all()

        # Update results table
        # Use OK rather than True always
        final_label = 'Pass' if is_first_last_u_ok and is_tau_correct else 'Fail'
        if self.results_table is not None:
            self.results_table['QH analysis'] = final_label

        # Display features
        # Display comparison
        display_html(first_last_u_comp_df.style.set_table_attributes("style='display:inline'") \
                     .set_caption('Comparison of the inititial and final voltage') \
                     .apply(lambda s: first_last_u_comp_df['result'] \
                            .map({False: 'background-color: red', True: ''}))._repr_html_() \
                     + tau_comp_df.style.set_table_attributes("style='display:inline'") \
                     .set_caption('Comparison of the discharge characteristic time to the reference') \
                     .apply(lambda s: tau_comp_df['result'] \
                            .map({False: 'background-color: red', True: ''}))._repr_html_(), raw=True)

        timestamp_qds_ref = Reference.get_quench_heater_reference_discharge(self.circuit_type, circuit_name)
        print('Reference QH discharge timestamp is: %s' % Time.to_string_short(timestamp_qds_ref))
        print('The QH discharges are labeled: %s.' % final_label)


class QuenchHeaterRqAnalysis(CircuitAnalysis):
    """ Class with methods for RQ QH analysis

    """

    def analyze_qh(self, source_timestamp_qh_df: pd.DataFrame, circuit_type: str,
                   u_hds_rq_dfs: List[List[pd.DataFrame]], u_hds_rq_ref_dfs: List[List[pd.DataFrame]]) -> None:
        """ Method analyzing QH voltage signals in RQ circuit. It is plotting all signals and calculating:
            - initial voltage
            - final voltage
            - time constant based on the charge approach

        :param source_timestamp_qh_df: dataframe with PM source and timestamp for QH events
        :type source_timestamp_qh_df: pd.DataFrame
        :param circuit_type: circuit type to analyze ('RQ')
        :type circuit_type: str
        :param u_hds_rq_dfs: list of lists of QH voltage signals. There are as many outer lists as quenched
            magnets and as many elements in the inner list as there are QH voltage signals to analyze
        :type u_hds_rq_dfs: List[List[pd.DataFrame]]
        :param u_hds_rq_ref_dfs: list of lists of reference QH voltage signals. There are as many outer lists as
            quenched magnets and as many elements in the inner list as there are reference QH voltage signals to analyze
        :type u_hds_rq_ref_dfs: List[List[pd.DataFrame]]
        """

        for index, row in source_timestamp_qh_df.iterrows():
            source_qds, timestamp_qds = row[['source', 'timestamp']]
            timestamp_qds = Time.to_unix_timestamp(timestamp_qds)
            timestamp_ref = Reference.get_quench_heater_reference_discharge(circuit_type, source_qds)

            u_hds_rq_df = u_hds_rq_dfs[index]
            u_hds_rq_ref_df = u_hds_rq_ref_dfs[index]

            # Plot
            title = "Magnet: {}, Time Stamp: {}, Reference {}, U_HDS(t)" \
                .format(source_qds, Time.to_string_short(timestamp_qds + 1000000), Time.to_string_short(timestamp_ref))

            PlotBuilder().with_signal(u_hds_rq_df, title=title, figsize=(15, 7.5), grid=True) \
                .with_ylabel(ylabel='U_HDS, [V]') \
                .with_signal(u_hds_rq_ref_df, is_twinx=False, grid=True, style='--') \
                .with_ylabel(ylabel='U_HDS, [V]', color='C0') \
                .plot()

            PlotBuilder().with_signal(u_hds_rq_df, title=title, figsize=(15, 7.5), grid=True) \
                .with_ylabel(ylabel='U_HDS, [V]') \
                .with_signal(u_hds_rq_ref_df, is_twinx=False, grid=True, style='--') \
                .with_ylabel(ylabel='U_HDS, [V]', color='C0') \
                .plot(yscale='log')

            # Calculate features
            # # compare to reference values
            first_last_u_df = FeatureBuilder().with_signal(u_hds_rq_df) \
                .calculate_features(features=['first', 'last20mean'], index=timestamp_qds)

            first_last_u_comp_df = Reference.compare_features_to_reference(first_last_u_df, self.circuit_type,
                                                                           'QH', wildcard={'CELL': source_qds})
            is_first_last_u_ok = first_last_u_comp_df['result'].all()

            # # compare to the reference discharge
            tau_df = FeatureBuilder().with_signal(u_hds_rq_df) \
                .calculate_features(features=['tau_charge'], index=timestamp_qds)

            tau_ref_df = FeatureBuilder().with_signal(u_hds_rq_ref_df) \
                .calculate_features(features=['tau_charge'], index=timestamp_ref)

            tau_comp_df = Reference.compare_difference_of_features_to_reference(tau_df, tau_ref_df, self.circuit_type,
                                                                                'QH', wildcard={'CELL': source_qds})
            is_tau_ok = tau_comp_df['result'].all()

            # Display features
            # Display comparison
            display_html(first_last_u_comp_df.style.set_table_attributes("style='display:inline'") \
                         .set_caption('Comparison of the inititial and final voltage') \
                         .apply(lambda s: first_last_u_comp_df['result'] \
                                .map({False: 'background-color: red', True: ''}))._repr_html_() \
                         + tau_comp_df.style.set_table_attributes("style='display:inline'") \
                         .set_caption('Comparison of the discharge characteristic time to the reference') \
                         .apply(lambda s: tau_comp_df['result'] \
                                .map({False: 'background-color: red', True: ''}))._repr_html_(), raw=True)

            # Update results table
            # Use OK rather than True always
            if self.results_table is not None:
                self.results_table.loc[index, 'QH analysis'] = 'Pass' if is_first_last_u_ok and is_tau_ok else 'Fail'


def get_df_col(dfs: List[pd.DataFrame], col: str) -> Tuple[pd.DataFrame, str]:
    """ Method returning a dataframe matching an input column from a list of dataframes

    :param dfs: list of dataframes
    :param col: column name
    :return: a tuple with a dataframe and its column
    """
    for df in dfs:
        if col in df.columns[0]:
            return df, df.columns[0]


def calculate_resistance(u_i_hds_dfs: List[pd.DataFrame]) -> List[pd.DataFrame]:
    """ Function calculating a list of resistances of a signal from a list of QH voltages and currents

    :param u_i_hds_dfs: list of QH voltages and currents
    :return: list of resistances
    """
    no_signal = int(len(u_i_hds_dfs) / 2)
    r_hds_dfs = []
    for index in range(1, no_signal + 1):
        u_hds_df, col_u_hds_df = get_df_col(u_i_hds_dfs, 'U_HDS_{}'.format(index))
        i_hds_df, col_i_hds_df = get_df_col(u_i_hds_dfs, 'I_HDS_{}'.format(index))

        r_value = u_hds_df.values / i_hds_df.values
        column = col_u_hds_df.replace('U', 'R')
        time = u_hds_df.index
        r_hds_dfs.append(pd.DataFrame(data=r_value, index=time, columns=[column]))
    return r_hds_dfs


def has_window_mean_std(x: np.ndarray, mean=50, std=0.1) -> bool:
    """ Function checking if a window vector has certain mean and std values

    :param x: input vector
    :param mean: threshold mean value
    :param std: threshold standard deviation value
    :return: True if vector has a mean value greater or equal to the mean threshold and standard deviation smaller or
    equal to the std threshold
    """
    return (x.mean() >= mean) and (x.std() <= std)


def get_start_index_with_current_mean_std(i_hds_df: pd.DataFrame, index_increment: int, mean_start_value: int) -> int:
    """ Function returning start index of an exponential current decay based on its mean and st value

    :param i_hds_df: QH current signal
    :param index_increment: index increment for the rolling check of the mean and std values
    :param mean_start_value: mean value at the beginning
    :return: 0 if index not found, otherwise the found index
    """
    with_mean_std_df = i_hds_df.rolling(index_increment) \
        .apply(has_window_mean_std, args=(mean_start_value,), raw=True) \
        .shift(-index_increment + 1)

    with_mean_std_df.dropna(inplace=True)
    mask = with_mean_std_df[with_mean_std_df[with_mean_std_df.columns[0]] == 1.0]
    if len(mask) == 0:
        return 0
    else:
        return with_mean_std_df.index.get_loc(mask.index[0])


def get_decay_start_index(hds_dfs: List[pd.DataFrame], mean_start_value=50, index_increment=20) -> int:
    """ Function calculating the start index of the voltage and current decay in a QH supply unit

    :param hds_dfs: QH voltage and current signals
    :param mean_start_value: mean value at the beginning
    :param index_increment: index increment for the rolling check of the mean and std values
    :return: start index of an exponential decay
    """
    decay_start_index = 0
    i_hds_x = ['I_HDS_1', 'I_HDS_2', 'I_HDS_3', 'I_HDS_4']
    for i in i_hds_x:
        df_col = get_df_col(hds_dfs, i)
        if df_col is not None:
            i_hds_df, col_i_hds_df = df_col
            i_decay_start_index = get_start_index_with_current_mean_std(i_hds_df, index_increment, mean_start_value)
            decay_start_index = i_decay_start_index if i_decay_start_index > decay_start_index else decay_start_index
            if decay_start_index == 0:
                u_hds_df_col = get_df_col(hds_dfs, 'U_HDS_1')
                if u_hds_df_col is not None:
                    u_hds_df, col_u_hds_df = u_hds_df_col
                    decay_start_index = get_index_with_deviation_from_initial_value(u_hds_df[col_u_hds_df])
                    if (decay_start_index + 1) == len(u_hds_df[col_u_hds_df].index):
                        return 0
                    else:
                        return decay_start_index

    return decay_start_index


def get_index_with_deviation_from_initial_value(origin_series: pd.Series, initial_value=0, allowed_deviation=0.998):
    """ Function returning an index for which the initial value is below an initial value

    :param origin_series: a QH signal for which the deviation is searched
    :param initial_value: if 0, the mean of 20 points is considered
    :param allowed_deviation: allowed deviation from the initial value
    :return: index if found, otherwise 0
    """
    if initial_value == 0:
        initial_value = np.mean(origin_series.iloc[0:19].rolling(window=3).median())

    # get df with initial values
    filtered_series = SignalAnalysis.apply_frequency_filter(origin_series)
    initial_series = filtered_series[filtered_series.values > initial_value * allowed_deviation]
    if len(initial_series) > 0:
        return np.flatnonzero(origin_series.index == initial_series.index[-1])[0]
    else:
        return 0


def get_decay_only(hds_dfs: List[pd.DataFrame], mean_start_value=50) -> List[pd.DataFrame]:
    """ Method returning signals with exponential decays only

    :param hds_dfs: list of QH signals
    :param mean_start_value: mean start value for current
    :return: list of QH signals with exponential decays only
    """
    decay_start_index = get_decay_start_index(hds_dfs, mean_start_value=mean_start_value)
    if decay_start_index == 0:
        return hds_dfs
    else:
        return [hds_df.iloc[(decay_start_index + 1):len(hds_df)] for hds_df in hds_dfs]


def plot_qh_discharge(timestamp_qh: int, source_qh: str, u_i_hds_dfs: List[pd.DataFrame], r_hds_dfs: List[pd.DataFrame],
                      u_i_hds_ref_dfs: List[pd.DataFrame], r_hds_ref_dfs: List[pd.DataFrame]) -> None:
    """ Function plotting QH signals: voltage, current, resistance for actual and reference timestamp.

    :param timestamp_qh: actual QH PM timestamp
    :param source_qh: magnet name
    :param u_i_hds_dfs: voltages and currents of a QH
    :param r_hds_dfs: resistances of a QH
    :param u_i_hds_ref_dfs: reference voltages and currents of a QH
    :param r_hds_ref_dfs: reference resistances of a QH
    :return: None
    """
    no_signal = int(len(u_i_hds_dfs) / 2)

    fig, ax = plt.subplots(1, 3, figsize=(20, 7))
    for index in range(1, no_signal + 1):
        u_hds_df = get_df_col(u_i_hds_dfs, 'U_HDS_{}'.format(index))[0]
        u_hds_ref_df = get_df_col(u_i_hds_ref_dfs, 'U_HDS_{}'.format(index))[0]
        u_hds_df.plot(ax=ax[0])
        u_hds_ref_df.plot(ax=ax[0], style='--')

    ax[0].grid(True)
    ax[0].tick_params(labelsize=15)
    ax[0].set_xlabel("time, [s]", fontsize=15)
    ax[0].set_ylabel("U_HDS, [V]", fontsize=15)

    title = "Magnet: {}, Time Stamp: {}, U_HDS(t), I_HDS(t), and R_HDS(t)". \
        format(source_qh, Time.to_string_short(timestamp_qh))

    for index in range(1, no_signal + 1):
        i_hds_df = get_df_col(u_i_hds_dfs, 'I_HDS_{}'.format(index))[0]
        i_hds_ref_df = get_df_col(u_i_hds_ref_dfs, 'I_HDS_{}'.format(index))[0]
        i_hds_df.plot(ax=ax[1], title=title)
        i_hds_ref_df.plot(ax=ax[1], style='--')

    ax[1].title.set_size(20)
    ax[1].grid(True)
    ax[1].tick_params(labelsize=15)
    ax[1].set_xlabel("time, [s]", fontsize=15)
    ax[1].set_ylabel("I_HDS, [A]", fontsize=15)

    for r_hds_df, r_hds_ref_df in zip(r_hds_dfs, r_hds_ref_dfs):
        r_hds_df.plot(ax=ax[2])
        r_hds_ref_df.plot(ax=ax[2], style='--')
    ax[2].grid(True)
    ax[2].tick_params(labelsize=15)
    ax[2].set_xlabel("time, [s]", fontsize=15)
    ax[2].set_ylabel("R_HDS (Calculated), [Ohm]", fontsize=15)

    plt.show()


def plot_qh_discharge_hwc(timestamp_qh: int, source_qh: str, u_i_hds_dfs: List[pd.DataFrame],
                          r_hds_dfs: List[pd.DataFrame], u_i_hds_ref_dfs: List[pd.DataFrame],
                          r_hds_ref_dfs: List[pd.DataFrame]) -> None:
    """ Function plotting QH signals: voltage, current, resistance for actual and reference timestamp.
    The layout is suited for HWC analysis.

    :param timestamp_qh: actual QH PM timestamp
    :param source_qh: magnet name
    :param u_i_hds_dfs: voltages and currents of a QH
    :param r_hds_dfs: resistances of a QH
    :param u_i_hds_ref_dfs: reference voltages and currents of a QH
    :param r_hds_ref_dfs: reference resistances of a QH
    :return: None
    """
    no_signal = int(len(u_i_hds_dfs) / 2)

    title = "Magnet: {}, Time Stamp: {}, U_HDS(t), I_HDS(t), and R_HDS(t)". \
        format(source_qh, Time.to_string_short(timestamp_qh))

    fig, ax = plt.subplots(2, 2, figsize=(20, 14))
    fig.suptitle(title, fontsize=20)

    for index in range(1, no_signal + 1):
        u_hds_df = get_df_col(u_i_hds_dfs, 'U_HDS_{}'.format(index))[0]
        u_hds_ref_df = get_df_col(u_i_hds_ref_dfs, 'U_HDS_{}'.format(index))[0]
        u_hds_df.plot(ax=ax[0][0])
        u_hds_ref_df.plot(ax=ax[0][0], style='--')

    ax[0][0].grid(True)
    ax[0][0].tick_params(labelsize=15)
    ax[0][0].set_xlabel("time, [s]", fontsize=15)
    ax[0][0].set_ylabel("U_HDS, [V]", fontsize=15)

    for index in range(1, no_signal + 1):
        u_hds_df = get_df_col(u_i_hds_dfs, 'U_HDS_{}'.format(index))[0]
        u_hds_ref_df = get_df_col(u_i_hds_ref_dfs, 'U_HDS_{}'.format(index))[0]
        u_hds_df.plot(ax=ax[1][0], logy=True)
        u_hds_ref_df.plot(ax=ax[1][0], logy=True, style='--')
    ax[1][0].grid(True)
    ax[1][0].tick_params(labelsize=15)
    ax[1][0].set_xlabel("time, [s]", fontsize=15)
    ax[1][0].set_ylabel("U_HDS, [V]", fontsize=15)

    for index in range(1, no_signal + 1):
        i_hds_df = get_df_col(u_i_hds_dfs, 'I_HDS_{}'.format(index))[0]
        i_hds_ref_df = get_df_col(u_i_hds_ref_dfs, 'I_HDS_{}'.format(index))[0]
        i_hds_df.plot(ax=ax[0][1])
        i_hds_ref_df.plot(ax=ax[0][1], style='--')
    ax[0][1].title.set_size(20)
    ax[0][1].grid(True)
    ax[0][1].tick_params(labelsize=15)
    ax[0][1].set_xlabel("time, [s]", fontsize=15)
    ax[0][1].set_ylabel("I_HDS, [A]", fontsize=15)

    for index in range(1, no_signal + 1):
        i_hds_df = get_df_col(u_i_hds_dfs, 'I_HDS_{}'.format(index))[0]
        i_hds_ref_df = get_df_col(u_i_hds_ref_dfs, 'I_HDS_{}'.format(index))[0]
        i_hds_df.plot(ax=ax[1][1], logy=True)
        i_hds_ref_df.plot(ax=ax[1][1], logy=True, style='--')
    ax[1][1].title.set_size(20)
    ax[1][1].grid(True)
    ax[1][1].tick_params(labelsize=15)
    ax[1][1].set_xlabel("time, [s]", fontsize=15)
    ax[1][1].set_ylabel("I_HDS, [A]", fontsize=15)

    plt.show()

    fig, ax = plt.subplots(1, 1, figsize=(20, 7))

    for r_hds_df, r_hds_ref_df in zip(r_hds_dfs, r_hds_ref_dfs):
        r_hds_df.plot(ax=ax)
        r_hds_ref_df.plot(ax=ax, style='--')
    ax.grid(True)
    ax.tick_params(labelsize=15)
    ax.set_xlabel("time, [s]", fontsize=15)
    ax.set_ylabel("R_HDS (Calculated), [Ohm]", fontsize=15)

    plt.show()


def analyze_single_qh(timestamp_qh: int, source_qh: str, u_hds_dfs: List[pd.DataFrame], i_hds_dfs: List[pd.DataFrame],
                      u_hds_ref_dfs: List[pd.DataFrame], i_hds_ref_dfs: List[pd.DataFrame], circuit_type='RB',
                      plot_qh_discharge=plot_qh_discharge, mean_start_value=50, current_offset=0.0) -> bool:
    """ Function analyzing a single QH discharge. It is finding the start of a pseudo-exponential decay of current
    and voltage and calculates resistance. All these signals are plotted against corresponding references. Plot function
    can be chosen (either for an FPA or HWC analysis). Initial and final voltage, time constant, initial resistance
    are calculated and compared to references.

    :param timestamp_qh: QH PM timestamp
    :param source_qh: magnet name
    :param u_hds_dfs: list of actual QH voltages
    :param i_hds_dfs: list of actual QH currents
    :param u_hds_ref_dfs: list of reference QH voltages
    :param i_hds_ref_dfs: list of reference QH currents
    :param circuit_type: type of a circuit
    :param plot_qh_discharge: function for plotting of voltages, currents, and resistances
    :param mean_start_value: mean start value of current
    :param current_offset: value of the current offset to subtract from the input currents for featurea and resistance
    calculation
    :return: True if QH discharge is labelled ok, otherwise False
    """

    # Take the raw signals for plotting the entire PM buffer
    u_i_hds_raw_dfs = u_hds_dfs + i_hds_dfs
    u_i_hds_raw_ref_dfs = u_hds_ref_dfs + i_hds_ref_dfs

    # Subtract current offset
    i_hds_no_offset_dfs = [i_hds_df - current_offset for i_hds_df in i_hds_dfs
                           if not i_hds_df.empty]
    i_hds_ref_no_offset_dfs = [i_hds_ref_df - current_offset for i_hds_ref_df in i_hds_ref_dfs
                               if not i_hds_ref_df.empty]

    # Get decay only
    u_i_hds_dfs = get_decay_only(u_hds_dfs + i_hds_no_offset_dfs, mean_start_value)
    u_i_hds_ref_dfs = get_decay_only(u_hds_ref_dfs + i_hds_ref_no_offset_dfs, mean_start_value)

    # Filter
    u_i_hds_dfs = [u_i_hds_df.rolling(window=3, min_periods=1).median() for u_i_hds_df in u_i_hds_dfs]
    u_i_hds_ref_dfs = [u_i_hds_ref_df.rolling(window=3, min_periods=1).median() for u_i_hds_ref_df in u_i_hds_ref_dfs]

    # Synchronize time to 0
    # # For plotting
    u_i_hds_raw_dfs = [SignalIndexConversion.synchronize_df(u_i_hds_raw_df, u_i_hds_df.index[0])
                       for u_i_hds_raw_df, u_i_hds_df in zip(u_i_hds_raw_dfs, u_i_hds_dfs)]
    u_i_hds_raw_ref_dfs = [SignalIndexConversion.synchronize_df(u_i_hds_raw_ref_df, u_i_hds_ref_df.index[0])
                           for u_i_hds_raw_ref_df, u_i_hds_ref_df in zip(u_i_hds_raw_ref_dfs, u_i_hds_ref_dfs)]

    # # For feature engineering
    u_i_hds_dfs = [SignalIndexConversion.synchronize_df(u_i_hds_df, u_i_hds_df.index[0]) for u_i_hds_df in u_i_hds_dfs]
    u_i_hds_ref_dfs = [SignalIndexConversion.synchronize_df(u_i_hds_ref_df, u_i_hds_ref_df.index[0]) for u_i_hds_ref_df
                       in u_i_hds_ref_dfs]

    # Calculate resistance
    r_hds_dfs = calculate_resistance(u_i_hds_dfs)
    r_hds_ref_dfs = calculate_resistance(u_i_hds_ref_dfs)

    timestamp_ref = Reference.get_quench_heater_reference_discharge('RB', source_qh)

    # Calculate features
    # # compare to reference values
    first_last_u_df = FeatureBuilder().with_signal(u_i_hds_dfs) \
        .calculate_features(features=['first', 'last20mean'], index=timestamp_qh)

    first_last_u_comp_df = Reference.compare_features_to_reference(first_last_u_df, circuit_type, 'QH',
                                                                   wildcard={'CELL': source_qh})

    first_r_df = FeatureBuilder().with_signal(r_hds_dfs) \
        .calculate_features(features='first20mean', index=timestamp_qh)
    first_r_ref_df = FeatureBuilder().with_signal(r_hds_ref_dfs) \
        .calculate_features(features='first20mean', index=timestamp_qh)
    first_r_comp_df = Reference.compare_difference_of_features_to_reference(first_r_df, first_r_ref_df, circuit_type,
                                                                            'QH', wildcard={'CELL': source_qh})

    # # compare to the reference discharge
    tau_df = FeatureBuilder().with_signal(u_i_hds_dfs) \
        .calculate_features(features=['tau_charge'], index=timestamp_qh)
    tau_ref_df = FeatureBuilder().with_signal(u_i_hds_ref_dfs) \
        .calculate_features(features=['tau_charge'], index=timestamp_ref)

    tau_comp_df = Reference.compare_difference_of_features_to_reference(tau_df, tau_ref_df, circuit_type,
                                                                        'QH', wildcard={'CELL': source_qh})

    # # Plot signals
    plot_qh_discharge(timestamp_qh, source_qh, u_i_hds_raw_dfs, r_hds_dfs, u_i_hds_raw_ref_dfs, r_hds_ref_dfs)

    # Display comparison
    display_html(first_last_u_comp_df.style.set_table_attributes("style='display:inline'").set_caption(
        'Comparison of the inititial and final voltage').apply(lambda s: first_last_u_comp_df['result'] \
                                                               .map(
        {False: 'background-color: red', True: ''}))._repr_html_()
                 + first_r_comp_df.style.set_table_attributes("style='display:inline'").set_caption(
        'Comparison of the initial resisitance to the reference').apply(lambda s: first_r_comp_df['result'] \
                                                                        .map(
        {False: 'background-color: red', True: ''}))._repr_html_()
                 + tau_comp_df.style.set_table_attributes("style='display:inline'").set_caption(
        'Comparison of the discharge characteristic time to the reference').apply(lambda s: tau_comp_df['result'] \
                                                                                  .map(
        {False: 'background-color: red', True: ''}))._repr_html_(), raw=True)

    # Compare features
    is_first_last_u_correct = first_last_u_comp_df['result'].all()
    is_first_r_correct = first_r_comp_df['result'].all()
    is_tau_correct = tau_comp_df['result'].all()

    return is_first_last_u_correct and is_first_r_correct and is_tau_correct


class QuenchHeaterRbAnalysis(CircuitAnalysis):
    """ Class with methods for RB QH analysis

    """

    def analyze_qh(self, source_timestamp_qh_df, u_hds_dfs, i_hds_dfs, u_hds_ref_dfs, i_hds_ref_dfs, current_offset=0.0):
        """ Method analyzing quench heater discharge with voltaga and current (for an actual and reference discharge)
        :param source_timestamp_qh_df: dataframe with PM source and timestamp for QH events
        :param u_hds_dfs: list of list of QH voltages (length of the outer list equal to the number of quenched mangets)
        :param i_hds_dfs: list of list of QH currents (length of the outer list equal to the number of quenched mangets)
        :param u_hds_ref_dfs: reference list of list of QH voltages (length of the outer list equal to the number of
        quenched mangets)
        :param i_hds_ref_dfs: reference list of list of QH voltages (length of the outer list equal to the number of
        quenched mangets)
        """
        index_max = source_timestamp_qh_df.index[-1]
        for index, row in source_timestamp_qh_df.iterrows():
            source_qh, timestamp_qh = row[['source', 'timestamp']]
            print('{}/{}: Analysing quench heater signals of {} on {}, {}' \
                  .format(index, index_max, source_qh, Time.to_string(timestamp_qh), timestamp_qh))

            is_qh_correct = analyze_single_qh(timestamp_qh, source_qh, u_hds_dfs[index], i_hds_dfs[index],
                                              u_hds_ref_dfs[index], i_hds_ref_dfs[index], current_offset=current_offset)

            # Update results table
            if self.results_table is not None:
                self.results_table.loc[index, 'QH analysis'] = 'Pass' if is_qh_correct else 'Fail'

            # Show next
            check_show_next(index, index_max, self.is_automatic)
