from typing import List, Union, Tuple

import pandas as pd
from tqdm.notebook import tqdm

from lhcsmapi.analysis.CircuitQuery import CircuitQuery, execution_count
from lhcsmapi.analysis.decorators import check_nan_timestamp
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.reference.Reference import Reference


class QuenchHeaterQuery(CircuitQuery):
    """ Class for query of quench heater events and signals (both voltage and current). Works for actual and reference.

    """
    @execution_count()
    @check_nan_timestamp(return_type=pd.DataFrame())
    def find_source_timestamp_qh(self, t_start: Union[int, str, float], duration=[(10, 's'), (400, 's')]
                                  ) -> pd.DataFrame:
        """ Method searching PM QH events
        If the start is NaN, then an empty dataframe is returned.

        :param t_start: start time of a search
        :type t_start: Union[str, int, float]
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: pd.DataFrame with source and timestamp of the PM QDS events
        """
        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='QH', source='*') \
            .event_query(verbose=self.verbose) \
            .filter_source(self.circuit_name, 'QH') \
            .drop_duplicate_source() \
            .sort_values(by='timestamp').df

    @execution_count()
    def query_qh_pm(self, source_timestamp_qh_df: pd.DataFrame, *, signal_names: List[str],
                    is_ref=False) -> List[List[pd.DataFrame]]:
        """ Method querying PM for QH signals (either voltage or current or both)
        Signals are synchronized to the QH PM event and index is converted to seconds.

        :param source_timestamp_qh_df: dataframe with PM source and timestamp for QH events
        :type source_timestamp_qh_df: pd.DataFrame
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param is_ref: boolean flag to determine whether a query of reference or actual discharge is necessary
        :type is_ref: bool
        :return: List of list of QH signals. The outer list has as many elements as QH events and the length
            of the inner is equal to the number of QH signals.
        """
        hds_dfs = []
        for index, row in tqdm(source_timestamp_qh_df.iterrows(), total=source_timestamp_qh_df.shape[0],
                               desc='Querying PM'):
            source_qds = row['source']
            if is_ref:
                timestamp_qh = Reference.get_quench_heater_reference_discharge(self.circuit_type, source_qds)
            else:
                timestamp_qh = row['timestamp']

            hds_df = QueryBuilder().with_pm() \
                .with_timestamp(timestamp_qh) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='QH', signal=signal_names, source=source_qds,
                               wildcard={'CELL': source_qds}) \
                .signal_query() \
                .synchronize_time(timestamp_qh) \
                .convert_index_to_sec().dfs

            hds_dfs.append(hds_df)

        return hds_dfs

