from typing import List, Tuple

from ipywidgets import HBox, Button, Label, IntProgress, Layout
import functools
from abc import ABC, abstractmethod

from lhcsmapi.Time import Time
from lhcsmapi.gui.pc.FgcBrowserAction import FgcBrowserAction
from lhcsmapi.gui.BaseModule import BaseModule


class FgcPmSearchBaseModule(BaseModule, ABC):
    """ Abstract class for search of FGC PM events by respective notebook browsers

    """

    def __init__(self):
        super().__init__()
        self.find_btn = Button(
            description='Find FGC PM entries',
            tooltip='Click to find FGC PM entries between indicated dates.',
            button_style='',  # 'success', 'info', 'warning', 'danger' or ''
        )
        self.find_btn.on_click(functools.partial(FgcPmSearchBaseModule.handle_find_button_event, self=self))

        # # Text field
        self.lbl_search = Label(value="Click button to search for FGC PM entries within the specified period of time")

        # # Progress bar
        self.progress_search = IntProgress(value=0, min=0, step=1, bar_style='success',
                                           orientation='horizontal', layout=Layout(width='20%', visibility='hidden'))

        self.widget = HBox(children=[self.find_btn, self.lbl_search, self.progress_search])

    def handle_find_button_event(b, self):
        self.mediator.notify(self, FgcBrowserAction.FIND_BUTTON_CLICKED)

    @abstractmethod
    def find_fgc_pm_entries_with_qps(self, start_time, end_time, circuit_name):
        """ Method finding FGC PM entries with QPS for a given time range and circuit name

        :param start_time: start time for the query
        :param end_time: end time for the query
        :param circuit_name: circuit name
        :return: a list of FGC source and timestamp tuples
        """
        pass

    def convert_fgc_pm_entries_to_options(self, fgc_pm_entries: List[Tuple[str, int]]) -> List[str]:
        """ Method converting a list of source and timestamp tuples into a list of strings to be displayed

        :param fgc_pm_entries: a list of FGC PM source and timestamp tuples
        :return: list of strings with FGC PM source and timestamp
        """

        options = []
        for fgc_pm_entry in fgc_pm_entries:
            if len(fgc_pm_entry) == 2:
                options.append('%s: %s' % (fgc_pm_entry[0], Time.to_string(fgc_pm_entry[1])))
            else:
                options.append('%s: %s; %s: %s' % (fgc_pm_entry[0], Time.to_string(fgc_pm_entry[1]),
                                                   fgc_pm_entry[2], Time.to_string(fgc_pm_entry[3])))
        return options

    def get_circuit_names(self, index: int, fgc_pm_entries: List[Tuple[str, int]]) -> str:
        """ Method returning circuit name for a selected element in the list

        :param index: index of a selected element in the FGC PM list
        :param fgc_pm_entries: a list of all source and timestamp FGC PM tuples
        :return: a selected circuit name
        """
        return fgc_pm_entries[index][0]

    def get_circuit_timestamps(self, index: int, fgc_pm_entries: List[Tuple[str, int]]) -> int:
        """ Method returning timestamp for a selected element in the list

        :param index: index of a selected element in the FGC PM list
        :param fgc_pm_entries: a list of all source and timestamp FGC PM tuples
        :return: a selected timestamp
        """
        return fgc_pm_entries[index][1]
