import getpass
import functools
from typing import List

from ipywidgets import Dropdown, Text, HBox, VBox, Select, Label, ToggleButtons, Layout
from IPython.display import display, HTML, clear_output, Image

from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.pyedsl.dbquery.DbQueryBuilder import explode
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.Time import Time
from lhcsmapi.gui.ModuleMediator import ModuleMediator
from lhcsmapi.gui.pc.FgcBrowserAction import FgcBrowserAction
from lhcsmapi.gui.pc.FgcPmSearchBaseModule import FgcPmSearchBaseModule
from lhcsmapi.gui.DateTimeBaseModule import DateTimeBaseModule
from lhcsmapi.gui.BrowserAction import BrowserAction
from lhcsmapi.metadata.SignalMetadata import SignalMetadata


def get_circuit_names_or_types(circuit_type):
    """ Method returns a list of circuit names or types if the list of names is very long
    Depending on the circuit type the list is modified accordingly:
    - independent circuits: RB, IT, IPD - full list of names
    - coupled circuits: RQ, 600A_RCBXHV, 600A_RCDO - list of reduced names to avoid duplications,
    e.g., RQ.A12 instead of RQD.A12 and RQF.A12
    - multiple circuits: 600A - family types in order to avoid very long lists,
    e.g. RU instead of RU.L4

    :param circuit_type: circuit type such as RB, RQ, 600A, IT, IPD, IPQ
    :type circuit_type: str
    :return: list of circuit names or types
    """
    if circuit_type in ['RB', 'IT']:
        return SignalMetadata.get_circuit_names(circuit_type)
    elif circuit_type == 'IPD':
        return SignalMetadata.get_circuit_names(['IPD2', 'IPD2_B1B2'])
    elif circuit_type == '60A':
        r60A_names = SignalMetadata.get_circuit_names('60A')
        r60A_names_prefix = list(set([r60A_name.split('.')[0] for r60A_name in r60A_names]))
        return sorted(r60A_names_prefix)
    elif circuit_type == '80-120A':
        r80_120A_names = SignalMetadata.get_circuit_names('80-120A')
        r80_120A_names_prefix = list(set([SignalMetadata.parse_80_120A_circuit_name_to_family_name(r80_120A_name)
                                          for r80_120A_name in r80_120A_names]))
        return sorted(r80_120A_names_prefix)
    elif circuit_type == 'IPQ':
        ipq_names = SignalMetadata.get_circuit_names(['IPQ2', 'IPQ4', 'IPQ8'])
        ipqs_names_prefix = list(set([ipq_name.split('.')[0] for ipq_name in ipq_names]))
        return sorted(ipqs_names_prefix)
    elif circuit_type == 'RQ':
        circuit_names = SignalMetadata.get_circuit_names('RQ')
        circuit_names_rqd = list(filter(lambda name: 'RQD' in name, circuit_names))
        circuit_names_rqd_set = {circuit_name.replace('RQD', 'RQ') for circuit_name in circuit_names_rqd}
        return sorted(list(circuit_names_rqd_set))
    elif circuit_type == '600A':
        circuit_names = SignalMetadata.get_circuit_names('600A')
        circuit_names_without_rcbx = list(filter(lambda name: 'RCBX' not in name, circuit_names))
        circuit_names_without_rcbx_rcd = list(filter(lambda name: 'RCD' not in name, circuit_names_without_rcbx))
        circuit_names_without_rcbx_rcd_rco = list(
            filter(lambda name: 'RCO' not in name, circuit_names_without_rcbx_rcd))
        return sorted(list({circuit_name.split('.')[0] for circuit_name in circuit_names_without_rcbx_rcd_rco}))
    elif circuit_type == '600A_RCBXHV':
        circuit_names = SignalMetadata.get_circuit_names('600A')
        circuit_names_rcbxh = list(filter(lambda name: 'RCBXH' in name, circuit_names))
        circuit_names_rcbxh_set = {circuit_name.replace('RCBXH', 'RCBX') for circuit_name in circuit_names_rcbxh}
        return sorted(list(circuit_names_rcbxh_set))
    elif circuit_type == '600A_RCDO':
        circuit_names = SignalMetadata.get_circuit_names('600A')
        circuit_names_rcd = list(filter(lambda name: 'RCD' in name, circuit_names))
        circuit_names_rcd_set = {circuit_name.replace('RCD', 'RC') for circuit_name in circuit_names_rcd}
        return sorted(list(circuit_names_rcd_set))
    else:
        raise KeyError('Circuit type {} is not supported'.format(circuit_type))


def _display_ipd_circuit(circuit_name: str) -> None:
    """ Method displaying in a Jupyter notebook a QPS schematic corresponding to an IPD circuit name.

    :param circuit_name: IPD circuit name
    :type circuit_name: str

    """
    circuit_class = MappingMetadata.get_qps_schematic_naming('IPD', circuit_name)

    if circuit_class == 'MBX':
        display(Image(url="https://gitlab.cern.ch/LHCData/lhc-sm-hwc/-/raw/master/figures/ipd/IPD_MBX_D1.png"))
    elif circuit_class == 'MBRC':
        display(Image(url="https://gitlab.cern.ch/LHCData/lhc-sm-hwc/-/raw/master/figures/ipd/IPD_MBRC_D2_MBRB_D4.png"))
    elif circuit_class == 'MBRS':
        display(Image(url="https://gitlab.cern.ch/LHCData/lhc-sm-hwc/-/raw/master/figures/ipd/IPD_MBRS_D3.png"))
    elif circuit_class == 'MBRB':
        display(Image(url="https://gitlab.cern.ch/LHCData/lhc-sm-hwc/-/raw/master/figures/ipd/IPD_MBRC_D2_MBRB_D4.png"))


def _display_ipq_circuit(circuit_name: str) -> None:
    """ Method displaying in a Jupyter notebook a QPS schematic corresponding to an IPQ circuit name.

    :param circuit_name: IPQ circuit name
    :type circuit_name: str

    """
    circuit_class = MappingMetadata.get_qps_schematic_naming('IPQ', circuit_name)

    if circuit_class == '2MQM':
        display(Image(url="https://gitlab.cern.ch/LHCData/lhc-sm-hwc/-/raw/master/figures/ipq/2MQM_QPS.png"))
    elif circuit_class == '2x2MQM':
        display(Image(url="https://gitlab.cern.ch/LHCData/lhc-sm-hwc/-/raw/master/figures/ipq/2x2MQM_QPS.png"))
    elif circuit_class == '2MQML':
        display(Image(url="https://gitlab.cern.ch/LHCData/lhc-sm-hwc/-/raw/master/figures/ipq/2MQML_QPS.png"))
    elif circuit_class == '2MQM+2MQML':
        display(Image(url="https://gitlab.cern.ch/LHCData/lhc-sm-hwc/-/raw/master/figures/ipq/2MQM_2MQML_QPS.png"))
    elif circuit_class == '2MQM+2MQMC':
        display(Image(url="https://gitlab.cern.ch/LHCData/lhc-sm-hwc/-/raw/master/figures/ipq/2MQM_2MQMC_QPS.png"))
    elif circuit_class == '2MQY':
        display(Image(url="https://gitlab.cern.ch/LHCData/lhc-sm-hwc/-/raw/master/figures/ipq/2MQY_QPS.png"))
    elif circuit_class == '2x2MQY':
        display(Image(url="https://gitlab.cern.ch/LHCData/lhc-sm-hwc/-/raw/master/figures/ipq/2x2MQY_QPS.png"))


@functools.lru_cache(maxsize=8)
def get_pc_to_circuit_metadata(selected_sector: str) -> List[List[str]]:
    """ Method returning power converter names belonging to a sector

    :param selected_sector: name of a selected sector
    :return: List of power converter names belonging to a sector
    """
    pc_to_circuit_metadata = []
    circuit_name_to_sector_df = MappingMetadata.read_circuit_name_to_sector()
    mask = circuit_name_to_sector_df['Sector name'] == selected_sector
    for index, row in circuit_name_to_sector_df[mask].iterrows():
        pc_name = SignalMetadata.get_circuit_name_metadata(row['Circuit type SIGMON'], row['Circuit name'])['PC'][
            '%PC%']
        if isinstance(pc_name, list):
            pc_to_circuit_metadata.append(pc_name)
        else:
            pc_to_circuit_metadata.append([pc_name])

    return pc_to_circuit_metadata


class FgcPmSearchModuleMediator(ModuleMediator):

    def __init__(self, date_time_module: DateTimeBaseModule, fgc_pm_search_module: FgcPmSearchBaseModule,
                 circuit_type='RB'):
        self._circuit_type = circuit_type
        # First row
        circuit_names = get_circuit_names_or_types(circuit_type)
        description = 'Circuit name prefix:' if circuit_type in ['600A', 'IPQ'] else 'Circuit name'
        self.select_circuit_name_dropdown = Dropdown(
            options=circuit_names,
            description=description,
            disabled=False,
        )

        self._date_time_module = date_time_module
        self._date_time_module.mediator = self

        self.done_by_txt = Text(placeholder='Your name', description='Done by:', value=getpass.getuser(),
                                disabled=True)

        self.analysis_mode_tgl_btn = ToggleButtons(
            options=['Automatic', 'Manual'],
            description='Analysis:',
            disabled=False,
            button_style='',
            tooltips=['Select automatic for automatic execution, comment, and acceptance.',
                      'Select manual to execute, comment, and accept each analysis manually.']
        )

        first_row = HBox(children=[self.select_circuit_name_dropdown, self._date_time_module.widget])

        # On select value change, update output
        self.select_circuit_name_dropdown.observe(
            functools.partial(FgcPmSearchModuleMediator.update_text_output_ee, self=self))

        # Second row
        second_row = HBox(children=[self.analysis_mode_tgl_btn, self.done_by_txt])

        # Third row
        self.fgc_pm_entries = []
        self._fgc_pm_search_module = fgc_pm_search_module
        self._fgc_pm_search_module.mediator = self
        third_row = fgc_pm_search_module.widget

        # Fourth row
        self.fgc_pm_sel = Select(
            options=[],
            rows=10,
            disabled=False,
            layout=Layout(width='50%')
        )

        # On select value change, update output
        self.fgc_pm_sel.observe(functools.partial(FgcPmSearchModuleMediator.update_text_output, self=self),
                                names=['value'])

        fourth_row = HBox(children=[Label(value='FGC PM Entries:'), self.fgc_pm_sel])

        # Fifth row - search for
        self.time_range = Text(placeholder='Time range for search of FGC', description='', value='5')
        self.time_range.observe(functools.partial(FgcPmSearchModuleMediator.update_text_output, self=self))
        self.time_range_label = Label(value='')
        fifth_row = HBox(
            children=[Label(value='Time range for search of FGC PM events around selected event [s]'), self.time_range,
                      self.time_range_label])

        # Final GUI
        self.widget = VBox(children=[first_row, second_row, third_row, fourth_row, fifth_row])

        display(self.widget)

        # For 600A display if the circuit contains EE or not
        self._print_ee_600A()

    def notify(self, sender: object, event: str) -> None:
        start_date_time = self._date_time_module.get_start_date_time()
        end_date_time = self._date_time_module.get_end_date_time()
        circuit_name = self.get_circuit_name()

        if event == BrowserAction.DATE_CHANGED:
            pass
        elif event == FgcBrowserAction.FIND_BUTTON_CLICKED:
            if not DateTimeBaseModule.check_date_time(start_date_time, end_date_time):
                print('Start time is not later than the end time.')
            else:
                self.fgc_pm_entries = self._fgc_pm_search_module.find_fgc_pm_entries_with_qps(start_date_time,
                                                                                              end_date_time,
                                                                                              circuit_name)
                options = self._fgc_pm_search_module.convert_fgc_pm_entries_to_options(self.fgc_pm_entries)
                self.fgc_pm_sel.options = options
                if not options:
                    print('No FGC PM events for %s from %s to %s' % (circuit_name, start_date_time, end_date_time))
        else:
            raise KeyError(f'Action {event} not supported!')

    @staticmethod
    def update_text_output_ee(value, self):
        self.fgc_pm_entries = []
        self.fgc_pm_sel.options = []
        clear_output(wait=True)
        display(self.widget)
        self._print_ee_600A()

    def _print_ee_600A(self):
        if self._circuit_type == '600A':
            circuit_family = self.select_circuit_name_dropdown.value
            circuit_name = list(filter(lambda col: circuit_family in col, SignalMetadata.get_circuit_names('600A')))[0]

            if 'EE' in SignalMetadata.get_system_types_per_circuit_name(self._circuit_type, circuit_name):
                print('Circuit family %s has an EE system.' % circuit_family)
            else:
                print('Circuit family %s does not have an EE system.' % circuit_family)

    @staticmethod
    def update_text_output(value, self):
        clear_output(wait=True)
        display(self.widget)
        if self.fgc_pm_entries:
            index = 0 if self.fgc_pm_sel.index is None else self.fgc_pm_sel.index

            if self.get_circuit_type() in ['RB', 'IT', 'IPD']:
                circuit_name = self.select_circuit_name_dropdown.value
                timestamp = self._fgc_pm_search_module.get_circuit_timestamps(index, self.fgc_pm_entries)
                pm_event_timestamp = '%s: %s' % (circuit_name, Time.to_string_short(timestamp))
            else:
                circuit_name = self.fgc_pm_entries[index][0]
                timestamp = self.fgc_pm_entries[index][1]
                options = self._fgc_pm_search_module.convert_fgc_pm_entries_to_options(self.fgc_pm_entries)
                pm_event_timestamp = options[index]

            if self.get_circuit_type() == 'IPQ':
                _display_ipq_circuit(circuit_name)
            elif self._circuit_type == 'IPD':
                _display_ipd_circuit(self.get_circuit_name())

            self._print_ee_600A()
            print('Selected FGC PM event %s.' % pm_event_timestamp)
            print('The analysis is executed in %s mode.' % self.analysis_mode_tgl_btn.value)

            # Displaying FGC PM timestamps from the same sector
            circuit_name_to_sector_df = MappingMetadata.read_circuit_name_to_sector()
            print(circuit_name)
            selected_sector = circuit_name_to_sector_df[circuit_name_to_sector_df['Circuit name'] == circuit_name][
                'Sector name'].values[0]
            print('FGC PM events in circuits of sector %s:' % selected_sector)

            self.time_range_label = 'Searching for PM events, please wait...'
            # for HWC 2021 and Run3 should be lhc_self_pmd
            source_timestamp_df = QueryBuilder().with_pm() \
                .with_duration(t_start=timestamp,
                               duration=[(float(self.time_range.value), 's'), (float(self.time_range.value), 's')]) \
                .with_query_parameters(system='FGC', className=['51_self_pmd', '92_self'], source='*') \
                .event_query().df

            circuit_name_to_sector_df = circuit_name_to_sector_df[
                circuit_name_to_sector_df['Sector name'] == selected_sector]
            circuit_name_to_sector_df['PC name'] = get_pc_to_circuit_metadata(selected_sector)
            circuit_name_to_sector_df = explode(circuit_name_to_sector_df, 'PC name')
            circuit_name_to_sector_df = circuit_name_to_sector_df[
                circuit_name_to_sector_df['PC name'].isin(source_timestamp_df['source'])]
            circuit_name_to_sector_df = circuit_name_to_sector_df \
                .merge(source_timestamp_df, left_on=['PC name'], right_on=['source']) \
                .drop(columns=['source', 'Circuit type Layout'], axis=1)
            circuit_name_to_sector_df['dt [ms]'] = (circuit_name_to_sector_df['timestamp'] - timestamp) / 1e6
            circuit_name_to_sector_df['timestamp'] = circuit_name_to_sector_df['timestamp'].apply(
                lambda col: Time.to_string_short(col))
            circuit_name_to_sector_df.rename(
                columns={'Circuit type SIGMON': 'Circuit type'}, inplace=True)
            circuit_name_to_sector_df.sort_values(by='dt [ms]', inplace=True)
            circuit_name_to_sector_df.reset_index(drop=True, inplace=True)
            display(HTML(circuit_name_to_sector_df.to_html()))

            self.time_range_label = 'Search for PM events is completed.'

    def get_circuit_type(self):
        return self._circuit_type

    def get_circuit_name(self):
        return self.select_circuit_name_dropdown.value

    def is_automatic_mode(self):
        return self.analysis_mode_tgl_btn.value == 'Automatic'

    def get_author(self):
        return self.done_by_txt.value

    def get_fgc_timestamp(self):
        if len(self.fgc_pm_sel.options) == 0:
            raise ValueError('The list of FGC PM Entries should contain at least one element. '
                             'Please adjust circuit signal as well as start and end date.')

        return self._fgc_pm_search_module.get_circuit_timestamps(self.fgc_pm_sel.index, self.fgc_pm_entries)

    def get_fgc_circuit(self):
        if len(self.fgc_pm_sel.options) == 0:
            raise ValueError('The list of FGC PM Entries should contain at least one element. '
                             'Please adjust circuit signal as well as start and end date.')

        return self._fgc_pm_search_module.get_circuit_names(self.fgc_pm_sel.index, self.fgc_pm_entries)
