from typing import List, Tuple

from lhcsmapi.Time import Time
from lhcsmapi.gui.pc.FgcPmSearchBaseModule import FgcPmSearchBaseModule
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder


def query_pm_convert_pc_to_circuit_name(start_time, end_time, circuit_name: str) -> List[Tuple[str, int]]:
    """ Method querying FGC PM source and timestamp tuple and taking only those corresponding to the given circuits

    :param start_time: start time for a query
    :param end_time: end time for a query
    :param circuit_name: circuit name
    :return: a list of PM source and timestamp tuples corresponding to circuit names
    """
    if circuit_name in SignalMetadata.get_circuit_names('IPD2'):
        circuit_type = 'IPD2'
    elif circuit_name in SignalMetadata.get_circuit_names('IPD2_B1B2'):
        circuit_type = 'IPD2_B1B2'
    else:
        raise ValueError('Circuit name %s does not map to internal metadata.' % circuit_name)

    # Query PM for the specified period of tim
    source_timestamp_df = QueryBuilder().with_pm() \
        .with_duration(t_start=start_time, t_end=end_time) \
        .with_circuit_type(circuit_type) \
        .with_metadata(circuit_name=circuit_name, system='PC', source='*') \
        .event_query().df

    source_timestamp_df = source_timestamp_df.drop_duplicates('timestamp')

    # Convert PC names to circuit names
    output = []
    for index, row in source_timestamp_df.iterrows():
        output.append((row['timestamp']))

    return output


class IpdFgcPmSearchBaseModule(FgcPmSearchBaseModule):
    """ A class with methods to search for IPD FGC PM events to fill-in a search field in the FGC PM search module.

    """

    def get_circuit_names(self, index, fgc_pm_entries):
        raise NotImplementedError('This method is not implemented!')

    def convert_fgc_pm_entries_to_options(self, fgc_pm_entries: List[Tuple[str, int]]) -> List[str]:
        """ Method converting a list of source and timestamp tuples into a list of strings to be displayed

        :param fgc_pm_entries: a list of FGC PM source and timestamp tuples
        :return: list of strings with FGC PM source and timestamp
        """
        return [Time.to_string(fgc_pm_entry) for fgc_pm_entry in fgc_pm_entries]

    def find_fgc_pm_entries_with_qps(self, start_time, end_time, circuit_name: str) -> List[Tuple[str, int]]:
        """ Method finding FGC PM entries with QPS for a given time range and circuit name

        :param start_time: start time for the query
        :param end_time: end time for the query
        :param circuit_name: circuit name
        :return: a list of FGC source and timestamp tuples
        """
        self.lbl_search.value = 'Querying PM for FGC events'
        self.progress_search.layout.visibility = 'visible'
        self.progress_search.max = 1
        self.progress_search.description = ''

        output = query_pm_convert_pc_to_circuit_name(start_time, end_time, circuit_name)

        self.lbl_search.value = 'Query completed'
        self.progress_search.layout.visibility = 'visible'
        self.progress_search.value = 1
        self.progress_search.description = ''

        return output
