from typing import List, Tuple, Union, Any

from lhcsmapi.Time import Time
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.dbsignal.post_mortem.PmDbRequest import PmDbRequest
from lhcsmapi.gui.pc.RbFgcPmSearchBaseModule import FgcPmSearchBaseModule


def check_quench_presence_around_fqc_timestamp(circuit_type: str, circuit_name: str,
                                               source_timestamp_qds: List[Tuple[str, int]]) -> List[Tuple[str, int]]:
    """ Method checking whether there is a quench around an FGC PM timestamp

    :param circuit_type: circuit type
    :param circuit_name: circuit name
    :param source_timestamp_qds: QDS PM source and timestamp
    :return: a list of tuples with FGC source and timestamp
    """
    source_timestamp_qds_filtered_act = []
    for source, timestamp in source_timestamp_qds:
        rq_circuits_with_cell = SignalMetadata.get_sector_names_from_components(circuit_type, 'MAGNET', source)
        # ['RQD.A12', 'RQF.A12'] -> 'RQD.A12' -> ['RQD', 'A12'] -> 'A12'
        rq_sector_with_cell = rq_circuits_with_cell[0].split('.')[1]
        rq_sector_with_circuit_name = circuit_name.split('.')[1]
        if rq_sector_with_cell == rq_sector_with_circuit_name:
            source_timestamp_qds_filtered_act.append((source, timestamp))
    return source_timestamp_qds_filtered_act


class RqFgcPmSearchBaseModule(FgcPmSearchBaseModule):
    """ A class with methods to search for RQ FGC PM events to fill-in a search field in the FGC PM search module.

    """

    def get_circuit_names(self, index: int, fgc_pm_entries: List[Tuple[str, int]]) -> List[str]:
        """ Method for getting a list of RQ circuit names (RQD.AXY, RQF.AXY)

        :param index: selected index in the list
        :param fgc_pm_entries: list of FGC PM events for RQD and RQF (source and timestamp available for both)
        :return: a two-element list with selected RQD and RQF names
        """
        return sorted(list([fgc_pm_entries[index][0], fgc_pm_entries[index][2]]))

    def get_circuit_timestamps(self, index, fgc_pm_entries):
        """ Method returning timestamp for a selected element in the list

        :param index: index of a selected element in the FGC PM list
        :param fgc_pm_entries: a list of all source and timestamp FGC PM tuples
        :return: a selected timestamp
        """
        if len(fgc_pm_entries[index]) == 2:
            return fgc_pm_entries[index][1], fgc_pm_entries[index][1]
        else:
            return fgc_pm_entries[index][1], fgc_pm_entries[index][3]

    def convert_fgc_pm_entries_to_options(self, fgc_pm_entries: List[Tuple[str, int]]) -> List[str]:
        """ Method converting a list of source and timestamp tuples with FGC PM events to a list of strings

        :param fgc_pm_entries: a list of source and timestamp tuples with FGC PM events
        :return: list of options to display
        """
        return ['{}: {}; {}: {}'.format(x[0], Time.to_string(x[1]), x[2], Time.to_string(x[3])) for x in fgc_pm_entries]

    def find_fgc_pm_entries_with_qps(self, start_time, end_time, circuit_name: str
                                     ) -> List[List[Union[Union[str, int], Any]]]:
        """ Method finding FGC PM entries with QPS for a given time range and circuit name

        :param start_time: start time for the query
        :param end_time: end time for the query
        :param circuit_name: circuit name
        :return: a list of FGC source and timestamp tuples
        """
        # convert RQ.AXX into [RQD.AXX, RQF.AXX]
        circuit_names = [circuit_name.replace('RQ', rq_circuit_name) for rq_circuit_name in ['RQD', 'RQF']]

        source_timestamp_fgc = self._find_fgc_pm_entries(start_time, end_time, 'RQ', circuit_names)
        return self._keep_fgc_rqd_with_qps_and_rqf('RQ', circuit_names, source_timestamp_fgc)

    def _find_fgc_pm_entries(self, start_time, end_time, circuit_type: str, circuit_names: List[str]
                             ) -> List[Tuple[str, int]]:
        """ Method finding FGC PM entries for a given start time and end time

        :param start_time: query start time
        :param end_time: query end time
        :param circuit_type: circuit type
        :param circuit_names: list of circuit names
        :return: a list of source and timestamp tuples
        """

        metadata_fgc_rqd = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_names[0], 'PC',
                                                                               'PM')

        quotient, remainder = Time.modulo_day_in_unix_timestamp(start_time, end_time)
        event_duration_nsec = int(24 * 60 * 60 * 1e9)
        source_timestamp_fgc = []

        self.progress_search.layout.visibility = 'visible'
        self.progress_search.max = quotient
        self.progress_search.description = f'Days: 0/{quotient}'

        for day in range(quotient):
            start_time_temp = Time.to_unix_timestamp(start_time) + int(day * event_duration_nsec)

            # Update GUI
            self.lbl_search.value = 'Querying PM for FGC entries on {}'.format(Time.to_string(start_time_temp))

            source_timestamp_fgc += PmDbRequest.find_events(metadata_fgc_rqd['source'], metadata_fgc_rqd['system'],
                                                            metadata_fgc_rqd['className'], t_start=start_time_temp,
                                                            duration=[(event_duration_nsec, 'ns')])
            self.progress_search.description = f'Days: {day + 1}/{quotient}'
            self.progress_search.value = day + 1

        if remainder > 0:
            start_time_temp = Time.to_unix_timestamp(end_time) - remainder
            self.lbl_search.value = 'Querying PM for FGC entries on {}'.format(Time.to_string(start_time_temp))
            source_timestamp_fgc += PmDbRequest.find_events(metadata_fgc_rqd['source'], metadata_fgc_rqd['system'],
                                                            metadata_fgc_rqd['className'], t_start=start_time_temp,
                                                            duration=[(remainder, 'ns')])

        return source_timestamp_fgc

    def _keep_fgc_rqd_with_qps_and_rqf(self, circuit_type: str, circuit_names: List[str],
                                       source_timestamp_fgc: List[Tuple[str, int]]
                                       ) -> List[List[Union[Union[str, int], Any]]]:
        """ Method keeping FGC PM events with QPS and RQF

        :param circuit_type: circuit type
        :param circuit_names: List of RQ circuit names (RQD.AXY, RQF.AXY)
        :param source_timestamp_fgc: a list of tuples with source and timestamp
        :return: a list of tuples with source and timestamp that has an QPS PM event around an FGC as well as a
        corresponding RQF FGC timestamp
        """
        # Check if an RQD FGC timestamp has
        # # a QPS discharge around
        # # an RQF FGC timestamp
        source_timestamp_fgc_with_qps = []
        metadata_fgc_rqf = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_names[1], 'PC',
                                                                               'PM')
        metadata_qds = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_names[0], 'QDS', 'PM')

        for index, elem in enumerate(source_timestamp_fgc):
            source_fgc, timestamp_fgc = elem
            # Update GUI
            self.lbl_search.value = 'Checking if there are QPS PM around an FGC timestamp {}'.format(
                Time.to_string(elem[1]))
            self.progress_search.description = f'FGC: {index + 1}/{len(source_timestamp_fgc)}'
            self.progress_search.value = index + 1

            # Check if there is at least one QPS timestamp around FGC timestamp
            source_timestamp_qds = PmDbRequest.find_events('*', metadata_qds['system'], metadata_qds['className'],
                                                           t_start=timestamp_fgc,
                                                           duration=[(60, 's'), (300, 's')])

            source_timestamp_qds_filtered = check_quench_presence_around_fqc_timestamp(circuit_type, circuit_names[0],
                                                                                       source_timestamp_qds)

            # Check if there is a corresponding RQF timestamp
            source_timestamp_fgc_rqf = PmDbRequest.find_events(metadata_fgc_rqf['source'], metadata_fgc_rqf['system'],
                                                               metadata_fgc_rqf['className'], t_start=timestamp_fgc,
                                                               duration=[(20, 'ms'), (20, 'ms')])
            # The FGC frequency is 50 Hz so we must allow a 20 ms time window (Zinur 30.03.2020)
            if source_timestamp_qds_filtered and source_timestamp_fgc_rqf:
                source_timestamp_fgc_with_qps.append(
                    [circuit_names[0], timestamp_fgc, circuit_names[1], source_timestamp_fgc_rqf[0][1]])

        self.lbl_search.value = 'Query completed.'

        return source_timestamp_fgc_with_qps
