from typing import List, Union

import pandas as pd
import numpy as np
from scipy.optimize import leastsq

from lhcsmapi.pyedsl.dbsignal import SignalIndexConversion


def flatten_nested_lists(nested_lists) -> List:
    """ Function recursively flattening nested lists/tuples into a single lists

    :param nested_lists: list, tuple, list of lists, tuple of tuples, and all possible permutations
    :return: flattened list
    """
    result = []
    for el in nested_lists:
        if isinstance(el, (list, tuple)):
            result.extend(flatten_nested_lists(el))
        else:
            result.append(el)
    return result


def convert_table_df_into_row(df: pd.DataFrame, index: int) -> pd.DataFrame:
    """ Function converting a table into a single row. The indices are joined with columns by an underscore

    :param df: input tabular DataFrame
    :type df: pd.DataFrame
    :param index: index for output row DataFrame
    :type index: int
    :return: single row dataframe with an index
    """
    df_out = pd.DataFrame(df.stack())
    df_out.index = df_out.index.map('_'.join)
    df_out = df_out.T
    df_out.index = [index]

    return df_out


def first(x: pd.Series) -> float:
    """ Function returning the first value of the input vector

    :param x: input vector
    :type x: pd.Series
    :return: first value of the input vector
    """
    return x.iloc[0]


def first20mean(x: pd.Series) -> np.ndarray:
    """ Function returning the mean value of the first 20 elements of the input vector

    :param x: input vector
    :type x: pd.Series
    :return: mean value of the first 20 elements of the input vector
    """
    return np.mean(x.iloc[0:20])


def last(x: pd.Series) -> float:
    """ Function returning the last value of the input vector

    :param x: input vector
    :type x: pd.Series
    :return: last value of the input vector
    """
    return x.iloc[-1]


def last20mean(x: pd.Series) -> np.ndarray:
    """ Function returning the mean value of the last 20 elements of the input vector

    :param x: input vector
    :type x: pd.Series
    :return: mean value of the last 20 elements of the input vector
    """
    return np.mean(x.iloc[-20:-1])


def tau_charge(df):
    """ Function returning the time constant of an exponential decay computed with the charge approach

    :param x: input vector
    :type x: np.array
    :return: time constant of an exponential decay computed with the charge approach
    """
    return calculate_integral_of_exp_decay(df, exponent=1)


def tau_energy(df: pd.DataFrame) -> float:
    """ Function returning the time constant of an exponential decay computed with the energy approach

    :param x: input vector
    :type x: np.array
    :return: time constant of an exponential decay computed with the energy approach
    """
    return calculate_integral_of_exp_decay(df, exponent=2)


def calculate_integral_of_exp_decay(exp_decay_in_sec: pd.DataFrame, exponent: int) -> float:
    """ Function calculating a characteristic time constant

    :param exp_decay_in_sec: input dataframe with index in seconds
    :type exp_decay_in_sec: pd.DataFrame
    :param exponent: exponent for integral calculation (either 1 or 2)
    :type exponent: int
    :return: integral of an exponential decay
    """
    exp_decay_in_sec_pow = exp_decay_in_sec.pow(exponent)
    return exponent * np.trapz(exp_decay_in_sec_pow, x=exp_decay_in_sec_pow.index, axis=0) / (
            exp_decay_in_sec_pow.iloc[0] - exp_decay_in_sec_pow.iloc[-1])


def tau_lin_reg(df: pd.DataFrame) -> float:
    """ Function returning the time constant of an exponential decay computed with linear fit to a logarithm of an
    exponential decay.

    :param df: input dataframe
    :type df: pd.DataFrame
    :return: characteristic time constant
    """
    right_exp = lambda p, x: -p[1] * x + p[0]
    err = lambda p, x, y: (right_exp(p, x) - y)
    time = df.index
    rate = np.log(df.values)
    v0 = [0., time.min()]
    out = leastsq(err, v0, args=(time, rate))
    v = out[0]  # fit parameters out []
    time_constant = 1 / v[1]
    return time_constant


def tau_exp_fit(df: pd.DataFrame) -> float:
    """ Function returning the time constant of an exponential decay computed with an exponential function fit.

    :param df: input dataframe
    :type df: pd.DataFrame
    :return: characteristic time constant
    """
    right_exp = lambda p, x: p[0] * np.exp(-p[2] * (x - p[1]))
    err = lambda p, x, y: (right_exp(p, x) - y)
    time = df.index
    rate = df.values
    v0 = [0., 0., 0.]
    out = leastsq(err, v0, args=(time, rate))
    v = out[0]  # fit parameters out []
    time_constant = 1 / v[2]
    return time_constant


def max_abs_min(df: pd.DataFrame) -> float:
    """ Function calculating the maximum absolute value of the signal.
    Let min_df and max_df be the global minimum and maximum value of the input dataframe.
    Then, the returned value is min_df if |min_df| > max_df, otherwise the reutrned value is max_df.
    As a result, should the absolute value of the minimum current be greater than the maximum value, then the
    minimum value with the negative sign is returned.

    :param df: input signal
    :type df: pd.DataFrame

    :return: maximum absolute value of the input signal
    """
    return df.max() if df.max() > abs(df.min()) else df.min()


def miits(i_meas_df: pd.DataFrame, no_digits=3) -> float:
    """ Method computing MIITs - time integral of a square of current divided by 1e6 in order to reduce the exponent

    :param i_meas_df: input dataframe with current
    :type i_meas_df: pd.DataFrame
    :param no_digits: number of digits after the decimal point for rounding
    :type no_digits: int
    :return: value of MIITs for the input current dataframe
    """
    miits = np.trapz(i_meas_df.values.reshape(len(i_meas_df.index)) ** 2, i_meas_df.index) * 1e-6
    no_digits = no_digits if miits > 1 else 5
    return round(miits, no_digits)


class Feature(object):
    """ Fundamental class built with FeatureBuilder.
    Its instances contain information about signals for which the features are computed. It has also methods to
    - take a subsignal window,
    - synchronize time,
    - subtract minimum value,
    - calculate features,
    - calculate multi-column features.

    """
    __name_to_function = {"first": first, "first20mean": first20mean, "last": last, "last20mean": last20mean,
                          "max": max, "min": min, "median": np.median, "std": np.std, "mean": np.mean, 'sum': sum,
                          "tau_charge": tau_charge, "tau_energy": tau_energy, "tau_lin_reg": tau_lin_reg,
                          "tau_exp_fit": tau_exp_fit, 'max_abs_min': max_abs_min, 'miits': miits}

    def set_dfs(self, dfs: List[pd.DataFrame]):
        """ Methods setting signals for which the features are calculated

        :param dfs: list of signals for which the features are calculated
        :return: None as it only sets the dfs
        """
        dfs = dfs if isinstance(dfs, (list, tuple)) else [dfs]
        self.dfs = flatten_nested_lists(dfs)

    def take_signal_window(self, t_start: float, t_end: float):
        """ Method updating signals with windows between t_start and t_end

        :param t_start: start time for a subwindow
        :type t_start: float
        :param t_end: end time for a subwindow
        :type t_end: float
        :return: None as it only updates the dfs
        """
        def get_window(df, t_start: float, t_end: float):
            """ Method taking a signal window between t_start and t_end

            :param t_start: start time for a subwindow
            :type t_start: float
            :param t_end: end time for a subwindow
            :type t_end: float
            :return: None as it only updates the dfs
            """
            return df[(df.index > t_start) & (df.index < t_end)]

        self.dfs = [get_window(df, t_start, t_end) for df in self.dfs]

    def synchronize_time(self, t_sync: float):
        """ Method updating signals with indices synchronized to t_sync

        :param t_sync: start time for synchronization
        :type t_sync: float
        :return: None as it only updates the dfs
        """
        self.dfs = [SignalIndexConversion.synchronize_df(df, t_sync) if not df.empty else df for df in self.dfs]

    def subtract_min_value(self):
        """ Method updating signals with values subtracted with its minimum value

        :return: None as it only updates the dfs
        """
        self.dfs = [df - df.min() if not df.empty else df for df in self.dfs]

    def calculate_features(self, features: Union[str, List[str]], index=0) -> pd.DataFrame:
        """ Method calculating features for the dataframes

        :param features: either a single feature name or a list of features
        :param index: for the output row with features
        :type index: int
        :return: a single-row DataFrame with computed features
        """
        features = features if isinstance(features, list) else [features]
        functions_to_apply = [self.__name_to_function[x] for x in features]

        features_dfs = []
        for df in self.dfs:
            if not df.empty:
                features_df = df.aggregate(functions_to_apply).T
                feature_col_names = {col: '{}:{}'.format(features_df.index[0], col) for col in features_df.columns}
                features_df = features_df.rename(columns=feature_col_names)
                features_df.index = [index]
                features_dfs.append(features_df)

        return pd.concat(features_dfs, axis=1)

    def calculate_multicolumn_features(self, features: Union[str, List[str]], prefix: str):
        """ Method calculating features for a multicolumn signal

        :param features: a single feature or a list of fetures to compute
        :param prefix: prefix for all columns
        :return: computed dataframe with multi-column features
        """
        features = features if isinstance(features, list) else [features]
        functions_to_apply = [self.__name_to_function[x] for x in features]
        df_out = self.dfs[0].agg(functions_to_apply).T
        df_out.columns = ['%s_%s' % (prefix, col) for col in df_out.columns]

        return df_out


class FeatureBuilder(object):
    """ Builder class for the Feature objects.
    It works either with signals or multi-column dataframes.

    """

    def __init__(self):
        self.feature = Feature()

    def with_signal(self, *dfs):
        """ Method setting a signal / a list of signals for feature calculation

        :param dfs: either a single signal or a list of signals to store in the FeatureBuilderSignal object
        :return: a FeatureBuilderSignal object
        """
        self.feature.set_dfs(dfs)

        return FeatureBuilderSignal(self.feature)

    def with_multicolumn_signal(self, *dfs):
        """ Method setting a multi-column signal / a list of signals for feature calculation

        :param dfs: either a multi-column single signal or a list of signals to store
        in the FeatureBuilderMultiColumnSignal object
        :return: a FeatureBuilderMultiColumnSignal object
        """
        self.feature.set_dfs(dfs)

        return FeatureBuilderMultiColumnSignal(self.feature)


class FeatureBuilderSignal(object):
    """ Class for feature calculation with signal(s) already set

    """

    def __init__(self, feature):
        self.feature = feature

    def with_duration(self, t_start: float, t_end: float):
        """ Method updating signals with windows between t_start and t_end

        :param t_start: start time for a subwindow
        :type t_start: float
        :param t_end: end time for a subwindow
        :type t_end: float
        :return: an updated FeatureBuilderSignal object
        """
        self.feature.take_signal_window(t_start, t_end)

        return self

    def synchronize_time(self, t_sync: float):
        """ Method updating signals with indices synchronized to t_sync

        :param t_sync: start time for synchronization
        :type t_sync: float
        :return: an updated FeatureBuilderSignal object
        """
        self.feature.synchronize_time(t_sync)

        return self

    def subtract_min_value(self):
        """ Method updating signals with values subtracted with its minimum value

        :return: an updated FeatureBuilderSignal object
        """
        self.feature.subtract_min_value()

        return self

    def calculate_features(self, features: Union[str, List[str]], index=0):
        """ Method calculating features for the signals

        :param features: either a single feature name or a list of features
        :param index: for the output row with features
        :type index: int
        :return: a single-row DataFrame with computed features
        """
        return self.feature.calculate_features(features, index)


class FeatureBuilderMultiColumnSignal(object):
    """ Class for feature calculation with multicolumn signal already set

    """

    def __init__(self, feature):
        self.feature = feature

    def calculate_features(self, features: Union[str, List[str]], prefix: str):
        """ Method calculating features for a multicolumn signal

        :param features: a single feature or a list of fetures to compute
        :param prefix: prefix for all columns
        :return: an updated FeatureBuilderMultiColumnSignal object
        """
        self.dfs = self.feature.calculate_multicolumn_features(features, prefix)
        return self

    def convert_into_row(self, index=0):
        """ Method converting a table-shape feature dataframe into a row

        param index: index for output row DataFrame
        :type index: int
        :return: an updated FeatureBuilderMultiColumnSignal object
        """
        self.dfs = convert_table_df_into_row(self.dfs, index)

        return self
