from typing import Dict, Union, Tuple, List
import warnings
import datetime

import numpy as np
import pandas as pd

from lhcsmapi.Time import Time
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.FeatureBuilder import convert_table_df_into_row
from lhcsmapi.pyedsl.dbquery.DbQueryBuilder import DbQueryBuilderFactory
from lhcsmapi.pyedsl.dbsignal.DbSignal import validate_key_in_kwargs
from lhcsmapi.pyedsl.dbsignal.SignalIndexConversion import SignalIndexConversion


def validate_time_definition(time_def_dct: Dict[str, Union[int, str, float]]) -> Tuple[int, int]:
    """ Function validating time definition given in three ways:
    - t_start, t_end
    - t_start, time range definition (time after t_start)
    - t_start, time range definition (time before and after t_start)

    :param time_def_dct: a dictionary with time definition
    :return: a tuple of (t_start, t_end) based on the time definition
    """

    # to correct for an error with checking type of a np.int64 variable by isinstance()
    if 't_start' in time_def_dct and isinstance(time_def_dct['t_start'], np.int64):
        time_def_dct['t_start'] = int(time_def_dct['t_start'])

    if 't_end' in time_def_dct and isinstance(time_def_dct['t_end'], np.int64):
        time_def_dct['t_end'] = int(time_def_dct['t_end'])

    t_start = validate_key_in_kwargs('t_start', int, pd.Timestamp, str, float, datetime, tuple, np.datetime64, np.int64,
                                     **time_def_dct) or 0
    t_end = validate_key_in_kwargs('t_end', int, pd.Timestamp, str, float, datetime, tuple, np.datetime64, np.int64,
                                   **time_def_dct)
    duration = validate_key_in_kwargs('duration', tuple, list, **time_def_dct)
    return Time.get_query_period_in_unix_time(t_start, t_end, duration, '')


class Query(object):
    """ Fundamental class built with the query builder.

    """

    def set_db(self, db: str) -> None:
        """ Method setting database name (PM, NXCALS, AFT)

        :param db: database name
        :type db: str
        :return: None
        """
        self.db = db

    def set_circuit_type(self, circuit_type: str) -> None:
        """ Method setting circuit type (e.g., RB)

        :param circuit_type: circuit type
        :type circuit_type: str
        :return: None
        """
        self.circuit_type = circuit_type

    def set_t_start(self, t_start: int) -> None:
        """ Method setting start time for a query

        :param t_start: start time
        :type t_start: int
        :return: None
        """
        self.t_start = t_start

    def set_t_end(self, t_end: int) -> None:
        """ Method setting end time for a query

        :param t_end: end time
        :type t_end: int
        :return: None
        """
        self.t_end = t_end

    def set_timestamp(self, timestamp: int) -> None:
        """ Method setting a timestamp for a query

        :param timestamp: timestamp for a query
        :type timestamp: int
        :return: None
        """
        self.timestamp = timestamp

    def set_dbconnector(self, dbconnector) -> None:
        """ Method setting a dbconnector for a query (NXCALS and AFT)

        :param dbconnector: dbconnector for either NXCALS or AFT
        :return: None
        """
        self.dbconnector = dbconnector

    def get_time_params(self) -> Dict[str, int]:
        """ Method returning time parameters as a dictionary
        - timestamp for PM signal query
        - t_start, t_end otherwise

        :return: a dictionary with time definition (either a t_start and t_end or timestamp)
        """
        if hasattr(self, 'timestamp') and self.db == 'PM':
            return {'eventTime': self.timestamp}
        elif hasattr(self, 't_start') and self.db in ['NXCALS', 'PM', 'AFT']:
            return {'t_start': self.t_start, 't_end': self.t_end}
        else:
            raise ValueError('Wrong time definition!')


class QueryBuilder(object):
    """ Builder class to query databases (PM, CALS, NXCALS, AFT).
        The general syntax for PM (NX)CALS is
        QueryBuilder().{with_pm()/with_aft()/with_nxcals()}.with_circuit_type().{with_duration()/with_timestamp()}.with_metadata().{signal_query()/feature_query()}

        For AFT the metadata is not provided which simplifies the syntax to
        QueryBuilder().with_aft().context_query()
        QueryBuilder().with_aft().with_duration().fault_query()

        - EXAMPLE PM (event query)::
        >>> from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
        >>> t_start = '2015-01-13 16:59:11+01:00'
        >>> t_end = '2015-01-13 17:15:46+01:00'

        >>> source_timestamp_df = QueryBuilder().with_pm()\
            .with_duration(t_start=t_start, t_end=t_end)\
            .with_circuit_type('RB')\
            .with_metadata(circuit_name='RB.A12', system='PC').event_query().df # doctest: +SKIP

        >>> print(source_timestamp_df) # doctest: +SKIP
                     source            timestamp
        0  RPTE.UA23.RB.A12  1421164827600000000

        - EXAMPLE PM (signal query)::
        >>> from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
        >>> timestamp_sync = 1421164827600000000

        >>> i_meas_df = QueryBuilder().with_pm().with_circuit_type('RB').with_timestamp(timestamp_sync)\
                .with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').signal_query()\
                .synchronize_time(timestamp_sync).convert_index_to_sec().dfs[0] # doctest: +SKIP

        >>> print(i_meas_df.head()) # doctest: +SKIP
                 STATUS.I_MEAS
        -359.98      -0.150693
        -359.96      -0.152328
        -359.94      -0.151348
        -359.92      -0.150366
        -359.90      -0.149385

        - Example NXCALS (assuming connection to NXCALS cluster)::
        >>> from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
        >>> t_start = '2015-01-13 16:59:11+01:00'
        >>> t_end = '2015-01-13 17:15:46+01:00'

        >>> i_meas_df = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=t_start, t_end=t_end) \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').signal_query() \
            .synchronize_time().convert_index_to_sec().dfs[0] # doctest: +SKIP

        print(i_meas_df.head()) # doctest: +SKIP
              I_MEAS
        0.0     0.07
        25.5    1.13
        26.0    1.72
        26.5    2.82
        27.0    4.32
        """

    _msg = 'Set database name using with_pm(), with_nxcals(spark), with_aft(session) methods.'

    def __init__(self):
        self.query = Query()

    def __getattr__(self, name) -> None:
        """ Method checking if query syntax is correct and raises a meaningful error if the syntax is violated

        :param name: name of field/method called for a QueryBuilder instance
        :return: None
        """
        if name not in ['query', 'with_pm', 'with_aft', 'with_nxcals']:
            raise AttributeError('{}() is not supported. {}'.format(name, self._msg))

    def __repr__(self):
        """ Method displaying a hint for available next step

        :return: message with a hint for available options
        """
        return self._msg

    def with_pm(self):
        """ Method setting PM as the database for a query

        :return: an instance of QueryBuilderWithDb with selected db
        """
        self.query.set_db('PM')
        self.query.dbconnector = None
        return QueryBuilderWithDb(self.query)

    def with_nxcals(self, dbconnector):
        """ Method setting NXCALS as the database for a query with a spark connector

        :param dbconnector: spark connector

        :return: an instance of QueryBuilderWithDb with selected db
        """
        self.query.set_db('NXCALS')
        self.query.set_dbconnector(dbconnector)
        return QueryBuilderWithDb(self.query)

    def with_aft(self, dbconnector):
        """ Method setting AFT as the database for a query with a spark connector

        :param dbconnector: AFT connector

        :return: an instance of QueryBuilderWithDb with selected db
        """
        self.query.set_db('AFT')
        self.query.set_dbconnector(dbconnector)
        return QueryBuilderWithDb(self.query)


class QueryBuilderWithDb(object):
    """ Class stores information about a selected database and allows for setting query time definition.
    There are checks for proper use of the language grammar.

    """
    _msg = 'Set time definition: for PM signal query, with_timestamp(),' \
           '\n for PM event query or AFT, NXCALS signal query, with_duration()'

    def __init__(self, query):
        self.query = query

    def __getattr__(self, name):
        """ Method checking if query syntax is correct and raises a meaningful error if the syntax is violated

        :param name: name of field/method called for a QueryBuilderWithDb instance
        :return: None
        """
        if name not in ['query', 'with_timestamp', 'with_duration']:
            raise AttributeError('{}() is not supported. {}'.format(name, self._msg))

    def __repr__(self):
        """ Method displaying a hint for available next steps

        :return: message with a hint for available options
        """
        return 'Database name properly set to {}. {}'.format(self.query.db, self._msg)

    def with_timestamp(self, timestamp):
        """ Method setting timestamp and checking if it is for query with db set to pm

        :param timestamp: timestamp, typically unix timestamp; for clarity, converted to unix timestamp anyhow
        :return: an instance of QueryBuilderWithDbTimeDef
        """
        if self.query.db == 'PM':
            self.query.set_timestamp(Time.to_unix_timestamp(timestamp))
            return QueryBuilderWithDbTimeDef(self.query)
        else:
            raise AttributeError('Timestamp can be only set for PM database, got {}'.format(self.query.db))

    def with_duration(self, **duration):
        """ Method setting time definition based on input keyword-argument

        :param duration: a keyword-argument time definition (t_start, t_end) or (t_start, duration)
        :return: an instance of QueryBuilderWithDbTimeDef
        """
        t_start, t_end = validate_time_definition(duration)
        t_start, t_end = Time.get_query_period_in_unix_time(t_start, t_end, duration, '')

        self.query.set_t_start(t_start)
        self.query.set_t_end(t_end)

        return QueryBuilderWithDbTimeDef(self.query)

    def context_query(self, context, context_id=None, **kwargs):
        """ Method querying context with AFT

        :param context: query context, e.g., 'accelerators/{acceleratorId}/systems'
        :param context_id: context id, e.g., 'LHC'
        :param kwargs: an optional list of keyword-arguments
        :return: a ContextProcessing object
        """
        if self.query.db is not 'AFT':
            raise NotImplementedError('This method is not implemented for this database %s' % self.query.db)

        is_timestamp = False
        is_t_start = True
        self.query.db_query_builder = DbQueryBuilderFactory.create('AFT', is_timestamp, is_t_start, None)
        time_params = {}
        kwargs['dbconnector'] = self.query.dbconnector
        kwargs['context_id'] = context_id
        return self.query.db_query_builder.execute_context_query(time_params, context, **kwargs)


class QueryBuilderWithDbTimeDef(object):
    """ Class stores information about a selected database and time definition allows for:
    - setting a circuit type (for metadata-based queries)
    - providing query parameters (for general purpose queries)
    - querying faults with AFT
    There are checks for proper use of the language grammar.

    """
    _msg = 'Set generic query parameter using with_query_parameters() method, ' \
           'or a circuit signal using with_circuit_type() method.'

    def __init__(self, query):
        self.query = query

    def __getattr__(self, name):
        """ Method checking if query syntax is correct and raises a meaningful error if the syntax is violated

        :param name: name of field/method called for a QueryBuilderWithDb instance
        :return: None
        """
        if name not in ['query', 'with_circuit_type', 'with_query_parameters']:
            raise AttributeError('{}() is not supported. \n{}'.format(name, self._msg))

    def __repr__(self):
        """ Method displaying time definition and a hint for available next steps

        :return: message with a hint for available options
        """
        if hasattr(self.query, 'timestamp'):
            return 'Timestamp set to timestamp={}. {}'.format(self.query.timestamp, self._msg)
        else:
            return 'Query duration set to t_start={}, t_end={}. {}' \
                .format(self.query.t_start, self.query.t_end, self._msg)

    def with_circuit_type(self, circuit_type: str):
        """ Method setting circuit type for further circuit-oriented metadata definition

        :param circuit_type: name of a circuit type
        :type circuit_type: str
        :return: a QueryBuilderWithDbTimeDefCircuitType object
        """
        if circuit_type not in SignalMetadata.get_circuit_types():
            raise ValueError('Circuit type %s not supported!' % circuit_type)

        self.query.set_circuit_type(circuit_type.upper())
        return QueryBuilderWithDbTimeDefCircuitType(self.query)

    def with_query_parameters(self, **kwargs):
        """ Method setting custom query parameters.
        User is responsible for validity check as no checks are performed.

        :param kwargs: a list of keyword-arguments for a database query
        :type kwargs: Dict
        :return: a QueryBuilderWithDbTimeDefCircuitType object
        """
        db = self.query.db
        is_timestamp = hasattr(self.query, 'timestamp')
        is_t_start = hasattr(self.query, 't_start')

        self.query.db_query_builder = DbQueryBuilderFactory.create(db, is_timestamp, is_t_start)
        self.query.db_query_builder.create_query_parameters(kwargs)

        return QueryBuilderWithDbTimeDefMetadata(self.query)

    def fault_query(self, **kwargs) -> pd.DataFrame:
        """ Method performing fault query for AFT and a givent time duration

        :param kwargs: a dictionary of keyword-arguments, e.g., acceleratorId='LHC', accessNeeded=True
        :return: pd.Dataframe with fault information
        """
        if self.query.db is not 'AFT':
            raise NotImplementedError('This method is not implemented for this database %s' % self.query.db)

        is_timestamp = hasattr(self.query, 'timestamp')
        is_t_start = hasattr(self.query, 't_start')

        self.query.db_query_builder = DbQueryBuilderFactory.create('AFT', is_timestamp, is_t_start, None)
        time_params = self.query.get_time_params()
        kwargs['dbconnector'] = self.query.dbconnector
        return self.query.db_query_builder.execute_fault_query(time_params_dict=time_params, **kwargs)


class QueryBuilderWithDbTimeDefCircuitType(object):
    """ Class stores information about a selected database, time definition, and circuit type allows for:
    - setting metadata (for metadata-based queries)
    There are checks for proper use of the language grammar.

    """
    _msg = 'Set metadata using with_metadata() method, e.g., ' \
           'with_metadata(circuit_name=\'RB.A12\', system=\'PC\', signal=\'I_MEAS\').'

    def __init__(self, query):
        self.query = query

    def __getattr__(self, name):
        """ Method checking if query syntax is correct and raises a meaningful error if the syntax is violated

        :param name: name of field/method called for a QueryBuilderWithDbTimeDefCircuitType instance
        :return: None
        """
        if name not in ['query', 'with_metadata']:
            raise AttributeError('{}() not supported. \n{}'.format(name, self._msg))

    def __repr__(self):
        """ Method displaying time definition and a hint for available next steps

        :return: message with a hint for available options
        """
        return 'Circuit type set to {}. \n{}'.format(self.query.circuit_type, self._msg)

    def with_metadata(self, **kwargs):
        """ Method setting metadata for a query in a convenient way based on:
        - circuit name
        - system
        - signal (for signal and feature query)
        - wildcard (for signal, feature, and event query)

        :param kwargs: keyword arguments for selection of metadata based on wildcards
        :return: an updated QueryBuilderWithDbTimeDefMetadata object
        """
        db = self.query.db
        is_timestamp = hasattr(self.query, 'timestamp')
        is_t_start = hasattr(self.query, 't_start')
        circuit_type = self.query.circuit_type

        # choose metadata according to the start of a query
        timestamp_query = self.query.t_start if is_t_start else self.query.timestamp
        self.query.db_query_builder = DbQueryBuilderFactory.create(db, is_timestamp, is_t_start, circuit_type)
        self.query.db_query_builder.create_query_metadata(kwargs, timestamp_query)
        self.query.db_query_builder.update_query_metadata(kwargs)

        return QueryBuilderWithDbTimeDefMetadata(self.query)


class QueryBuilderWithDbTimeDefMetadata(object):
    """ Class stores information about a selected database, time definition, circuit type, and metadata allows for:
    - signal query
    - event query
    - feature query
    - context query
    There are checks for proper use of the language grammar.

    """

    def __init__(self, query):
        self.query = query

    def _check_language_consistency(self, name):
        """ Method checking language consistency
        There are several, syntactically correct options:
        - For PM:
          - QueryBuilder().with_pm().with_timestamp().signal_query()
          - QueryBuilder().with_pm().with_duration().event_query()
          - QueryBuilder().with_pm().with_duration().signal_query() - for BLM signals
          - QueryBuilder().with_pm().with_duration().context_query() - for BLM signals

        - For NXCALS:
          - QueryBuilder().with_nxcals().with_duration().signal_query()
          - QueryBuilder().with_nxcals().with_duration().feature_query()

        :param name:
        :return:
        """
        if hasattr(self.query, 'timestamp') and (self.query.db == 'PM') \
                and name not in ['query', 'signal_query']:
            raise AttributeError('{}() not supported, for PM with timestamp you can call signal_query()'.format(name))
        elif hasattr(self.query, 't_start') and (self.query.db == 'PM') \
                and name not in ['query', 'event_query', 'signal_query', 'context_query']:
            raise AttributeError('{}() not supported, for PM with duration you can call '
                                 'event_query(), signal_query(), context_query(); the former two for BLM'.format(name))
        elif hasattr(self.query, 't_start') and (self.query.db == 'NXCALS') \
                and name not in ['query', 'signal_query', 'feature_query']:
            raise AttributeError('{}() not supported, for NXCALS with duration you can call '
                                 'signal_query(), feature_query()'.format(name))

    def __getattr__(self, name):
        """ Method checking if query syntax is correct and raises a meaningful error if the syntax is violated

        :param name: name of field/method called for a QueryBuilderWithDbTimeDefCircuitType instance
        :return: None
        """
        self._check_language_consistency(name)

    def __repr__(self):
        """ Method displaying time definition and a hint for available next steps

        :return: message with a hint for available options
        """
        if hasattr(self.query, 'timestamp') and self.query.db == 'PM':
            return 'Query PM signal using signal_query() method.'
        elif hasattr(self.query, 't_start') and self.query.db == 'PM':
            return 'Query PM event using event_query() method.\n' \
                   'Query PM BLM signal using signal_query() method.\n' \
                   'Query PM BLM context using context_query() method.'
        elif hasattr(self.query, 't_start') and self.query.db == 'NXCALS':
            return 'Query NXCALS signals using signal_query() method.\n' \
                   'Query NXCALS signal features using feature_query() method.'

    def signal_query(self, allow_duplicate_index=False, **kwargs):
        """ Method for executing a signal query for PM and NXCALS

        :param allow_duplicate_index: boolean flag for PM signal query of FGC (there are duplicated indices for FGC
        trigger due to a limited resolution)
        :param kwargs: an optional keyword-argument dictionary to further parametrize a query
        :return: a SignalProcessing object

        - EXAMPLE PM (signal query)::
        >>> from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
        >>> timestamp_sync = 1421164827600000000

        >>> i_meas_df = QueryBuilder().with_pm().with_circuit_type('RB').with_timestamp(timestamp_sync)\
                .with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').signal_query()\
                .synchronize_time(timestamp_sync).convert_index_to_sec().dfs[0] # doctest: +SKIP

        >>> print(i_meas_df.head()) # doctest: +SKIP
                 STATUS.I_MEAS
        -359.98      -0.150693
        -359.96      -0.152328
        -359.94      -0.151348
        -359.92      -0.150366
        -359.90      -0.149385

        - Example NXCALS (assuming connection to NXCALS cluster)::
        >>> from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
        >>> t_start = '2015-01-13 16:59:11+01:00'
        >>> t_end = '2015-01-13 17:15:46+01:00'

        >>> i_meas_df = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=t_start, t_end=t_end) \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').signal_query() \
            .synchronize_time().convert_index_to_sec().dfs[0] # doctest: +SKIP

        print(i_meas_df.head()) # doctest: +SKIP
              I_MEAS
        0.0     0.07
        25.5    1.13
        26.0    1.72
        26.5    2.82
        27.0    4.32
        """

        # language consistency check:
        self._check_language_consistency('signal_query')

        time_params = self.query.get_time_params()
        circuit_dfs = self.query.db_query_builder\
            .execute_signal_query(time_params, dbconnector=self.query.dbconnector,
                                  allow_duplicate_index=allow_duplicate_index, **kwargs)

        return SignalProcessing(circuit_dfs, self.query.db)

    def event_query(self, **kwargs):
        """ Method querying PM events

        :param kwargs: an optional keyword-argument dictionary to further parametrize a query
        :return: an EventProcessing object

        - EXAMPLE PM (event query)::
        >>> from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
        >>> t_start = '2015-01-13 16:59:11+01:00'
        >>> t_end = '2015-01-13 17:15:46+01:00'

        >>> source_timestamp_df = QueryBuilder().with_pm()\
            .with_duration(t_start=t_start, t_end=t_end)\
            .with_circuit_type('RB')\
            .with_metadata(circuit_name='RB.A12', system='PC').event_query().df # doctest: +SKIP

        >>> print(source_timestamp_df) # doctest: +SKIP
                     source            timestamp
        0  RPTE.UA23.RB.A12  1421164827600000000
        """
        # language consistency check:
        self._check_language_consistency('event_query')

        time_params = self.query.get_time_params()
        event_source_timestamp_df = self.query.db_query_builder.execute_event_query(time_params, **kwargs)

        return EventProcessing(event_source_timestamp_df)

    def feature_query(self, features, **kwargs):
        """ Method querying NXCALS signal features

        :param features: list of features to compute on the NXCALS cluster
        :param kwargs: an optional keyword-argument dictionary to further parametrize a query
        :return: an FeatureProcessing object

        - Example::
        >>> from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder

        >>> i_meas_df = QueryBuilder().with_nxcals(spark) \
                .with_duration(t_start=1526898157236000000, t_end=1526903352338000000) \
                .with_query_parameters(nxcals_system='CMW', nxcals_device='RPTE.UA23.RB.A12', nxcals_property='SUB', signal='I_MEAS') \
                .feature_query(['mean', 'std', 'max', 'min', 'count']).df  # doctest: +SKIP

        """

        # language consistency check:
        self._check_language_consistency('feature_query')

        time_params = self.query.get_time_params()

        return FeatureProcessing(self.query.db_query_builder.execute_feature_query(time_params, features, **kwargs,
                                                                                   dbconnector=self.query.dbconnector))

    def context_query(self, contexts: List[str], **kwargs):
        """ Method executing context query for BLM PM

        :param contexts: context variables to query
        :param kwargs: optional keyword-arguments for a query
        :return: an instance of ContextProcessing class

        - Example::
        >>> from lhcsmapi.Time import Time
        >>> from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
        >>> start_time_date = '2015-11-23 07:28:53+01:00'
        >>> t_start, t_end = Time.get_query_period_in_unix_time(start_time_date=start_time_date, duration_date=[(1, 's'), (2, 's')])

        >>> blm_names_df = QueryBuilder().with_pm() \
                .with_duration(t_start=t_start, duration=[(2, 's')]) \
                .with_query_parameters(system='BLM', className='BLMLHC', source='HC.BLM.SR6.C') \
                .context_query(contexts=["blmNames"]).df # doctest: +SKIP
        """
        # language consistency check:
        self._check_language_consistency('context_query')

        time_params = self.query.get_time_params()
        context_df = self.query.db_query_builder.execute_context_query(time_params, contexts, **kwargs)

        return ContextProcessing(context_df)


class ContextProcessing(object):
    """ Placeholder class for processing of results of context queries

    """
    def __init__(self, df):
        self.df = df


class FeatureProcessing(object):
    """  Class for processing of results of feature queries

    """
    def __init__(self, df):
        self.df = df

    def sort_busbar_location(self, circuit_type: str, circuit_name: str, split=':'):
        """ Method sorting busbar location for RB and RQ circuits according to their physical location in the
        LHC tunnel.

        :param circuit_type: circuit type (RB and RQ supported)
        :param circuit_name: name of a circuit, e.g., RB.A12
        :param split: character on which signal name is split to recover busbar name for sorting
        :return: an updated FeatureProcessing object
        """
        if circuit_type not in ['RB', 'RQ']:
            return self

        if 'nxcals_variable_name' in self.df.columns:
            circuit_name = SignalMetadata.get_circuit_names(circuit_type) if circuit_type == 'RB' else circuit_name
            self.df['name_prefix'] = self.df.apply(lambda row: row['nxcals_variable_name'].split(split)[0], axis=1)
            busbars_ordered_df = MappingMetadata.read_busbar_to_magnet_mapping(circuit_type)
            circuit_name = circuit_name if isinstance(circuit_name, list) else [circuit_name]
            is_busbar_in_circuit_name = busbars_ordered_df['Circuit'].isin(circuit_name)
            busbars_ordered_per_circuit_df = busbars_ordered_df[is_busbar_in_circuit_name]
            self.df['name_prefix'] = pd.Categorical(self.df['name_prefix'],
                                                    categories=busbars_ordered_per_circuit_df['Bus Bar Segment Name'],
                                                    ordered=True)
            self.df.sort_values(by='name_prefix', inplace=True)
            self.df.drop(columns='name_prefix', inplace=True)
            self.df.reset_index(inplace=True, drop=True)
        else:
            warnings.warn('This operation is only supported to feature DataFrame with nxcals_variable_name column,'
                          ' got {}'.format(self.df.columns))

        return self

    def correct_voltage_sign(self, variable_name='nxcals_variable_name',
                             class_compare=0, correction_col='mean', polarity=-1):
        """ Method providing alteration of polarity to groups of data, by comparing a selected columns
        value, from a selected class to a chosen expected polarity. E.g. Comparing the value of the 'mean' column for
        a Class representing inductive voltage at ramp, to a polarity of -1, will result in swapped polarity for all
        related classes (including selected class) if statement yields true

        Example of Usages:
            - Create default check and correction: Checks class == 0, column that is being checked and
            corrected is 'mean', and we are checking for negative polarity.
            feature_query takes in function 'translate_udf' that creates the 'class'-column needed to complete
            correction of voltage sign

            >>> df = QueryBuilder().with_nxcals(spark) \
                .with_duration(t_start=t_start_injs, t_end=t_end_sbs) \
                .with_circuit_type(circuit_type) \
                .with_metadata(circuit_name=['RB.A12', 'RB.A45'], system='BUSBAR', signal='U_MAG', wildcard={'BUSBAR': ['DCBB.8L2.R', 'DCBA.15R4.R']}) \
                .feature_query(['mean'], function=translate_udf) \
                .correct_voltage_sign() \
                .df # doctest: +SKIP

            >>> print(df) # doctest: +SKIP
            nxcals_variable_name 	class 	mean
            DCBA.15R4.R:U_MAG 	2 	-0.001142
            DCBA.15R4.R:U_MAG 	0 	0.632161
            DCBA.15R4.R:U_MAG 	1 	-0.001239
            DCBB.8L2.R:U_MAG 	1 	0.001861
            DCBB.8L2.R:U_MAG 	0 	0.636315
            DCBB.8L2.R:U_MAG 	2 	0.001983


        Assumes:
            - feature_query includes function to classify, and add, classes to dataframe
            - For each variable_name there is only one of each class type
            - Data from selected correction_col contains numbers value
            - There are groups of data with same variable name


        :param variable_name: Name of column used to group data points together for value comparison. Default: nxcals_variable_name
        :type variable_name: str

        :param class_compare: Class checked for polarity, action performed if polarity chosen equals True. Default: Class == 0
        :type class_compare: str or num

        :param correction_col: Chosen value for polarity change. Default: mean
        :type correction_col: str

        :param polarity: Check for selected polarity: Default: -1
        :type polarity: whole number, either -1 or 1

        :return: self.df with altered polarity to selected column
        """
        if (correction_col not in self.df) or ('class' not in self.df):
            warnings.warn('This operation is only supported to feature DataFrame with class column, and the selected '
                          '%s column. Got %s' % (correction_col, self.df.columns))

        def correction(df, class_cmpr, cor_col, pol, num_of_rows=3):
            if len(df) < num_of_rows:
                return df

            class_change = df['class'].unique()
            idx = df.index.unique()

            if (class_cmpr in class_change) & (df[cor_col].isnull().values.any() == False):
                if np.sign(df[df['class'] == class_cmpr][cor_col].iloc[0]) == pol:
                    for i in range(len(class_change)):
                        df = df.assign(mean=df[cor_col].replace({df[cor_col][idx[int(class_change[i])]]:
                                                                     df[cor_col][idx[int(class_change[i])]] * -1}))

            return df

        self.df = self.df.groupby(variable_name).apply(lambda row: correction(row, class_compare, correction_col, polarity))
        return self

    def calculate_max_abs(self):
        """ Method calculating maximum absolute value from 'min' and 'max' columns. The result is stored in 'max_abs'
        column. E.g. if min = -2 and max = 1, then max_abs = 2; if min = 0, max = 3, then max_abs = 3.

        :return: an updated FeatureProcessing object with an additional column 'max_abs' if 'max' and 'min' columns are
        present, otherwise the same object is returned
        """
        if ('min' in self.df.columns) and ('max' in self.df.columns):
            self.df['max_abs'] = self.df.apply(lambda row: max(row['max'], abs(row['min'])), axis=1)
        else:
            warnings.warn('This operation is only supported to feature DataFrame with min, max columns,'
                          ' got {}'.format(self.df.columns))
        return self

    def sort_values(self, by: str):
        """ Method sorting values by selected column name. In addition the index is reset.

        :param by: column by which the sorting is performed
        :return: an updated FeatureProcessing object with sorted DataFrame by selected column
        """
        self.df.sort_values(by=by, inplace=True)
        self.df.reset_index(inplace=True, drop=True)
        return self

    def convert_into_row(self, index=0):
        """ Method converting a columnar dataframe into a single row with an index.
        It is assumed that there exists two columns:
        - 'nxcals_variable_name' with feature names
        - another one with feature values

        :param index: index for the output row, 0 by default
        :return: an updated FeatureProcessing object with dataframe coverted into a row
        """
        # table format
        if 'nxcals_variable_name' in self.df.columns:
            self.df = self.df.set_index(self.df['nxcals_variable_name'], drop=True) \
                .drop(columns='nxcals_variable_name')
            # row format
            self.df = convert_table_df_into_row(self.df, index)
        else:
            warnings.warn('This operation is only supported to feature DataFrame with nxcals_variable_name column,'
                          ' got {}'.format(self.df.columns))
        return self


class EventProcessing(object):
    """ Class provides a set of frequently used methods for event processing. Each method always returns a reference
    to an object which allows for chaining of method calls and reduce the need for using local variables.

    """

    def __init__(self, df):
        self.df = df

    def filter_source(self, circuit_name: str, system: str):
        """ Method filtering out events for QH and QDS in RB and RQ that do not come from a given circuit, e.g., in case
        there are several circuits experiencing an FPA at the same time.

        :param circuit_name: name of a circuit with magnets to keep
        :param system: name of a system to consider
        :return: an updated EventProcessing object with events belonging to a given circuit name and system
        """
        if self.df.empty:
            return self

        if system in ['QH', 'QDS']:
            if 'RQ.' in circuit_name:
                self.df['is_circuit_ok'] = self.df.apply(lambda row: (SignalMetadata.get_sector_names_from_components(
                    'RQ', 'MAGNET', row['source'])[0] == circuit_name) |
                                                                     (SignalMetadata.get_sector_names_from_components(
                                                                         'RQ', 'MAGNET', row['source'])[
                                                                          1] == circuit_name), axis=1)
            elif 'RB.' in circuit_name:
                self.df['is_circuit_ok'] = self.df.apply(lambda row: (
                        SignalMetadata.get_sector_names_from_components('RB', 'MAGNET', row['source']) == circuit_name),
                                                         axis=1)
            else:
                return self

            self.df = self.df[self.df['is_circuit_ok']]
            self.df.drop(columns=['is_circuit_ok'], inplace=True)
            self.df.reset_index(drop=True, inplace=True)
        return self

    def drop_duplicate_source(self):
        """ Method drops rows that contain duplicate sources, e.g., when events come from redundant boards.

        :return: an updated EventProcessing object without duplicate sources
        """
        if self.df.empty:
            return self

        self.df = self.df.drop_duplicates(subset=['source']).sort_values(by=['timestamp']).reset_index(drop=True)
        return self

    def sort_values(self, by: str):
        """Method sorting values by selected column name. In addition the index is reset.

        :param by: column by which the sorting is performed
        :return: an updated FeatureProcessing object with sorted DataFrame by selected column
        """
        if self.df.empty:
            return self

        self.df.sort_values(by=by, inplace=True)
        self.df.reset_index(inplace=True, drop=True)
        return self


class SignalProcessing(object):
    """ Class provides a set of frequently used methods for signal processing. Each method always returns a reference
    to an object which allows for chaining of method calls and reduce the need for using local variables.

    """

    def __init__(self, dfs, db):
        self.dfs = dfs
        self.db = db

    def synchronize_time(self, t0=None):
        """ Method synchronizing time to an input initial timestamp. Method assumes that the time unit of indices and
        t0 are the same (e.g., ns and ns).

        :param t0: initial timestamp for synchronization
        :return: an updated SignalProcessing object with all signals synchronized to the initial timestamp
        """
        self.dfs = [SignalIndexConversion.synchronize_df(df, t0) if not df.empty else df for df in self.dfs]
        return self

    def convert_index_to_sec(self):
        """ Method converting indices from nanoseconds to seconds

        :return: an updated SignalProcessing object with all signals having indices expressed in seconds
        """
        self.dfs = [SignalIndexConversion.convert_indices_to_sec(df) if not df.empty else df for df in self.dfs]
        return self

    def create_col_from_index(self, col:str):
        """ Method creating a column from index

        :param col: column name
        :return: an updated SignalProcessing object with a new column made as a copy of the index
        """
        for df in self.dfs:
            df[col] = df.index
        return self

    def filter_median(self, window=3):
        """ Method applying a median filter of a given window size to each signal

        :param window: median filter window size
        :return: an updated SignalProcessing object with each signal subject to median filtering
        """
        if window > 1:
            self.dfs = [df.rolling(window=window, min_periods=1).median() if not df.empty else df for df in self.dfs]
        return self

    def remove_values_for_time_less_than(self, time_threshold):
        """ Method removing values for rows less or equal to the time threshold.

        :param time_threshold: time threshold for row removal (rows greater than the threshold are maintained)
        :return: an updated SignalProcessing object with updated rows
        """
        self.dfs = [df[df.index > time_threshold] if not df.empty else df for df in self.dfs]
        return self

    def remove_initial_offset(self, time_threshold: float):
        """ Method removing an initial offset from the signal. The offset is calculated as an average value from the
        beginning of a signal until the input time_threshold (excluded)

        :param time_threshold: time threshold value for offset calculation
        :return: an updated SignalProcessing object with offset value subtracted from each signal
        """
        self.dfs = [df.subtract(df[df.index < time_threshold].mean()) if not df.empty else df for df in self.dfs]
        return self

    def filter_values(self, values):
        """ Method filtering values, i.e., provided list of values is maintained in the first column of each dataframe

        :param values: list of values to maintain
        :return: an updated SignalProcessin object with dataframes containing in the first column only the provided values
        """
        self.dfs = [df[df[df.columns[0]].isin(values)] for df in self.dfs]
        return self

    def map_values(self, mapping):
        """ Method mapping values in the first column of each dataframe based on the provided mapping dictionary

        :param mapping: a dictionary mapping values between existing and new ones
        :return: an updated SignalProcessing object with dataframes containing mapped values in the first column
        """
        self.dfs = [pd.DataFrame(df.apply(lambda col: mapping[col[df.columns[0]]], axis=1), columns=df.columns)
                    for df in self.dfs]
        return self

    def append_suffix_to_columns(self, suffix: str):
        """ Method appending a suffix to all columns. No extra character is added.

        :param suffix: suffix string
        :return: an updated SignalProcessing object with a suffix added to all columns in all dataframes
        """
        def append_suffix(df, suffix):
            df.columns = [col + suffix for col in df.columns]
            return df

        self.dfs = [append_suffix(df, suffix) if not df.empty else df for df in self.dfs ]
        return self

    def overwrite_sampling_time(self, t_sampling: float, t_end: float):
        """ Method overwriting an existing index with a new one given sampling time and new index duration.

        :param t_sampling: sampling time for a new index
        :param t_end: new end time for a new index
        :return: updated SignalProcessing object with new indices for all dataframes
        """
        def overwrite_sampling_time(df, t_sampling_, t_end_):
            new_index = np.arange(0, len(df), t_end_) * t_sampling_
            df.set_index(new_index, inplace=True)
            return df

        self.dfs = [overwrite_sampling_time(df, t_sampling, t_end) for df in self.dfs]
        return self
