import pandas as pd
import numpy as np
from sklearn.linear_model import LinearRegression

from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.QueryBuilder import FeatureProcessing


def convert_current_into_mean_vector(circuit_type: str, i_busbar_feature_df: pd.DataFrame) -> pd.DataFrame:
    """ Function converting current per device (power converter) and returning a vector of means

    :param circuit_type: circuit type (either RB or RQ)
    :param i_busbar_feature_df: a dataframe with an 'mean' column with a vector of mean voltage values
    :return:
    """

    gb = i_busbar_feature_df.groupby('device')
    x = {SignalMetadata.get_sector_names_from_components(circuit_type, 'PC', x): gb.get_group(x)['mean'].values for x
         in gb.groups}
    return pd.DataFrame(x)


def convert_voltage_into_mean_vector(u_res_feature_df: pd.DataFrame) -> pd.DataFrame:
    """ Function converting voltage per class (constant current part) into a mean vector

    :param u_res_feature_df: input dataframe with multiple columns containing average voltage
    :return: a dataframe with an 'out' column with a vector of mean voltage values
    """
    gb = u_res_feature_df.groupby('class')
    dfs = [gb.get_group(x) for x in gb.groups]

    out = dfs[0]
    columns = set(out.columns) - {'nxcals_variable_name'}
    out = out.rename(columns={col: '%s_0' % col for col in columns})
    out.index.names = [None]*len(out.index.names)
    for index in range(1, len(dfs)):
        dfs[index].index.names = [None] * len(dfs[index].index.names)
        out = out.merge(dfs[index], on='nxcals_variable_name')
        # add index suffix to columns
        out = out.rename(columns={col: '%s_%d' % (col, index) for col in columns})
    # Add sector name here
    out['mean_vector'] = out.filter(regex='mean_').values.tolist()
    return out


def calculate_resistance_from_lin_fit(circuit_type: str, busbar_voltage_df: pd.DataFrame,
                                      sector_to_i_meas_mean_df: pd.DataFrame) -> pd.DataFrame:
    """ Function calculating resistance from the slope of a linear fit of voltage and current

    :param circuit_type: type of a circuit
    :param busbar_voltage_df: dataframe with mean and std of voltage for linear fit computation
    :param sector_to_i_meas_mean_df: dataframe with mean and std of current for linear fit computation
    :return: a dataframe with a column containing one resistance per busbar
    """
    first_busbar = busbar_voltage_df.loc[0, 'nxcals_variable_name']
    RES_NAME = first_busbar.split(':')[1].replace('U_', 'R_')

    if circuit_type == 'RB':
        busbar_voltage_df['sector'] = busbar_voltage_df.apply(
            lambda row: SignalMetadata.get_sector_names_from_components(circuit_type, 'BUSBAR',
                                                                        row['nxcals_variable_name'.split(':')[0]]),
            axis=1)
    elif circuit_type == 'RQ':
        def get_sector_name(nxcals_variable_name):
            busbar_name = 'BUSBAR_RQD' if nxcals_variable_name[3] == 'D' else 'BUSBAR_RQF'
            sector_index = 0 if nxcals_variable_name[3] == 'D' else 1
            return SignalMetadata.get_sector_names_from_components(circuit_type, busbar_name,
                                                                   nxcals_variable_name.split(':')[0])[sector_index]

        busbar_voltage_df['sector'] = busbar_voltage_df.apply(lambda row:
                                                              get_sector_name(row['nxcals_variable_name']), axis=1)
    else:
        raise ValueError('Circuit type {} not supported!'.format(circuit_type))

    # Take slope coefficient of the linear fit of the V-I characteristics
    if len(sector_to_i_meas_mean_df) == 1:
        busbar_voltage_df[RES_NAME] = busbar_voltage_df.apply(
            lambda col: abs(col['mean_vector'][0]) / sector_to_i_meas_mean_df[col['sector']].values[0], axis=1)
    else:
        busbar_voltage_df[RES_NAME] = busbar_voltage_df.apply(lambda col: abs(
            LinearRegression().fit(sector_to_i_meas_mean_df[col['sector']].values.reshape((-1, 1)),
                                   col['mean_vector']).coef_[0]), axis=1)
    return busbar_voltage_df


class ResistanceBusbar(object):
    """ Base class with fields for voltage and current to be built with a ResistanceBuilder class.

    """

    def __init__(self, u_busbar_feature_df: pd.DataFrame):
        self.u_busbar_feature_df = u_busbar_feature_df
        self.i_busbar_feature_df = None

    def set_current(self, i_busbar_feature_df: pd.DataFrame):
        """ Method setting current for busbar resistance calculation

        :param i_busbar_feature_df: mean and std current per power converter
        :return: None
        """
        self.i_busbar_feature_df = i_busbar_feature_df


class ResistanceBuilder(object):
    """ Resistance Builder class with a method to set busbar voltage.

    """

    def with_busbar_voltage(self, u_busbar_feature_df: pd.DataFrame):
        """ Method setting voltage for busbar resistance calculation

        :param i_busbar_feature_df: mean and std voltage per busbar
        :return: an updated ResistanceBuilderBusbarVoltage object
        """
        u_busbar_feature_df = u_busbar_feature_df[
            u_busbar_feature_df['class'] != 0] if 'class' in u_busbar_feature_df else u_busbar_feature_df
        resistance_busbar = ResistanceBusbar(u_busbar_feature_df)

        return ResistanceBuilderBusbarVoltage(resistance_busbar)


class ResistanceBuilderBusbarVoltage(object):
    """ Resistance Builder class with a method to set busbar current (voltage already set).

    """


    def __init__(self, resistance_busbar):
        self.resistance_busbar = resistance_busbar

    def with_busbar_current(self, i_busbar_feature_df: pd.DataFrame):
        """ Method setting current for busbar resistance calculation

        :param i_busbar_feature_df: mean and std current per power converter
        :return: None
        """
        i_busbar_feature_df = i_busbar_feature_df[
            i_busbar_feature_df['class'] != 0] if 'class' in i_busbar_feature_df else i_busbar_feature_df
        self.resistance_busbar.set_current(i_busbar_feature_df)

        return ResistanceBuilderBusbarVoltageCurrent(self.resistance_busbar)


class ResistanceBuilderBusbarVoltageCurrent(object):
    """ Class provides methods to compute resistance from current and voltage dataframes.

    """

    def __init__(self, resistance_busbar):
        self.resistance_busbar = resistance_busbar
        self.df = None

    def calculate_mean_resistance(self, circuit_type: str):
        """ Method calculating mean resistance from mean voltage and current.

        :param circuit_type: circuit type (either RB or RQ)
        :return:
        """
        i_busbar_mean_vector_df = convert_current_into_mean_vector(circuit_type,
                                                                   self.resistance_busbar.i_busbar_feature_df)
        u_busbar_mean_vector_df = convert_voltage_into_mean_vector(self.resistance_busbar.u_busbar_feature_df)

        self.df = calculate_resistance_from_lin_fit(circuit_type, u_busbar_mean_vector_df, i_busbar_mean_vector_df)
        return self

    def convert_to_row(self, index: int):
        """ Method converting a column of mean voltage and current into a row. Method only considers two steps of the
        current profile (at injection and at a stable beams).

        :param index: index for the output row
        :return: an updated ResistanceBuilderBusbarVoltageCurrent object
        """
        if 'R_RES' in self.df.columns:
            out_df = self.df[['nxcals_variable_name', 'mean_0', 'std_0', 'mean_1', 'std_1', 'R_RES']]
        elif 'R_MAG' in self.df.columns:
            out_df = self.df[['nxcals_variable_name', 'mean_0', 'std_0', 'mean_1', 'std_1', 'R_MAG']]
        else:
            raise KeyError('R_RES or R_MAG not present in DF: {}'.format(self.df.columns))

        pd.set_option('mode.chained_assignment', None)
        out_df.loc[:, 'nxcals_variable_name_prefix'] = out_df['nxcals_variable_name'].apply(
            lambda row: row.split(':')[0])
        out_df.index = out_df['nxcals_variable_name_prefix'].values
        out_df.drop(columns=['nxcals_variable_name', 'nxcals_variable_name_prefix'], inplace=True)
        pd.set_option('mode.chained_assignment', 'warn')

        out_df = pd.DataFrame(out_df.stack())
        out_df.index = out_df.index.map('_'.join)
        out_df = out_df.T
        out_df.reset_index(inplace=True, drop=True)
        out_df.index = pd.Index([index])

        self.df = out_df
        return self

    def fill_missing_columns(self, signal_name: str, circuit_type='RB', split='_'):
        """ Method filling missing columns corresponding to missing voltage measurements. Missing data is filled with
        NaNs.

        :param signal_name: name of a signal to fill with NaNs
        :param circuit_type: circuit type
        :param split: character used to separate busbar name and signal name
        :return: an updated ResistanceBuilderBusbarVoltageCurrent object
        """
        # Get all signal names that was asked for
        busbars_ordered_df = MappingMetadata.read_busbar_to_magnet_mapping(circuit_type)

        # Get all column names in df
        list_df_columns = list(self.df.filter(regex='{}$'.format(signal_name), axis=1).columns)
        list_df_columns = [elem.split(split)[0] for elem in list_df_columns]

        columns_missing = list(set(busbars_ordered_df['Bus Bar Segment Name']) - set(list_df_columns))

        if len(columns_missing) > 0:
            for col in columns_missing:
                self.df.insert(0, ('{}{}{}'.format(col, split, signal_name)), np.nan)

        return self

    def sort_columns_busbar_location(self, signal_name: str, circuit_type='RB', circuit_name='*', split='_'):
        """ Method sorting busbar location for RB and RQ circuits according to their physical location in the
        LHC tunnel.
        :param signal_name: name of a signal to sort
        :param circuit_type: circuit type (RB and RQ supported)
        :param circuit_name: name of a circuit, e.g., RB.A12
        :param split: character on which signal name is split to recover busbar name for sorting
        :return: an updated ResistanceBuilderBusbarVoltageCurrent object
        """
        self.df.drop([col for col in self.df.columns if col[-1] not in '_%s' % signal_name], axis=1, inplace=True)
        res_df_t = self.df.T
        res_df_t['nxcals_variable_name'] = res_df_t.index
        res_df_t.reset_index(inplace=True, drop=True)

        res_df_t = FeatureProcessing(res_df_t). \
            sort_busbar_location(circuit_type=circuit_type, circuit_name=circuit_name, split=split).df

        res_df_t.set_index(res_df_t['nxcals_variable_name'], inplace=True)
        res_df_t.rename(columns={'nxcals_variable_name': 'Index'}, inplace=True)
        res_df_sorted = res_df_t.T
        res_df_sorted.drop(['Index'], inplace=True)

        self.df = res_df_sorted

        return self
