from typing import Tuple, List

import pandas as pd
from copy import deepcopy
import scipy.signal as spsig
from scipy import signal as s
import scipy as sp
import numpy as np
from scipy.spatial.distance import pdist, squareform

from lhcsmapi.pyedsl.dbsignal.SignalAnalysis import SignalAnalysis


def add_two_dataframes(df1: pd.DataFrame, df2: pd.DataFrame, col_name: str) -> pd.DataFrame:
    """ Function adding two dataframes and from the result creating a new DataFrame with a given column.
    It is assumed that both DataFrames have only one column.

    :param df1: first dataframe to add
    :param df2: second dataframe to add
    :param col_name: column name for the result
    :return: a dataframe with a sum of two input dataframes, interpolated index and a given column name
    """
    idx = df1.index.union(df2.index)
    idx_df1 = df1.reindex(idx).interpolate('index')
    idx_df2 = df2.reindex(idx).interpolate('index')

    return pd.DataFrame(index=idx_df1.index, data=idx_df1.values + idx_df2.values, columns=[col_name])


def calculate_char_time_profile(i_meas_df: pd.DataFrame, timestamp_first_quench=0) -> pd.DataFrame:
    """ Function calculating a time evolution characteristic decay time of a current profile

    :param i_meas_df: current profile
    :param timestamp_first_quench: relative time of the first quench to exclude from calculation values of current
    before the quench
    :return: a DataFrame with time evolution of a characteristic decay time
    """
    signal = i_meas_df.columns[0]
    der_name = 'd{}'.format(signal)
    tau_i_meas_df = calculate_time_derivative(deepcopy(i_meas_df), signal, der_name)

    # Calculate time constant as a ratio of a signal by its derivative
    char_time_name = '{0}/d{0}'.format(signal)
    tau_i_meas_df[char_time_name] = -i_meas_df[signal] / sp.signal.medfilt(tau_i_meas_df[der_name], 3)
    tau_i_meas_df.loc[tau_i_meas_df.index < timestamp_first_quench, [char_time_name]] = np.nan

    return pd.DataFrame(tau_i_meas_df[char_time_name])


def calculate_time_derivative(df: pd.DataFrame, source_signal: str, derivative_name: str,  kernel_size=51
                              ) -> pd.DataFrame:
    """ Function calculating time derivatives for signal contained in the passed DataFrame object.
    The returned signal name contains prefix d w.r.t. the input signal name, e.g. I_MEAS -> dI_MEAS.

    :param df: input DataFrame
    :param kernel_size: kernel size for gradient calculation
    :type kernel_size: int

    :return: pd.DataFrame with signal derivative
    """
    if isinstance(df, pd.DataFrame) and not df.empty:
        return pd.DataFrame(spsig.medfilt(np.gradient(df[source_signal]) / np.gradient(df.index), kernel_size),
                            df.index, columns=[derivative_name])
    else:
        raise ValueError('Wrong input type, please check the input assumptions')


def flip_u_qs0_signal(u_qs0_ref: pd.DataFrame, threshold=0.1) -> pd.DataFrame:
    """ Function flipping the U_QS0 signal such that the first signal increase has positive slope.

    :param u_qs0_ref: U_QS0 QPS signal
    :type u_qs0_ref: pd.DataFrame
    :param threshold: value of the quench detection threshold (in V)
    :type threshold: float

    :return: flipped U_QS0 signal
    """

    # # if there is only a negative peak, then flip the signal
    # # if there are two peaks and the first is negative, then flip the signal
    # # added the case when U_QS0 is saturated only to negative!
    # # otherwise, do nothing
    if all(u_qs0_ref.max() < threshold) and all(u_qs0_ref.min() < -threshold):
        return -deepcopy(u_qs0_ref)
    elif (u_qs0_ref.max().values[0] > threshold) & (u_qs0_ref.min().values[0] < -threshold) & (
            u_qs0_ref.idxmin().values[0] < u_qs0_ref.idxmax().values[0]):
        return -deepcopy(u_qs0_ref)
    else:
        return deepcopy(u_qs0_ref)


def calculate_last_const_di_dt(i_meas_df: pd.DataFrame, const_di_dt_duration=100) -> Tuple[float, float]:
    """ Function calculating the value and duration of the last di/dt prior to an event in the input signal.

    :param i_meas_df: pd.DataFrame with power converter current
    :type i_meas_df: pd.DataFrame
    :param const_di_dt_duration: minimum duration of constant di/dt in samples
    :type const_di_dt_duration: int

    :return: a tuple of last di/dt value and duration
    """
    # Calculate the first derivative
    signal = i_meas_df.columns[0]
    dsignal = 'd{}'.format(signal)
    d_i_meas_df = calculate_time_derivative(i_meas_df[i_meas_df.index < 0], signal, dsignal, kernel_size=51)
    # To remove noise round the derivative to a single digit
    d_i_meas_df['round_dSTATUS.I_MEAS'] = d_i_meas_df.apply(lambda row: round(row['dSTATUS.I_MEAS'], 1), axis=1)

    # Calculate the 2nd derivative
    signal = 'round_dSTATUS.I_MEAS'
    dsignal = 'd{}'.format(signal)
    dd_i_meas_df = calculate_time_derivative(d_i_meas_df, signal, dsignal, kernel_size=1)
    # Count duration of 2nd derivative equal to 0
    dd_i_meas_df['count'] = (dd_i_meas_df['dround_dSTATUS.I_MEAS'].diff(1) != 0).astype('int').cumsum()
    # Only take 2nd derivatives longer then a given number of points
    thresh = dd_i_meas_df['count'].value_counts() > const_di_dt_duration
    max_counts = thresh[thresh == True].index
    # Take indices of constant 2nd values
    index_max_counts = [dd_i_meas_df[dd_i_meas_df['count'] == max_count].index[0] for max_count in max_counts]

    # Sort according to the index_max_counts
    index_max_counts_sorted = sorted(index_max_counts, reverse=True)

    # Return the last constant didt
    last_di_dt = d_i_meas_df[dd_i_meas_df.index == max(index_max_counts_sorted)]['round_dSTATUS.I_MEAS'].values[0]
    index_max_count = index_max_counts_sorted.index(max(index_max_counts_sorted))
    if (index_max_count == 0) or (index_max_count == len(index_max_counts_sorted) - 1):
        last_di_dt_duration = -max(index_max_counts)
    else:
        last_di_dt_duration = index_max_counts_sorted[index_max_count] - index_max_counts_sorted[index_max_count + 1]

    return last_di_dt, last_di_dt_duration


def compare_curve_to_curve(series_decay: pd.DataFrame, series_decay_ref: pd.DataFrame, envelope_amplitude: float,
                           envelope_characteristic_time=False, apply_filter=False, filter_cut_off_frequency=100
                           ) -> float:
    """Function comparing a curve to its reference with a given envelope amplitude

    :param series_decay: actual signal
    :param series_decay_ref: reference signal
    :param envelope_amplitude: maximum allowed envelope amplitude
    :param envelope_characteristic_time: if not False, then an exponential decay of the envelope amplitude is considered
    the value of the envelope characteristic time is taken as a decay time constant.
    :param apply_filter: if not False, a frequency filter is applied (filter type is passed to the function)
    :param filter_cut_off_frequency: cutoff frequency for the applied filter
    :return: a ratio of points outside of the envelope to all points
    """
    end = np.min([len(series_decay), len(series_decay_ref)])

    series_decay = series_decay.iloc[0:end]
    series_decay_ref = series_decay_ref.iloc[0:end]

    if apply_filter:
        series_decay = SignalAnalysis.apply_frequency_filter(series_decay, filter_cut_off_frequency,
                                                             filter_type=apply_filter)
        series_decay_ref = SignalAnalysis.apply_frequency_filter(series_decay_ref, filter_cut_off_frequency,
                                                                 filter_type=apply_filter)

    index = series_decay.index
    if envelope_characteristic_time:
        envelope = np.exp(-index / envelope_characteristic_time)
        substract_array = np.abs(series_decay_ref.values - series_decay.values) / envelope
    else:
        substract_array = np.abs(series_decay_ref.values - series_decay.values)

    out_of_envelope_ratio = np.sum(substract_array > envelope_amplitude) / len(substract_array)
    return out_of_envelope_ratio


def calculate_similarity_matrix(df: pd.DataFrame) -> pd.DataFrame:
    """ Method calculating a similarity matrix across the columns in a DataFrame

    :param df: a DataFrame for which columns are calculated
    :return: a DataFrame with similarity matrix
    """
    sim_m = np.full(np.math.factorial(len(df.columns) - 1), -1)
    similarity = np.nonzero(np.triu(squareform(pdist(df.T, metric='euclidean'))).flatten())[0]
    if len(similarity) > 0:
        sim_m = np.triu(squareform(pdist(df.T, metric='euclidean'))).flatten()
    return pd.DataFrame(data=sim_m[np.nonzero(sim_m)], index=np.nonzero(sim_m)[0],
                        columns=[df.columns.values[0][:-2] + '_SimilarityMatrix'])


def calculate_filter_bank_from_spectrogram(spectrogram: Tuple[np.ndarray, np.ndarray, np.ndarray]) -> np.ndarray:
    """ Function calculating a filter bank feature from a spectrogram

    :param spectrogram: a tuple with the spectrogram features
    :return: filter bank feature
    """

    if spectrogram is not None:
        f, t, sxx = spectrogram
        filter_bank_max_values = np.zeros(len(sxx[:,1])//8)
        for i in range(len(filter_bank_max_values)):
            k = i*8
            filter_bank_max_values[i] = np.amax(sxx[k:k+8,:])

        return np.mean(filter_bank_max_values[1:5])


def normalize_df_of_signal_group(df: pd.DataFrame, signal_names: List[str]) -> None:
    """ Function normalizing a DataFrame with multiple columns based on a list of signal (column names)

    :param df: input DataFrame with multiple columns to normalize
    :param signal_names: list of signal (column) names to normalize
    :return: None as the df is updated in-place
    """
    for signals in signal_names:
        df[signals] = SignalAnalysis.normalize_df_with_global_max_and_min(df[signals])


def calculate_spectrogram_from_series(origin_series: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """ Function calculating a spectrogram from a series

    :param origin_series: input series to calculate a spectrogram
    :return: a three-element tuple representing a spectrogram
    """

    origin_series = origin_series.to_frame(name=None).set_index((origin_series.index - origin_series.index[0]))[
        origin_series.name]
    ts = (origin_series.index[1] - origin_series.index[0])  # Sample distance
    t = (origin_series.index[-1] - origin_series.index[0])  # Time period
    n = int(t / ts)  # Number of sample points
    if len(origin_series) == n:  # are the samples continuous
        new_series = origin_series
    else:
        new_series = SignalAnalysis.resample(origin_series)
    nperseg = n // 128
    try:
        f, t, sxx = s.spectrogram(new_series.values, 1 / ts, window=('hann'), nperseg=nperseg, noverlap=nperseg // 2)
        return f, t, sxx
    except Exception as e:
        print('Could not calculate spectrogram: ' + str(e))
