from abc import ABC, abstractmethod
import math
import re
from typing import List, Union, Dict

import pandas as pd
import numpy as np

from lhcsmapi.Time import Time
from lhcsmapi.pyedsl.dbsignal.Signal import Signal
from lhcsmapi.pyedsl.dbsignal.aft.AftDbRequest import AftDbRequest
from lhcsmapi.pyedsl.dbsignal.post_mortem.PmDbRequest import PmDbRequest
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.dbquery.FeatureNxcalsDbQueryBuilder import DevicePropertyNxcalsFeatureQuery, \
    VariableNxcalsFeatureQuery


def explode(query_params: pd.DataFrame, column: str) -> pd.DataFrame:
    """ Function exploding a column with entries as lists by propagating the same value across the rows

    :param query_params: a DataFrame with query parameters to explode
    :param column: name of a column to explode
    :return: an updated DataFrame with exploded rows for a given column
    """
    if any([row_type == list for row_type in query_params[column].apply(type).values]):
        return query_params.reindex(query_params.index.repeat(query_params[column].str.len())) \
            .assign(**{column: np.concatenate(query_params[column].values)}) \
            .reset_index(drop=True)
    else:
        return query_params


def create_query_params_template(circuit_name: str, system: str) -> pd.DataFrame:
    """ Function creating a template with query parameters (circuit name and system name)

    :param circuit_name: circuit name
    :param system: system name
    :return: DataFrame with circuit_name and system columns exploaded to account for vectorization
    """
    circuit_name = circuit_name if isinstance(circuit_name, list) else [circuit_name]
    query_params = pd.DataFrame(columns=['circuit_name'], data=circuit_name)

    # add system column with systems
    # system = [system] if isinstance(system, list) else [[system]]
    query_params['system'] = [system for i in query_params.index]
    # if system is a list, explode the row
    return explode(query_params, 'system')


def update_signal_wildcard(circuit_type: str, circuit_name: str, system: str, db: str, signal, signal_act: str,
                           kwargs: dict) -> Union[str, List[str]]:
    """ Function updating signal wildcards in query parameters

    :param circuit_type: circuit type
    :param circuit_name: circuit name
    :param system: system name
    :param db: database name
    :param signal: signal name
    :param signal_act: signal with wildcard to update
    :param kwargs: keyword-arguments with optional wildcard to update
    :return: an update signal signal or list of signal names
    """
    if '%' in signal_act:
        wildcard = re.findall('(?<=\%)(.*?)(?=\%)', signal_act)[0]

        if ('wildcard' in kwargs) and (wildcard in kwargs['wildcard']):
            wc_values = get_wildcard_values(circuit_name, circuit_type, db, kwargs, signal, system, wildcard)
            return [signal_act.replace('%{}%'.format(wildcard), wc_value) for wc_value in wc_values]
        else:
            raise KeyError('In order to fully define a signal name ({}), please provide a wildcard as '
                           'with_metadata(..., wildcard={{\'{}\': \'value\'}})'.format(signal_act, wildcard))
    else:
        return signal_act


def get_wildcard_values(circuit_name: str, circuit_type: str, db: str, kwargs: dict, signal: str, system: str,
                        wildcard: str) -> List[str]:
    """ Function providing wildcard values for query parameters, e.g., all busbar names for a given circuit type if
    * set as a wildcard.

    :param circuit_name: circuit name
    :param circuit_type: circuit type
    :param db: database name
    :param kwargs: keyword-argument database with wildcards
    :param signal: signal name
    :param system: system name
    :param wildcard: wildcard name
    :return: List of wildcard values for a signal
    """
    if '*' == kwargs['wildcard'][wildcard]:
        return SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit_type, circuit_name,
                                                                        system, db, signal)[1]
    else:
        wc_values = kwargs['wildcard'][wildcard]
        return wc_values if isinstance(wc_values, list) else [wc_values]


def get_signal_from_metadata(query_params: pd.DataFrame, circuit_type: str, db: str, signal: str, timestamp_query=None
                             ) -> pd.DataFrame:
    """ Function returning a column with signal name for a given circuit type

    :param query_params: DataFrame with query parameters
    :param circuit_type: circuit type
    :param db: database name
    :param signal: signal name
    :param timestamp_query: timestamp query to take an appropriate signal name from metadata module
    :return: a DataFrame with signal names for a given circuit type, name, system, and signal
    """
    return query_params.apply(lambda col: SignalMetadata.get_signal_name(circuit_type, col['circuit_name'],
                                                                         col['system'], db, signal, timestamp_query),
                              axis=1)


def get_metadata_keyword(query_params: pd.DataFrame, circuit_type: str, db: str, keyword: str, timestamp_query=None
                         ) -> pd.DataFrame:
    """ Function providing metadata for a given circuit type and db, metadata keyword taking into account timestamp of
    the query to choose an appropriate metadata.

    :param query_params: a DataFrame with query parameters
    :param circuit_type: circuit type
    :param db: database name
    :param keyword: metadata keyword
    :param timestamp_query: timestamp query to choose an appropriate metadata
    :return: a DataFrame with a given keyword of database metadata
    """
    return query_params.apply(lambda col: SignalMetadata.get_circuit_signal_database_metadata(
        circuit_type, col['circuit_name'], col['system'], db, timestamp_query)[keyword], axis=1)


def get_db_metadata(circuit_type: str, db: str, query_params: pd.DataFrame, timestamp_query=None) -> None:
    """ Function getting metadata database

    :param circuit_type: circuit type
    :param db: database name
    :param query_params: DataFrame with query parameters
    :param timestamp_query: timestamp query to choose an appropriate metadata
    :return: None as the query_params DataFrame is updated inplace
    """
    # get metadata
    # get metadata keywords that start with a small-case letter
    circuit_name = query_params['circuit_name'].values[0]
    system = query_params['system'].values[0]
    # The method assumes that all systems are the same
    metadata = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_name, system, db,
                                                                   timestamp_query)
    metadata = {k: v for k, v in metadata.items() if k[0].islower()}
    for key in metadata:
        query_params['metadata_{}'.format(key)] = get_metadata_keyword(query_params, circuit_type, db, keyword=key)


class DbQueryBuilder(ABC):
    """ Base abstract class providing a template for creation query parameters/metadata and query execution for pyeDSL.
    """
    _query_params = pd.DataFrame()

    def __init__(self, circuit_type):
        self.circuit_type = circuit_type

    @abstractmethod
    def create_query_parameters(self, query_params_dict):
        raise NotImplementedError("Not implemented")

    @abstractmethod
    def create_query_metadata(self, query_metadata_dict, timestamp_query):
        raise NotImplementedError("Not implemented")

    @abstractmethod
    def update_query_metadata(self, update_metadata_dict):
        raise NotImplementedError("Not implemented")

    @abstractmethod
    def execute_event_query(self, time_params_dict, **kwargs):
        raise NotImplementedError("Not implemented")

    @abstractmethod
    def execute_signal_query(self, time_params_dict, **kwargs):
        raise NotImplementedError("Not implemented")

    @abstractmethod
    def execute_feature_query(self, time_params_dict, features, **kwargs):
        raise NotImplementedError("Not implemented")

    @abstractmethod
    def execute_context_query(self, time_params_dict: Dict[str, int], context_info: str, **kwargs) -> pd.DataFrame:
        raise NotImplementedError("Not implemented")

    @abstractmethod
    def execute_fault_query(self, time_params_dict, **kwargs):
        raise NotImplementedError("Not implemented")


class DbQueryBuilderFactory(object):
    """ Class builds objects for query in a selected database and query type

    """

    @staticmethod
    def create(db, is_timestamp: bool, is_t_start: bool, circuit_type=None):
        """ Method creating a query builder object for a given database and query definition

        :param db: database name
        :param is_timestamp: true if timestamp is defined for a query
        :param is_t_start: true if t_start is defined for a query
        :param circuit_type: circuit type to initiate a query builder
        :return: a QueryBuilder database object for a given database and query type
        """
        if db == 'PM' and is_timestamp:
            return SignalPmDbQueryBuilder(circuit_type)
        elif db == 'PM' and is_t_start:
            return EventPmDbQueryBuilder(circuit_type)
        elif db == 'NXCALS' and is_t_start:
            return NxcalsDbQueryBuilder(circuit_type)
        elif db == 'AFT' and is_t_start:
            return AftDbQueryBuilder(circuit_type)
        else:
            raise ValueError('Database {} not supported for the provided time definition'.format(db))


class AftDbQueryBuilder(DbQueryBuilder):
    """ An implementation of the DbQueryBuilder abstract class for AFT queries

    """
    _db = 'AFT'

    def __init__(self, circuit_type):
        super().__init__(circuit_type)

    def create_query_parameters(self, query_params_dict):
        raise NotImplementedError('This method is not implemented for this class')

    def create_query_metadata(self, query_metadata_dict, timestamp_query):
        raise NotImplementedError('This method is not implemented for this class')

    def update_query_metadata(self, update_metadata_dict):
        raise NotImplementedError('This method is not implemented for this class')

    def execute_event_query(self, time_params_dict, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')

    def execute_signal_query(self, time_params_dict, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')

    def execute_feature_query(self, time_params_dict, features, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')

    def execute_context_query(self, time_params_dict: Dict[str, int], context_info: str, **kwargs) -> pd.DataFrame:
        """ Method executing an AFT context query

        :param time_params_dict: a dictionary with time parameters
        :param context_info: context information
        :param kwargs: optional keyword/arguments to parametrize a query
        :return: a DataFrame with context query results
        """
        dbconnector = kwargs['dbconnector']
        kwargs.pop('dbconnector', None)
        if 'context_id' in kwargs:
            context_id = kwargs['context_id']
            kwargs.pop('context_id', None)
            return AftDbRequest(dbconnector).context_query(context=context_info, context_id=context_id, **kwargs)
        else:
            return AftDbRequest(dbconnector).context_query(context=context_info, **kwargs)

    def execute_fault_query(self, time_params_dict: Dict[str, int], **kwargs) -> pd.DataFrame:
        """ Method executing an AFT fault query

        :param time_params_dict: a dictionary with time parameters
        :param kwargs: optional keyword/arguments to parametrize a query
        :return: a DataFrame with context fault results
        """
        dbconnector = kwargs['dbconnector']
        kwargs.pop('dbconnector', None)
        return AftDbRequest(dbconnector).fault_query(time_params_dict['t_start'], time_params_dict['t_end'], **kwargs)


class PmDbQueryBuilder(DbQueryBuilder):
    """ An implementation of the DbQueryBuilder abstract class for PM queries

    """
    _db = 'PM'

    def __init__(self, circuit_type):
        super().__init__(circuit_type)

    def create_query_parameters(self, query_params_dict: Dict[str, str]) -> None:
        """ Method creating query parameters for PM (system, className, source) after a DbQueryBuilder object is
        instantiated for queries with explicit query parameters.

        :param query_params_dict: PM query parameters
        :return: None as the method updates an instance of the class
        """
        dct = {'metadata_system': query_params_dict['system'], 'metadata_className': query_params_dict['className'],
               'metadata_source': query_params_dict['source']}
        if all([not isinstance(value, list) for value in dct.values()]):
            self._query_params = pd.DataFrame(dct, index=[0])
        else:
            self._query_params = pd.DataFrame(dct)

    def create_query_metadata(self, query_metadata_dict: Dict[str, str], timestamp_query=None) -> None:
        """ Method creating a query metadata after a DbQueryBuilder object is instantiated for metadata-based query

        :param query_metadata_dict: metadata (circuit name and syste) to get a metadata template
        :param timestamp_query: timestamp to select an appropriate metadata
        :return: None as the method updates an instance of the class
        """
        # initiate DataFrame with circuit names
        self._query_params = create_query_params_template(query_metadata_dict['circuit_name'],
                                                          query_metadata_dict['system'])
        # get metadata
        get_db_metadata(self.circuit_type, self._db, self._query_params, timestamp_query)

        self._query_params = explode(self._query_params, 'metadata_source')

    def update_query_metadata(self, update_metadata_dict: Dict[str, str]) -> None:
        """ Method updating wildcards in query metadata

        :param update_metadata_dict: a dictionary with wildcards to update
        :return: None as the method updates an instance of the class
        """

        def update_source(source_act: str, kwargs: dict):
            if '%' in source_act:
                wildcard = re.findall('(?<=\%)(.*?)(?=\%)', source_act)[0]

                if 'source' in kwargs:
                    return source_act.replace('%{}%'.format(wildcard), kwargs['source'])
                else:
                    raise KeyError(
                        'In order to fully define a source ({}), please provide a source (* for a wildcard) as '
                        'with_metadata(..., source=\'{}\')'.format(source_act, wildcard))
            else:
                return source_act

        self._query_params['metadata_source'] = \
            self._query_params.apply(lambda col: update_source(col['metadata_source'], update_metadata_dict), axis=1)

    def execute_event_query(self, time_params_dict, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')

    def execute_signal_query(self, time_params_dict, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')

    def execute_feature_query(self, time_params_dict, features, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')

    def execute_context_query(self, time_params_dict, context_info, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')

    def execute_fault_query(self, time_params_dict, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')


class EventPmDbQueryBuilder(PmDbQueryBuilder):
    """Implementation of DBQueryBuilder for PM event query
    """

    def create_query_parameters(self, query_params_dict: Dict[str, str]) -> None:
        """ Method creating query parameters for PM (system, className, source) after a DbQueryBuilder object is
        instantiated for queries with explicit query parameters.

        :param query_params_dict: PM query parameters
        :return: None as the method updates an instance of the class
        """
        super().create_query_parameters(query_params_dict)

        signal_ = query_params_dict['signal'] if 'signal' in query_params_dict else None

        if isinstance(signal_, str):
            self._query_params['signal'] = signal_
        elif isinstance(signal_, list):
            raise ValueError('The vectorized input is not supported for signal query of BLMs')

    def execute_event_query(self, time_params_dict: Dict[str, int], **kwargs) -> pd.DataFrame:
        """ Method executing a PM event query

        :param time_params_dict: a dictionary with t_start and t_end
        :param kwargs: optional keyword-arguments to parametrize the method
        :return: a DataFrame with source and timestamp for obtained PM events
        """
        if 'verbose' in kwargs and kwargs['verbose']:
            three_dots = '...' if len(self._query_params) > 5 else ''
            print('\tQuerying PM event timestamps for system: {}, className: {}, source: {} from {} to {}' \
                  .format(', '.join(set(self._query_params['metadata_system'].values[0:5])) + three_dots,
                          ', '.join(set(self._query_params['metadata_className'].values[0:5])) + three_dots,
                          ', '.join(set(self._query_params['metadata_source'].values[0:5])) + three_dots,
                          Time.to_string_short(time_params_dict['t_start']),
                          Time.to_string_short(time_params_dict['t_end'])))

        event_source_timestamps = []
        for index, row in self._query_params.iterrows():
            dct = {'source': row['metadata_source'],
                   'className': row['metadata_className'],
                   'system': row['metadata_system']}
            read_params = {**dct, **time_params_dict}
            event_source_timestamps += PmDbRequest.find_events_between_dates(**read_params)

        event_source_timestamp_df = pd.DataFrame(event_source_timestamps)
        event_source_timestamp_df.rename(columns={0: 'source', 1: 'timestamp'}, inplace=True)

        return event_source_timestamp_df

    def execute_signal_query(self, time_params_dict: Dict[str, int], **kwargs) -> List[pd.DataFrame]:
        """ Method executing a BLM signal query for a query with t_start and t_end (unlike typical PM query with an
        event timestamp)

        :param time_params_dict: a dictionary with t_start and t_end for the query
        :param kwargs: optinal keyword arguments to further parametrise a query
        :return: a list of DataFrames with the requested signal
        """
        if len(self._query_params) > 1:
            raise ValueError('execute_context_query() only supports scalar query parameters')

        dct = {'source': self._query_params.loc[0, 'metadata_source'],
               'className': self._query_params.loc[0, 'metadata_className'],
               'system': self._query_params.loc[0, 'metadata_system'],
               'signal': self._query_params.loc[0, 'signal']}
        read_params = {**dct, **time_params_dict}

        return [PmDbRequest.get_blm_signal(**read_params)]

    def execute_feature_query(self, time_params_dict, features, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')

    def execute_context_query(self, time_params_dict: Dict[str, int], contexts: List[str], **kwargs) -> pd.DataFrame:
        """ Method executing a BLM PM  context query

        :param time_params_dict: a dictionary with t_start and t_end for the query
        :param contexts: a list of context variables
        :param kwargs: a list of optional keyword/arguments
        :return: a DataFrame with BLM context parameters
        """
        if len(self._query_params) > 1:
            raise ValueError('execute_context_query() only supports scalar query parameters')

        dct = {'source': self._query_params.loc[0, 'metadata_source'],
               'className': self._query_params.loc[0, 'metadata_className'],
               'system': self._query_params.loc[0, 'metadata_system'],
               'parameters': contexts}
        read_params = {**dct, **time_params_dict}

        return PmDbRequest.query_context(**read_params)

    def execute_fault_query(self, time_params_dict, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')


class SignalPmDbQueryBuilder(PmDbQueryBuilder):
    """ Implementation of DBQueryBuilder for PM signal query
    """

    def __init__(self, circuit_type):
        super().__init__(circuit_type)

    def create_query_parameters(self, query_params_dict: Dict[str, str]) -> None:
        """ Method creating query parameters for PM (system, className, source) after a DbQueryBuilder object is
        instantiated for queries with explicit query parameters. It allows for a vectorized code execution.

        :param query_params_dict: PM query parameters
        :return: None as the method updates an instance of the class
        """
        super().create_query_parameters(query_params_dict)
        signal_ = query_params_dict['signal']
        if isinstance(signal_, str):
            self._query_params['signal'] = signal_
        elif isinstance(signal_, list) and (self._query_params.shape[0] == len(signal_)):
            self._query_params['signal'] = signal_
        elif isinstance(signal_, list) and (self._query_params.shape[0] != len(signal_)):
            self._query_params['signal'] = [signal_]
            self._query_params = explode(self._query_params, 'signal')
        else:
            raise ValueError('The polymorphic input does not match')

    def create_query_metadata(self, query_metadata_dict: Dict[str, str], timestamp_query=None) -> None:
        """ Method creating a query metadata after a DbQueryBuilder object is instantiated for metadata-based query

        :param query_metadata_dict: metadata (circuit name and system) to get a metadata template
        :param timestamp_query: timestamp to select an appropriate metadata
        :return: None as the method updates an instance of the class
        """
        super().create_query_metadata(query_metadata_dict, timestamp_query)

        self._query_params['signal'] = get_signal_from_metadata(self._query_params, self.circuit_type, self._db,
                                                                query_metadata_dict['signal'], timestamp_query)
        self._query_params = explode(self._query_params, 'signal')

    def update_query_metadata(self, update_metadata_dict: Dict[str, str]) -> None:
        """ Method updating wildcards in query metadata

        :param update_metadata_dict: a dictionary with wildcards to update
        :return: None as the method updates an instance of the class
        """
        self._query_params['signal'] = self._query_params.apply(
            lambda col: update_signal_wildcard(self.circuit_type, col['circuit_name'], col['system'], self._db,
                                               update_metadata_dict['signal'],
                                               col['signal'], update_metadata_dict), axis=1)

        self._query_params = explode(self._query_params, 'signal')
        super().update_query_metadata(update_metadata_dict)

    def execute_event_query(self, time_params_dict, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')

    def execute_signal_query(self, time_params_dict: Dict[str, int], **kwargs) -> List[pd.DataFrame]:
        """ Method executing a PM signal query for a given timestamp. Optionally it allows for displaying query metadata
        and timestamp to facilitate query reproduction. For some specific queries duplicated timestamps may be
        maintained.

        :param time_params_dict: a dictionary with query timestamp.
        :param kwargs: optional keyword/arguments to parametrize the query
        :return: a list of DataFrames for all considered systems and signals
        """
        if 'verbose' in kwargs and kwargs['verbose']:
            three_dots = '...' if len(self._query_params) > 5 else ''
            print('\tQuerying PM event signal(s) {} for system: {}, className: {}, source: {} at {}'
                  .format(', '.join(set(self._query_params['signal'].values[0:5])) + three_dots,
                          ', '.join(set(self._query_params['metadata_source'].values[0:5])) + three_dots,
                          ', '.join(set(self._query_params['metadata_className'].values[0:5])) + three_dots,
                          ', '.join(set(self._query_params['metadata_system'].values[0:5])) + three_dots,
                          Time.to_string_short(time_params_dict['eventTime'])))
        circuit_dfs = []
        for index, row in self._query_params.iterrows():
            dct = {'source': row['metadata_source'],
                   'className': row['metadata_className'],
                   'system': row['metadata_system'],
                   'signal': row['signal'],
                   'allow_duplicate_index': kwargs['allow_duplicate_index']}
            read_params = {**dct, **time_params_dict}
            circuit_dfs.append(Signal().read('pm', **read_params))

        return circuit_dfs

    def execute_feature_query(self, time_params_dict, features, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')

    def execute_context_query(self, time_params_dict, context_info, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')

    def execute_fault_query(self, time_params_dict, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')


class NxcalsDbQueryBuilder(DbQueryBuilder):
    """ Implementation of DBQueryBuilder for NXCALS signal and feature query
    """

    _db = 'NXCALS'

    def create_query_parameters(self, query_params_dict: Dict[str, str]) -> None:
        """ Method creating query parameters for an NXCALS query (raw metadata query)

        :param query_params_dict: a dictionary with NXCALS query parameters
        :return: None as the method updates an instance of the class
        """
        if 'nxcals_device' in query_params_dict:
            dct = {'metadata_system': query_params_dict['nxcals_system'],
                   'metadata_device': query_params_dict['nxcals_device'],
                   'metadata_property': query_params_dict['nxcals_property'],
                   'signal': query_params_dict['signal']}
        else:
            dct = {'metadata_system': query_params_dict['nxcals_system'],
                   'signal': query_params_dict['signal']}
        if all([not isinstance(value, list) for value in dct.values()]):
            self._query_params = pd.DataFrame(dct, index=[0])
        else:
            self._query_params = pd.DataFrame(dct)

    def create_query_metadata(self, query_metadata_dict: Dict[str, str], timestamp_query=None) -> None:
        """ Method creating a query metadata after a DbQueryBuilder object is instantiated for metadata-based query

        :param query_metadata_dict: metadata (circuit name and system) to get a metadata template
        :param timestamp_query: timestamp to select an appropriate metadata
        :return: None as the method updates an instance of the class
        """
        if '*' == query_metadata_dict['circuit_name']:
            query_metadata_dict['circuit_name'] = SignalMetadata.get_circuit_names(self.circuit_type, timestamp_query)

        self._query_params = create_query_params_template(query_metadata_dict['circuit_name'],
                                                          query_metadata_dict['system'])

        get_db_metadata(self.circuit_type, self._db, self._query_params)
        self._query_params['signal'] = get_signal_from_metadata(self._query_params, self.circuit_type,
                                                                self._db, query_metadata_dict['signal'],
                                                                timestamp_query)

    def update_query_metadata(self, update_metadata_dict: Dict[str, str]) -> None:
        """ Method updating wildcards in query metadata

        :param update_metadata_dict: a dictionary with wildcards to update
        :return: None as the method updates an instance of the class
        """

        self._query_params['signal'] = self._query_params.apply(
            lambda col: update_signal_wildcard(self.circuit_type, col['circuit_name'],
                                               col['system'], self._db, update_metadata_dict['signal'],
                                               col['signal'], update_metadata_dict), axis=1)
        if 'metadata_device' in self._query_params.columns:
            self._query_params = explode(self._query_params, 'metadata_device')
        self._query_params = explode(self._query_params, 'signal')

    def execute_event_query(self, time_params_dict, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')

    def execute_signal_query(self, time_params_dict: Dict[str, int], **kwargs) -> List[pd.DataFrame]:
        """ Method executing an NXCALS signal query

        :param time_params_dict: t_start and t_end for a query
        :param kwargs: optional keyword/arguments for a query
        :return: a list of signal DataFrames
        """
        if 'verbose' in kwargs and kwargs['verbose']:
            three_dots = '...' if len(self._query_params) > 5 else ''
            print('\tQuerying NXCALS signal(s) {} from {} to {}'
                  .format(', '.join(set(self._query_params['signal'].values[0:5])) + three_dots,
                          Time.to_string_short(time_params_dict['t_start']),
                          Time.to_string_short(time_params_dict['t_end'])))

        circuit_dfs = []
        for index, row in self._query_params.iterrows():
            dct = {'signal': row['signal'],
                   'spark': kwargs['dbconnector']}

            # Add metadata: it could be either (system, device, property) or only (system)
            for metadata_col in self._query_params.filter(regex='metadata').columns:
                metadata_suffix = metadata_col.split('_')[1]
                dct['nxcals_{}'.format(metadata_suffix)] = row[metadata_col]

            read_params = {**dct, **time_params_dict}
            circuit_dfs.append(Signal().read('nxcals', **read_params))

        return circuit_dfs

    @staticmethod
    def fill_missing_rows(results_df: pd.DataFrame, all_row_names: List[str]) -> pd.DataFrame:
        """ Method filling missing rows in a feature query w.r.t. all requested rows

        :param results_df: feature DataFrame
        :param all_row_names: a list of all requested rows
        :return: a DataFrame with NaN-filled empty rows
        """
        if 'device' in results_df:
            col = 'device'
        elif 'nxcals_variable_name' in results_df:
            col = 'nxcals_variable_name'
        else:
            raise ValueError('DataFrame does not contain device or nxcals_variable_name columns')

        names_missing = set(all_row_names) - set(results_df[col])
        if len(names_missing) > 0:
            df_missing = pd.DataFrame(list(names_missing), columns=[col]).sort_values(
                by=col)  # Needs to be sortet here, if not might get error based upon random selection of names_missing
            results_df = results_df.append(df_missing, ignore_index=True, sort=False)

        if 'class' in results_df:
            num_of_classes = results_df.drop_duplicates('class')
            if (len(num_of_classes) > 1) & (num_of_classes.isnull().values.any()):
                nan_df = results_df[results_df['class'].isnull()]

                # Add in N x nan_df to result_df
                for i in range(len(num_of_classes) - 2):
                    results_df = results_df.append(nan_df, ignore_index=True)

                range_tmp = list(
                    range(len(num_of_classes) - 1))  # List of class names/numbers (can only be incremantal from 0...N)
                idx_list = list(nan_df.index)

                # update with class numbers, based upon how many classes exist
                for idx in idx_list:
                    results_df.loc[results_df[col] == results_df[col][idx], 'class'] = range_tmp
                    if 'mean' in results_df:
                        results_df.loc[results_df[col] == results_df[col][idx], 'mean'] = 0

        return results_df

    def execute_feature_query(self, time_params_dict: Dict[str, int], features: List[str], **kwargs) -> pd.DataFrame:
        """ Method executing a feature query

        :param time_params_dict: start and end time for a feature query
        :param features: list of features to query
        :param kwargs: optional keyword/arguments to parametrise a query
        :return: a DataFrame with signal features
        """

        feature_query_results_df = self.nxcals_feature_query(time_params_dict, features, kwargs)

        # Fill with NaN missing entries
        if 'metadata_device' in self._query_params:
            feature_query_results_df = self.fill_missing_rows(feature_query_results_df,
                                                              self._query_params['metadata_device'].values)
        else:
            feature_query_results_df = self.fill_missing_rows(feature_query_results_df,
                                                              self._query_params['signal'].values)

        return feature_query_results_df

    def nxcals_feature_query(self, time_params_dict: Dict[str, int], features: List[str], kwargs) -> pd.DataFrame:
        """ Method executing a feature query

        :param time_params_dict: start and end time for a feature query
        :param features: list of features to query
        :param kwargs: optional keyword/arguments to parametrise a query
        :return: a DataFrame with signal features
        """
        spark = kwargs['dbconnector']
        system = self._query_params.loc[0, 'metadata_system']
        t_start = int(time_params_dict['t_start'])
        t_end = int(time_params_dict['t_end'])
        feature_query = DevicePropertyNxcalsFeatureQuery() if 'metadata_device' in self._query_params \
            else VariableNxcalsFeatureQuery()
        features = list(set(features))
        col_names = feature_query.get_col_names(kwargs)
        feature_query_results_df = pd.DataFrame(columns=col_names + features)

        # Handle limit of 1000
        index = 0
        MAX_NXCALS_SIGNALS = 1000
        query_divisor = math.ceil(len(self._query_params) / MAX_NXCALS_SIGNALS)
        max_list_size = math.ceil(len(self._query_params) / query_divisor)
        while index < len(self._query_params):
            feature_query.initialize_query(spark, system, t_start, t_end)
            for _, row in self._query_params[index:index + max_list_size].iterrows():
                feature_query.populate_query(row)

            feature_query.build_dataset(self._query_params)
            feature_query.decimate_columns(kwargs)
            feature_query.translate_column(kwargs)
            feature_query.group_by(col_names)
            features_pyspark = feature_query.get_features(features, self._query_params)
            aggregates = feature_query.aggregate(features_pyspark)

            feature_query_results_df_temp = pd.DataFrame(aggregates, columns=feature_query_results_df.columns)
            feature_query_results_df = pd.concat([feature_query_results_df, feature_query_results_df_temp], axis=0)
            index += max_list_size
        return feature_query_results_df

    def execute_context_query(self, time_params_dict, context_info, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')

    def execute_fault_query(self, time_params_dict, **kwargs):
        raise NotImplementedError('This method is not implemented for this class')
