import pandas as pd
from copy import deepcopy

from lhcsmapi.pyedsl.dbsignal.nxcals.NxcalsDbSignal import NxcalsDbSignal
from lhcsmapi.pyedsl.dbsignal.post_mortem.PmDbSignal import PmDbSignal
from lhcsmapi.pyedsl.dbsignal.SignalIndexConversion import SignalIndexConversion


class Signal(object):
    """Signal class encapsulates behavior for working with signal databases.
    It provides read and write methods for accessing Post Mortem, NXCALS, and InfluxDB APIs.
    """

    __time_units = {'PM': 'ns', 'NXCALS': 'ns'}

    def __init__(self, df=None):

        self.__df_state = None
        self.__df = pd.DataFrame()
        if df is not None:
            self.df = df

    # region df property
    @property
    def df(self):
        return self.__df

    @df.setter
    def df(self, df):
        """Method validates that a DataFrame is passed
        """
        if not isinstance(df, pd.DataFrame):
            raise TypeError("You can only assign a pandas DataFrame to Signal.df")

        self.__df = df

    # endregion
    def read(self, api, append=False, **kwargs):
        """Method allowing users to read signal from a given database api

            :param api: input signal database signal ("nxcals", "pm", "influx")
            :type api: string
            :param append: flag indicating whether the signal should be appended to the object's dataframe (True) or not (False). False emulates static behavior
            :type append: boolean
            :param kwargs: dictionary of (keyword: argument) pairs which parametrize an db API query
            :type kwargs: dictionary
            :return: pandas.DataFrame

            - Example PM::

                  >>> from lhcsmapi.pyedsl.dbsignal.Signal import Signal
                  >>> system = "QPS"
                  >>> className = "DQAMCNMB_PMHSU"
                  >>> magnet = "B20L5"
                  >>> signal = "I_HDS_4"
                  >>> timestamp = 1426220469491000000
                  >>> i_hds_4_df = Signal().read('pm', system=system, className=className, source=magnet,
                  ...                           signal=signal, eventTime=timestamp) # doctest: +SKIP
                  >>> print(i_hds_4_df.head()) # doctest: +SKIP
                                        I_HDS_4
                  1426220469491020832  0.089648
                  1426220469491026040  0.091555
                  1426220469491031248  0.091555
                  1426220469491036456  0.093463
                  1426220469491041664  0.091555

            - Expected Fail Response PM::

                      >>> from lhcsmapi.pyedsl.dbsignal.Signal import Signal
                      >>> system = "QPS"
                      >>> className = "DQAMCNMB_PMHSU"
                      >>> magnet = "B20L5"
                      >>> signal = "I_HDS_4"
                      >>> timestamp = 1426220469491000000
                      >>> i_hds_4_df = Signal().read('pm', system=system, className=className, source=magnet,
                      ...                              signal=signal, eventTime=timestamp) # doctest: +SKIP
                      >>> print(i_hds_4_df.head()) # doctest: +SKIP
                      Querying Post Mortem failed using the following query:
                      http://pm-api-pro/v2/pmdata/signal?system=QPS&className=DQAMCNMB_PMHS&source=B20L5&timestampInNanos=1426220469491000000&signal=I_HDS_4
        """

        api = api.upper()

        df = Signal._database_access_handler(api, 'read', **kwargs)
        if append:
            df = pd.concat([df, self.df], axis=1)
        else:
            self.__df_state = api

        self.df = df

        return self.df

    @staticmethod
    def write(api, **kwargs):
        """
            **Writes df into a database**

            Method writes dataframe associated with an object to a selected database.

            :param api: input signal database signal ("influx")
            :type api: string
            :param kwargs: dictionary of (keyword: argument) pairs which parametrize an db API query
            :type kwargs: dictionary

        """
        return Signal._database_access_handler(api, 'write', **kwargs)

    @staticmethod
    def _database_access_handler(api, access_method_identifier, **kwargs):

        db_access = Signal.execute_dbsignal_factory(api, **kwargs)
        try:
            access_method = getattr(db_access, access_method_identifier, None)
            return access_method(**kwargs)
        except AttributeError:
            raise AttributeError('{} Signal has no {} method'.format(api, access_method_identifier))

    @staticmethod
    def execute_dbsignal_factory(signal_type, **kwargs):
        """**Returns an object of one of the derived DbSignal classes according to the passed signal_type**

            Method executes a factory producing signal database APIs.

            :param signal_type: input signal database signal ("nxcals", "pm")
            :type signal_type: string
            :param kwargs: dictionary of (keyword: argument) pairs which parametrize an db API query
            :type kwargs: dictionary
            :return: datafrane

            - Example PM::

                  >>> from lhcsmapi.pyedsl.dbsignal.Signal import Signal
                  >>> Signal.execute_dbsignal_factory("pm", t_start=1, duration = [(0, 's')]) # doctest: +SKIP
                  <lhcsmapi.post_mortem.PmDbSignal.PmDbSignal at 0x7f4ec387e780>

                  >>> from lhcsmapi.pyedsl.dbsignal.Signal import Signal
                  >>> Signal.execute_dbsignal_factory("PM", t_start=1, duration = [(0, 's')]) # doctest: +SKIP
                  <lhcsmapi.post_mortem.PmDbSignal.PmDbSignal at 0x7f4ec387e780>

            - Expected Fail Response::

                  >>> from lhcsmapi.pyedsl.dbsignal.Signal import Signal
                  >>> Signal.execute_dbsignal_factory("aft", duration = [(0, 's')])
                  Traceback (most recent call last):
                  ...
                  TypeError: AFT key does not match any known APIs: PM, NXCALS
                """

        signal_type_upper = signal_type.upper()
        if signal_type_upper == "PM":
            return PmDbSignal.build_signal(**kwargs)
        elif signal_type_upper == "NXCALS":
            return NxcalsDbSignal.build_signal(**kwargs)
        else:
            raise TypeError('{} key does not match any known APIs: PM, NXCALS'.format(signal_type_upper))

    def get_synchronized_time_df(self, t0=None):
        """ Method synchronizing DataFrame time to a given t0 time

        :param t0: time to which a signal is synchronized (in ns)
        :return: DataFrame with synchronized time
        """

        output = SignalIndexConversion.synchronize_df(deepcopy(self.df), t0)
        return output
