from typing import List, Tuple

import numpy as np
from scipy import signal as s
import pandas as pd


class SignalAnalysis:
    """ Class with methods for signal and dataframe analysis

    """

    @staticmethod
    def find_rows_with_exact_pattern_in_series(series: pd.DataFrame, pattern: List[float]
                                               ) -> Tuple[pd.DataFrame, List[int]]:
        """ Method finding rows in a signal that match the input pattern

        :param series: input series/signal in which a pattern is searched
        :param pattern: pattern to be found
        :return: a tuple with a DataFrame containing rows matching exactly the pattern and indices of rows with the
        requested pattern.
        """
        series_with_last_index_of_pattern = series.rolling(len(pattern)).apply(lambda x: all(np.equal(x, pattern)),
                                                                          raw=False)
        df_with_pattern = series_with_last_index_of_pattern.fillna(0)
        # indices with the last found element
        last_indices_of_pattern = np.where(df_with_pattern)[0]
        # all indices corresponding to the pattern
        all_indices_of_pattern = [range(match - len(pattern) + 1, match + 1) for match in last_indices_of_pattern]

        data = []
        for subs in all_indices_of_pattern:
            data.append(series.iloc[subs].index)

        # flatten indices
        flattened_indices = [val for indices_of_pattern in all_indices_of_pattern for val in indices_of_pattern]

        return pd.DataFrame(data, columns=pattern), flattened_indices

    @staticmethod
    def resample(origin_series: pd.Series, ts=0, t0=None, t_end=None) -> pd.Series:
        """ Function resampling a series by interpolating values to get continuous sample frequency

        :param origin_series: input series to be resampled
        :param ts: sampling time (if zero, the sampling time is deduced from the first to indices)
        :param t0: start time for the resampling (if None, then it is deduced as the first index)
        :param t_end: end time for the resampling (if None, then it is deduced as the last index)
        :return: resampled series
        """

        def _get_time_point_distribution(ts: float, t0: float, t_end: float) -> np.ndarray:

            t = t_end - t0  # Time period
            n = int(t / ts)  # Number of sample points
            return np.linspace(t0, t + t0, n)

        # takes sample time of first sample if not given
        if ts == 0:
            ts = (origin_series.index[1] - origin_series.index[0])  # Sample distance

        if t0 is not None and t_end is not None:
            index_list = _get_time_point_distribution(ts, t0, t_end)
        else:
            index_list = _get_time_point_distribution(ts, origin_series.index[0], origin_series.index[-1])

        # interpolate data in order to have the same sample time everywhere
        interpolated_series = origin_series.reindex(origin_series.index.union(index_list)).\
            interpolate('index', limit_direction='forward')

        # delete those points which are not in in the continuous interval
        del_index = [x for x in origin_series.index if x not in index_list]
        return interpolated_series.drop(del_index, axis=0)

    @staticmethod
    def apply_frequency_filter(origin_series: pd.Series, sample_frequency=0, cut_off_frequency=0, filter_type='low'
                               ) -> pd.Series:
        """ Method applying a low pass filter to an input series

        :param origin_series: input series
        :param sample_frequency: sampling frequency (if 0, it is an invers of the difference of the first two indices)
        :param cut_off_frequency: cut off frequency (if 0, it is sampling frequency divided by 100)
        :param filter_type: 'lowpass', 'highpass', 'bandpass', 'bandstop'
        :return: filtered dataframe
        """

        # Set cuttoff frequency if not given
        if sample_frequency == 0:
            sample_frequency = 1 / (origin_series.index[1] - origin_series.index[0])  # Sample distance
        if cut_off_frequency == 0:
            cut_off_frequency = sample_frequency / 100  # Cut-off frequency of the filter
        normalized_frequency = cut_off_frequency / (sample_frequency / 2)  # Normalize the frequency

        b, a = s.butter(5, normalized_frequency, filter_type)
        filtered_series = pd.DataFrame(data=s.filtfilt(b, a, origin_series.values, axis=0), index=origin_series.index,
                                      columns=['filtered'])
        return filtered_series['filtered']

    @staticmethod
    def normalize_df_with_global_max_and_min(df: pd.DataFrame) -> pd.DataFrame:
        """ Method normalizing DataFrame based on its global maximum and minimum values

        :param df: input DataFrame to normalize
        :return: normalized DataFrame
        """
        df = df.dropna()
        return (df - np.min(df.values)) / (np.max(df.values) - np.min(df.values))
