import warnings

import pandas as pd
import numpy as np
from datetime import datetime
from lhcsmapi.Time import Time
from lhcsmapi.pyedsl.dbsignal.DbSignal import DbSignal, validate_key_in_kwargs

try:
    from cern.nxcals.pyquery.builders import VariableQuery, DevicePropertyQuery
except:
    pass


class NxcalsDbSignal(DbSignal):
    """NxcalsDbSignal class encapsulates parsing of a signal obtained from the NXCALS database.
    """

    def __init__(self, signal=None, t_start=0, t_end=0, nxcals_system=None, nxcals_device=None, nxcals_property=None,
                 spark=None):
        super().__init__('NXCALS')
        self.signal = signal
        self.t_start = t_start
        self.t_end = t_end
        self.nxcals_system = nxcals_system
        self.nxcals_device = nxcals_device
        self.nxcals_property = nxcals_property
        self.spark = spark

        if self.spark is None:
            raise Exception('You have to provide an instance of spark connector via the spark keyword')

    @staticmethod
    def build_signal(**kwargs):
        """ Method building a signal object based on input keyword/arguments.
        Method is validating the input dictionary in order to properly set the arguments

        :param kwargs: keyword/arguments to parametrize a query
        :return: a NxcalsDbSignal object
        """
        # look through kwargs for constructor arguments
        signal_name = validate_key_in_kwargs('signal', str, **kwargs) or None

        t_start = validate_key_in_kwargs('t_start', int, pd.Timestamp, str, float, datetime, tuple, np.datetime64,
                                         **kwargs) or 0
        t_end = validate_key_in_kwargs('t_end', int, pd.Timestamp, str, float, datetime, tuple, np.datetime64,
                                       **kwargs) or None
        duration = validate_key_in_kwargs('duration', tuple, list, **kwargs)
        tz = validate_key_in_kwargs('tz', str, **kwargs)

        nxcals_system = validate_key_in_kwargs('nxcals_system', str, **kwargs) or None
        nxcals_device = validate_key_in_kwargs('nxcals_device', str, **kwargs) or None
        nxcals_property = validate_key_in_kwargs('nxcals_property', str, **kwargs) or None

        # calculate start and end time based of parsed arguments
        t_start, t_end = Time.get_query_period_in_unix_time(t_start, t_end, duration, tz)

        # convert start and end to unix time in seconds with us precision for CALS api
        t_start_nxcals = pd.Timestamp(t_start).to_datetime64()
        t_end_nxcals = pd.Timestamp(t_end).to_datetime64()

        # different syntax since the spark instance can't be validated without being a dependency
        spark = kwargs['spark'] if 'spark' in kwargs else None

        return NxcalsDbSignal(signal_name, t_start_nxcals, t_end_nxcals, nxcals_system,
                              nxcals_device, nxcals_property, spark)

    def read(self, signal=None, **kwargs) -> pd.DataFrame:
        """ Method returning a pandas DataFrame containing all signals with an index formatted to ns

        :param signal: signal to read
        :param kwargs: optional keyword/arguments to parametrise a query
        :return: a DataFrame with requested signal
        """

        return NxcalsDbSignal.query_signal(self.signal, self.t_start, self.t_end, self.nxcals_system,
                                           self.nxcals_device, self.nxcals_property, self.spark)

    @staticmethod
    def query_signal(signal: str, t_start, t_end, nxcals_system: str, nxcals_device: str, nxcals_property: str, spark
                     ) -> pd.DataFrame:
        """ Method querying NXCALS between start and end time either with system only or system, device and property.
        In case of failing queries, an appropriate warning is displayed and an empty DataFrame returned

        :param signal: signal name
        :param t_start: start time
        :param t_end: end time
        :param nxcals_system: NXCALS system name
        :param nxcals_device: NXCALS device name (optional)
        :param nxcals_property: NXCALS property name (optional)
        :param spark: spark connector
        :return: a DataFrame with a requested signal
        """
        if (nxcals_device is not None) and (nxcals_property is not None):

            try:
                signal_df = DevicePropertyQuery.builder(spark).system(nxcals_system) \
                    .startTime(t_start).endTime(t_end) \
                    .entity().device(nxcals_device).property(nxcals_property) \
                    .buildDataset().select('acqStamp', signal).toPandas()

                signal_df.set_index(signal_df['acqStamp'], inplace=True)
                signal_df.drop(columns='acqStamp', inplace=True)
                signal_df.dropna(inplace=True)
                signal_df.sort_index(inplace=True)
                return signal_df
            except NameError:
                exception = """NXCALS query failed with the parameters, signal: {0}, start time: {1}, end time: {2}, 
                nxcals_system {3}, nxcals_device {4}, nxcals_property {5}\n
                Empty signal returned.
                ___________________________________________________________________________________
                You can reproduce the error using the raw NXCALS API:\n
                from cern.nxcals.pyquery.builders import *
                DevicePropertyQuery.builder(spark)\\
                .system('{3}')\\
                .startTime('{1}').endTime('{2}').entity()\\
                .device('{4}').property('{5}').buildDataset()\\
                .select('acqStamp', {0}).toPandas()
                """.format(signal, t_start, t_end, nxcals_system, nxcals_device, nxcals_property)

                warnings.warn(exception)
                return pd.DataFrame()
        else:
            try:
                signal_df = VariableQuery.builder(spark) \
                    .system(nxcals_system) \
                    .startTime(t_start).endTime(t_end) \
                    .variable(signal) \
                    .buildDataset().select('nxcals_timestamp', 'nxcals_value').toPandas()

                signal_df.set_index(signal_df['nxcals_timestamp'], inplace=True)
                signal_df.drop(columns='nxcals_timestamp', inplace=True)
                signal_df.dropna(inplace=True)
                signal_df.sort_index(inplace=True)
                signal_df.rename(columns={'nxcals_value': signal}, inplace=True)
                return signal_df
            except NameError:
                exception = """NXCALS query failed with the parameters, signal: {0}, start time: {1}, end time: {2},
                nxcals_system {3} failed\n
                Empty signal returned.
                ___________________________________________________________________________________
                You can reproduce the error using the raw NXCALS API:\n
                from cern.nxcals.pyquery.builders import *\n
                VariableQuery.builder(spark)\\
                .system({3})\\
                .startTime({1}).endTime({2})\\
                .variable({0})\\
                .buildDataset()\\
                .select('nxcals_timestamp', 'nxcals_value').toPandas()
                """.format(signal, t_start, t_end, nxcals_system)

                warnings.warn(exception)
                return pd.DataFrame()
