from pathlib import Path
import os
import warnings
import unittest
from unittest import mock

import pandas as pd
import numpy as np

from lhcsmapi.Time import Time
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder, validate_time_definition, Query, FeatureProcessing, \
    SignalProcessing, EventProcessing
from lhcsmapi.metadata.SignalMetadata import SignalMetadata

# === USED FOR MOCK
from lhcsmapi.pyedsl.dbquery.DbQueryBuilder import NxcalsDbQueryBuilder
from unittest.mock import patch

from test.test_pyedsl.test_dbsignal.test_post_mortem.test_PmDbRequest import mocked_pmdbrequest_requests_get
from test.test_pyedsl.test_dbsignal.test_post_mortem.test_PmDbSignal import mocked_pmdbsignal_requests_get


def read_csv(file_name):
    path = Path(os.path.dirname(__file__))
    full_path = os.path.join(str(path.parent), 'resources/pyedsl/{}'.format(file_name))
    return pd.read_csv(full_path, index_col=0)


class TestQueryBuilder(unittest.TestCase):

    def test_validate_time_definition(self):
        # arrange
        time_def_dct = {'t_start': np.int64(1544631694792000000),
                        't_end': np.int64(1544631694800000000)}

        # act
        time_definition = validate_time_definition(time_def_dct)

        # assert
        self.assertEqual(1544631694792000000, time_definition[0])
        self.assertEqual(1544631694800000000, time_definition[1])

    def test_query_get_time_params_timestamp(self):
        # arrange
        query = Query()
        timestamp = 1544631694792000000

        # act
        query.set_db('PM')
        query.set_timestamp(timestamp)

        # assert
        self.assertEqual({'eventTime': 1544631694792000000}, query.get_time_params())

    def test_query_get_time_params_t_start_t_end(self):
        # arrange
        query = Query()
        t_start = 1544631694792000000
        t_end = 1544631694800000000

        # act
        query.set_db('NXCALS')
        query.set_t_start(t_start)
        query.set_t_end(t_end)

        # assert
        self.assertEqual({'t_start': 1544631694792000000, 't_end': 1544631694800000000}, query.get_time_params())

    def test_query_get_time_params_timestamp_error(self):
        # arrange
        query = Query()
        timestamp = 1544631694792000000

        # act
        query.set_db('NXCALS')
        query.set_timestamp(timestamp)

        # assert
        with self.assertRaises(ValueError) as context:
            query.get_time_params()

        self.assertTrue('Wrong time definition!' in context.exception.args[0])

    def test_query_builder_wrong_attr_error(self):
        # arrange
        # act

        # assert
        with self.assertRaises(AttributeError) as context:
            QueryBuilder().with_cals()

        self.assertTrue('with_cals() is not supported. Set database name using with_pm(), '
                        'with_nxcals(spark), with_aft(session) methods.' in context.exception.args[0])

    def test_query_builder_repr(self):
        # arrange
        qb = QueryBuilder()

        # act
        # assert
        self.assertEqual('Set database name using with_pm(), '
                        'with_nxcals(spark), with_aft(session) methods.', str(qb))


    def test_query_builder_with_aft(self):
        # arrange
        qb = QueryBuilder().with_aft(None)

        # act
        # assert
        self.assertEqual('AFT', qb.query.db)
        self.assertEqual(None, qb.query.dbconnector)

    def test_query_builder_with_db_repr(self):
        # arrange
        qb = QueryBuilder().with_aft(None)

        # act
        # assert
        self.assertEqual('Database name properly set to AFT. Set time definition: for PM signal query, with_timestamp(),'
                         '\n for PM event query or AFT, NXCALS signal query, with_duration()', str(qb))

    def test_query_builder_with_db_wrong_attr(self):
        # arrange

        # act
        # assert
        with self.assertRaises(AttributeError) as context:
            QueryBuilder().with_aft(None).with_t_start()

        self.assertTrue('with_t_start() is not supported. Set time definition: for PM signal query, with_timestamp(),\n'
                        ' for PM event query or AFT, NXCALS signal query, with_duration()' in context.exception.args[0])

    def test_query_builder_with_db_wrong_attr_db_mismatch(self):
        # arrange

        # act
        # assert
        with self.assertRaises(AttributeError) as context:
            QueryBuilder().with_aft(None).with_timestamp(1)

        self.assertTrue('Timestamp can be only set for PM database, got AFT' in context.exception.args[0])

    def test_query_builder_context_query_pm_error(self):
        # arrange
        qb = QueryBuilder().with_pm()

        # act
        # assert
        with self.assertRaises(NotImplementedError) as context:
            qb.context_query(None)

        self.assertTrue('This method is not implemented for this database PM' in context.exception.args[0])

    def test_query_builder_context_query_aft(self):
        # arrange
        qb = QueryBuilder().with_aft(None)

        # act
        # assert
        with self.assertRaises(ValueError) as context:
            qb.context_query(None)

        self.assertTrue('Context None is not supported by AFT REST API. For more details, '
                        'please consult https://aft.cern.ch/docs' in context.exception.args[0])

    def test_query_builder_with_db_time_def_error(self):
        # arrange
        qb = QueryBuilder().with_pm().with_timestamp(1544631694792000000)

        # act
        # assert
        with self.assertRaises(AttributeError) as context:
            qb.signal_query()

        self.assertTrue('signal_query() is not supported. \nSet generic query parameter using with_query_parameters() '
                        'method, or a circuit signal using with_circuit_type() method.' in context.exception.args[0])

    def test_query_builder_with_db_time_def_repr_timestamp(self):
        # arrange
        qb = QueryBuilder().with_pm().with_timestamp(1544631694792000000)

        # act
        # assert

        self.assertEqual('Timestamp set to timestamp=1544631694792000000. Set generic query parameter using '
                         'with_query_parameters() method, or a circuit signal using with_circuit_type() method.',
                         str(qb))

    def test_query_builder_with_db_time_def_repr_t_start_t_end(self):
        # arrange
        qb = QueryBuilder().with_pm().with_duration(t_start=1544631694792000000, t_end=1544631694800000000)

        # act
        # assert

        self.assertEqual('Query duration set to t_start=1544631694792000000, t_end=1544631694800000000. Set generic '
                         'query parameter using with_query_parameters() method, or a circuit signal using '
                         'with_circuit_type() method.', str(qb))

    def test_query_builder_with_db_time_def_repr_timestamp_with_circuit_type(self):
        # arrange
        qb = QueryBuilder().with_pm().with_timestamp(1544631694792000000)

        # act
        # assert
        with self.assertRaises(ValueError) as context:
            qb.with_circuit_type('ITT')

        self.assertTrue('Circuit type ITT not supported!' in context.exception.args[0])

    def test_query_builder_fault_query_pm_error(self):
        # arrange
        qb = QueryBuilder().with_pm().with_timestamp(1)

        # act
        # assert
        with self.assertRaises(NotImplementedError) as context:
            qb.fault_query()

        self.assertTrue('This method is not implemented for this database PM' in context.exception.args[0])

    def test_query_builder_fault_query_aft(self):
        # arrange
        qb = QueryBuilder().with_aft(None).with_duration(t_start=0, t_end=1)

        # act
        # assert
        with self.assertRaises(AttributeError) as context:
            qb.fault_query()

        self.assertTrue('\'NoneType\' object has no attribute \'get\'' in context.exception.args[0])

    def test_query_builder_with_db_time_def_circuit_type_get_attr_error(self):
        # arrange
        qb = QueryBuilder().with_pm().with_timestamp(1544631694792000000).with_circuit_type('RB')

        # act
        # assert
        with self.assertRaises(AttributeError) as context:
            qb.signal_query()

        self.assertTrue('signal_query() not supported. \nSet metadata using with_metadata() method, e.g., '
                        'with_metadata(circuit_name=\'RB.A12\', system=\'PC\', signal=\'I_MEAS\').'
                        in context.exception.args[0])

    def test_query_builder_with_db_time_def_circuit_type_pm_timestamp_repr(self):
        # arrange
        qb = QueryBuilder().with_pm().with_timestamp(1544631694792000000).with_circuit_type('RB')

        # act
        # assert
        self.assertEqual('Circuit type set to RB. \nSet metadata using with_metadata() method, e.g., '
                         'with_metadata(circuit_name=\'RB.A12\', system=\'PC\', signal=\'I_MEAS\').', str(qb))

    # QueryBuilderWithDbTimeDefMetadata
    def test_query_builder_with_db_time_def_metadata_pm_timestamp_event_query(self):
        # arrange
        qb = QueryBuilder().with_pm().with_timestamp(1544631694792000000).with_circuit_type('RB')\
            .with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS')

        # act
        # assert
        with self.assertRaises(AttributeError) as context:
            qb.event_query()

        self.assertTrue('event_query() not supported, for PM with timestamp you can call signal_query()'
                        in context.exception.args[0])

    def test_query_builder_with_db_time_def_metadata_pm_duration_feature_query(self):
        # arrange
        qb = QueryBuilder().with_pm().with_duration(t_start=0, t_end=1).with_circuit_type('RB')\
            .with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS')

        # act
        # assert
        with self.assertRaises(AttributeError) as context:
            qb.feature_query(['max'])

        self.assertTrue('feature_query() not supported, for PM with duration you can call '
                        'event_query(), signal_query(), context_query(); the former two for BLM'
                        in context.exception.args[0])

    def test_query_builder_with_db_time_def_metadata_nxcals_duration_event_query(self):
        # arrange
        qb = QueryBuilder().with_nxcals(None).with_duration(t_start=0, t_end=1).with_circuit_type('RB')\
            .with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS')

        # act
        # assert
        with self.assertRaises(AttributeError) as context:
            qb.event_query()

        self.assertTrue('event_query() not supported, for NXCALS with duration you can call '
                        'signal_query(), feature_query()'
                        in context.exception.args[0])

    def test_query_builder_with_db_time_def_metadata_nxcals_duration_query_event(self):
        # arrange
        qb = QueryBuilder().with_nxcals(None).with_duration(t_start=0, t_end=1).with_circuit_type('RB')\
            .with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS')

        # act
        # assert
        with self.assertRaises(AttributeError) as context:
            qb.query_event()

        self.assertTrue('query_event() not supported, for NXCALS with duration you can call '
                        'signal_query(), feature_query()'
                        in context.exception.args[0])

    def test_query_builder_with_db_time_def_metadata_pm_timestamp_repr(self):
        # arrange
        qb = QueryBuilder().with_pm().with_timestamp(1544631694792000000).with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS')

        # act
        # assert
        self.assertEqual('Query PM signal using signal_query() method.', str(qb))

    def test_query_builder_with_db_time_def_metadata_pm_duration_repr(self):
        # arrange
        qb = QueryBuilder().with_pm().with_duration(t_start=0, t_end=1).with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A12', system='PC')

        # act
        # assert
        self.assertEqual('Query PM event using event_query() method.\n'
                         'Query PM BLM signal using signal_query() method.\n'
                         'Query PM BLM context using context_query() method.', str(qb))

    def test_query_builder_with_db_time_def_metadata_nxcals_duration_repr(self):
        # arrange
        qb = QueryBuilder().with_nxcals(None).with_duration(t_start=0, t_end=1).with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS')

        # act
        # assert
        self.assertEqual('Query NXCALS signals using signal_query() method.\n'
                         'Query NXCALS signal features using feature_query() method.', str(qb))

    @mock.patch('requests.get', side_effect=mocked_pmdbsignal_requests_get)
    def test_query_builder_pm_timestamp_signal_query(self, mock_get):
        # act
        i_meas_df = QueryBuilder().with_pm().with_timestamp(1426220469520000000).with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A45', system='PC', signal='I_MEAS').signal_query().dfs[0]

        # assert
        i_meas_df_ref = pd.DataFrame({'STATUS.I_MEAS': {1426220409500000000: 9571.501,
                                                        1426220409520000000: 9571.699,
                                                        1426220409540000000: 9571.899,
                                                        1426220409560000000: 9572.1,
                                                        1426220409580000000: 9572.3}})
        self.assertEqual(len(mock_get.call_args_list), 1)
        pd.testing.assert_frame_equal(i_meas_df_ref, i_meas_df.head())

    @mock.patch('requests.get', side_effect=mocked_pmdbsignal_requests_get)
    def test_query_builder_pm_timestamp_signal_query_append_suffix_to_column(self, mock_get):
        # act
        i_meas_df = QueryBuilder().with_pm().with_timestamp(1426220469520000000).with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A45', system='PC', signal='I_MEAS') \
            .signal_query().append_suffix_to_columns('_REF').dfs[0]

        # assert
        i_meas_df_ref = pd.DataFrame({'STATUS.I_MEAS_REF': {1426220409500000000: 9571.501,
                                                        1426220409520000000: 9571.699,
                                                        1426220409540000000: 9571.899,
                                                        1426220409560000000: 9572.1,
                                                        1426220409580000000: 9572.3}})
        pd.testing.assert_frame_equal(i_meas_df_ref, i_meas_df.head())

    @mock.patch('requests.get', side_effect=mocked_pmdbsignal_requests_get)
    def test_query_builder_pm_timestamp_signal_query_sync_time(self, mock_get):
        # act
        timestamp = 1426220469520000000
        i_meas_df = QueryBuilder().with_pm().with_timestamp(timestamp).with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A45', system='PC', signal='I_MEAS') \
            .signal_query().synchronize_time(timestamp).dfs[0]

        # assert
        i_meas_df_ref = pd.DataFrame({'STATUS.I_MEAS': {-60020000000: 9571.501,
                                                        -60000000000: 9571.699,
                                                        -59980000000: 9571.899,
                                                        -59960000000: 9572.1,
                                                        -59940000000: 9572.3}})
        self.assertEqual(len(mock_get.call_args_list), 0)
        pd.testing.assert_frame_equal(i_meas_df_ref, i_meas_df.head())

    @mock.patch('requests.get', side_effect=mocked_pmdbsignal_requests_get)
    def test_query_builder_pm_timestamp_signal_query_sync_time_convert_to_sec(self, mock_get):
        # act
        timestamp = 1426220469520000000
        i_meas_df = QueryBuilder().with_pm().with_timestamp(timestamp).with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A45', system='PC', signal='I_MEAS') \
            .signal_query().synchronize_time(timestamp).convert_index_to_sec().dfs[0]

        # assert
        i_meas_df_ref = pd.DataFrame({'STATUS.I_MEAS': {-60.02: 9571.501,
                                                        -60.00: 9571.699,
                                                        -59.98: 9571.899,
                                                        -59.96: 9572.1,
                                                        -59.94: 9572.3}})
        self.assertEqual(len(mock_get.call_args_list), 0)
        pd.testing.assert_frame_equal(i_meas_df_ref, i_meas_df.head())

    @mock.patch('requests.get', side_effect=mocked_pmdbsignal_requests_get)
    def test_query_builder_pm_timestamp_signal_query_sync_time_convert_to_sec_remove_values_for_time(self, mock_get):
        # act
        timestamp = 1426220469520000000
        i_meas_df = QueryBuilder().with_pm().with_timestamp(timestamp).with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A45', system='PC', signal='I_MEAS') \
            .signal_query().synchronize_time(timestamp).convert_index_to_sec() \
            .remove_values_for_time_less_than(0).dfs[0]

        # assert
        i_meas_df_ref = pd.DataFrame({'STATUS.I_MEAS': {0.02: 10171.523,
                                                        0.04: 10171.574,
                                                        0.06: 10171.418,
                                                        0.08: 10171.383,
                                                        0.1: 10171.342}})
        pd.testing.assert_frame_equal(i_meas_df_ref, i_meas_df.head())

    @mock.patch('requests.get', side_effect=mocked_pmdbsignal_requests_get)
    def test_query_builder_pm_timestamp_signal_query_sync_time_convert_to_sec_filter_median(self, mock_get):
        # arrange
        timestamp = 1426220469520000000
        # act
        i_meas_df = QueryBuilder().with_pm().with_timestamp(timestamp).with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A45', system='PC', signal='I_MEAS') \
            .signal_query().synchronize_time(timestamp).convert_index_to_sec().filter_median().dfs[0]

        # assert
        i_meas_df_ref = pd.DataFrame({'STATUS.I_MEAS': {-60.02: 9571.501,
                                                        -60.00: 9571.6,
                                                        -59.98: 9571.699,
                                                        -59.96: 9571.899,
                                                        -59.94: 9572.1}})
        pd.testing.assert_frame_equal(i_meas_df_ref, i_meas_df.head())

    @mock.patch('requests.get', side_effect=mocked_pmdbsignal_requests_get)
    def test_query_builder_pm_timestamp_signal_query_sync_time_convert_to_sec_remove_initial_offset(self, mock_get):
        # arrange
        # arrange
        timestamp = 1426220469491000000
        # act
        i_hds_df = QueryBuilder().with_pm().with_timestamp(timestamp) \
            .with_query_parameters(system='QPS', className='DQAMCNMB_PMHSU', source='B20L5', signal='I_HDS_4') \
            .signal_query().remove_initial_offset(1426220469491041664).dfs[0]

        # assert
        i_hds_df_ref = pd.DataFrame({'I_HDS_4': {1426220469491020832: -0.0019074000000000035,
                                                  1426220469491026040: 0.0,
                                                  1426220469491031248: 0.0,
                                                  1426220469491036456: 0.0019074000000000035,
                                                  1426220469491041664: 0.0}})
        pd.testing.assert_frame_equal(i_hds_df_ref, i_hds_df.head())


    @mock.patch('requests.get', side_effect=mocked_pmdbrequest_requests_get)
    def test_query_builder_pm_duration_event_query(self, mock_get):
        # arrange
        # act
        source_timestamp_act_df = QueryBuilder().with_pm() \
            .with_duration(t_start=1426201200000000000, duration=(24 * 60 * 60, 's')) \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A45', system='QH', source='*').event_query().df

        # assert
        source_timestamp_ref_df = pd.DataFrame({'source': {0: 'B20L5', 1: 'C20L5', 2: 'A20L5', 3: 'A21L5', 4: 'B21L5',
                                                           5: 'C23L4', 6: 'B23L4', 7: 'A23L4', 8: 'C22L4', 9: 'C15R4',
                                                           10: 'B15R4', 11: 'A15R4', 12: 'B34L8', 13: 'C34L8',
                                                           14: 'A34L8', 15: 'C33L8', 16: 'C34R7', 17: 'A34R7',
                                                           18: 'A20R3', 19: 'B20R3', 20: 'C20R3', 21: 'B18L5',
                                                           22: 'A18L5', 23: 'C18L5', 24: 'A19L5'},
                                                'timestamp': {0: 1426220469491000000, 1: 1426220517100000000,
                                                              2: 1426220518112000000, 3: 1426220625990000000,
                                                              4: 1426220866112000000, 5: 1426236802332000000,
                                                              6: 1426236839404000000, 7: 1426236839832000000,
                                                              8: 1426236949841000000, 9: 1426251285711000000,
                                                              10: 1426251337747000000, 11: 1426251388741000000,
                                                              12: 1426258716281000000, 13: 1426258747672000000,
                                                              14: 1426258747370000000, 15: 1426258835955000000,
                                                              16: 1426258853947000000, 17: 1426258854113000000,
                                                              18: 1426267931956000000, 19: 1426267983579000000,
                                                              20: 1426268004144000000, 21: 1426277626360000000,
                                                              22: 1426277679838000000, 23: 1426277680496000000,
                                                              24: 1426277903449000000}})

        pd.testing.assert_frame_equal(source_timestamp_ref_df, source_timestamp_act_df)

    @mock.patch('requests.get', side_effect=mocked_pmdbrequest_requests_get)
    def test_query_builder_pm_duration_event_query_filter_source(self, mock_get):
        # arrange
        # act
        source_timestamp_act_df = QueryBuilder().with_pm() \
            .with_duration(t_start=1426201200000000000, duration=(24 * 60 * 60, 's')) \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A45', system='QH', source='*') \
            .event_query().filter_source(circuit_name='RB.A45', system='QH').df

        # assert
        source_timestamp_ref_df = pd.DataFrame({'source': {0: 'B20L5', 1: 'C20L5', 2: 'A20L5', 3: 'A21L5', 4: 'B21L5',
                                                           5: 'C15R4', 6: 'B15R4', 7: 'A15R4', 8: 'B18L5', 9: 'A18L5',
                                                           10: 'C18L5', 11: 'A19L5'},
                                                'timestamp': {0: 1426220469491000000, 1: 1426220517100000000,
                                                              2: 1426220518112000000, 3: 1426220625990000000,
                                                              4: 1426220866112000000, 5: 1426251285711000000,
                                                              6: 1426251337747000000, 7: 1426251388741000000,
                                                              8: 1426277626360000000, 9: 1426277679838000000,
                                                              10: 1426277680496000000, 11: 1426277903449000000}})

        pd.testing.assert_frame_equal(source_timestamp_ref_df, source_timestamp_act_df)

    @mock.patch('requests.get', side_effect=mocked_pmdbrequest_requests_get)
    def test_query_builder_pm_duration_event_query_filter_source_rq(self, mock_get):
        # arrange
        # act
        source_timestamp_act_df = QueryBuilder().with_pm() \
            .with_duration(t_start=1426201200000000000, duration=(24 * 60 * 60, 's')) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A45', system='QH', source='*') \
            .event_query().filter_source(circuit_name='RQ.A45', system='QH').df

        # assert
        source_timestamp_ref_df = pd.DataFrame(pd.DataFrame({'source': pd.Series([], dtype='str'),
                                                             'timestamp': pd.Series([], dtype='int')}))

        pd.testing.assert_frame_equal(source_timestamp_ref_df, source_timestamp_act_df)

    @mock.patch('requests.get', side_effect=mocked_pmdbrequest_requests_get)
    def test_query_builder_pm_duration_event_query_filter_source_drop_duplicate_source(self, mock_get):
        # arrange
        # act
        source_timestamp_act_df = QueryBuilder().with_pm() \
            .with_duration(t_start=1426201200000000000, duration=(24 * 60 * 60, 's')) \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A45', system='QH', source='*') \
            .event_query().filter_source(circuit_name='RB.A45', system='QH').drop_duplicate_source().df

        # assert
        source_timestamp_ref_df = pd.DataFrame({'source': {0: 'B20L5', 1: 'C20L5', 2: 'A20L5', 3: 'A21L5', 4: 'B21L5',
                                                           5: 'C15R4', 6: 'B15R4', 7: 'A15R4', 8: 'B18L5', 9: 'A18L5',
                                                           10: 'C18L5', 11: 'A19L5'},
                                                'timestamp': {0: 1426220469491000000, 1: 1426220517100000000,
                                                              2: 1426220518112000000, 3: 1426220625990000000,
                                                              4: 1426220866112000000, 5: 1426251285711000000,
                                                              6: 1426251337747000000, 7: 1426251388741000000,
                                                              8: 1426277626360000000, 9: 1426277679838000000,
                                                              10: 1426277680496000000, 11: 1426277903449000000}})

        pd.testing.assert_frame_equal(source_timestamp_ref_df, source_timestamp_act_df)

    @mock.patch('requests.get', side_effect=mocked_pmdbrequest_requests_get)
    def test_query_builder_pm_duration_event_query_filter_source_sort_values(self, mock_get):
        # arrange
        # act
        source_timestamp_act_df = QueryBuilder().with_pm() \
            .with_duration(t_start=1426201200000000000, duration=(24 * 60 * 60, 's')) \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A45', system='QH', source='*') \
            .event_query().filter_source(circuit_name='RB.A45', system='QH').sort_values(by='timestamp').df

        # assert
        source_timestamp_ref_df = pd.DataFrame({'source': {0: 'B20L5', 1: 'C20L5', 2: 'A20L5', 3: 'A21L5', 4: 'B21L5',
                                                           5: 'C15R4', 6: 'B15R4', 7: 'A15R4', 8: 'B18L5', 9: 'A18L5',
                                                           10: 'C18L5', 11: 'A19L5'},
                                                'timestamp': {0: 1426220469491000000, 1: 1426220517100000000,
                                                              2: 1426220518112000000, 3: 1426220625990000000,
                                                              4: 1426220866112000000, 5: 1426251285711000000,
                                                              6: 1426251337747000000, 7: 1426251388741000000,
                                                              8: 1426277626360000000, 9: 1426277679838000000,
                                                              10: 1426277680496000000, 11: 1426277903449000000}})

        pd.testing.assert_frame_equal(source_timestamp_ref_df, source_timestamp_act_df)

    @mock.patch('requests.get', side_effect=mocked_pmdbrequest_requests_get)
    def test_query_builder_pm_duration_context_query(self, mock_get):
        # arrange
        start_time_date = '2015-11-23 07:28:53+01:00'
        t_start, t_end = Time.get_query_period_in_unix_time(start_time_date=start_time_date,
                                                            duration_date = [(1, 's'), (2, 's')])
        # act
        context_df_act = QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, duration=[(2, 's')]) \
            .with_query_parameters(system='BLM', className='BLMLHC', source='*') \
            .context_query(["pmFillNum"]).df

        # assert
        context_df_ref = pd.DataFrame({'pmFillNum': {1448260133517488525: 4647}})
        self.assertEqual(len(mock_get.call_args_list), 1)

        pd.testing.assert_frame_equal(context_df_ref, context_df_act)

    # with_db().with_query_parameters()
    # PM: event query
    def test_eq_query_parameters_qps_16l2_source(self):
        """Tests PM metadata for 1 signal, 1 system, 1 source, 1 className"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm()\
            .with_duration(t_start=1544622149620000000, duration=[(10, 's'), (400, 's')]) \
            .with_query_parameters(system='QPS', source='16L2', className='DQAMCNMQ_PMHSU') \
            .query.db_query_builder._query_params

        # assert metadata
        metadata_ref = read_csv('query_builder/pm_eq/QPS_system_source_className.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    # PM: signal query
    def test_sq_query_parameters_rqda12_fgc_i_meas(self):
        """Tests PM metadata for 1 signal, 1 system, 1 source, 1 className"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(1426220469520000000) \
            .with_query_parameters(signal='STATUS.I_MEAS', system='FGC', source='RPTE.UA47.RB.A45',
                                   className='51_self_pmd') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/PC_system_source_className_I_MEAS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    # PM: signal query
    def test_sq_query_parameters_rqda12_fgc_i_meas_i_ref(self):
        """Tests PM metadata for 2 signals, 2 systems, 2 sources, 2 classNames"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(1426220469520000000) \
            .with_query_parameters(signal=['STATUS.I_MEAS', 'STATUS.I_REF'], system=['FGC', 'FGC'],
                                   source=['RPTE.UA47.RB.A45', 'RPTE.UA47.RB.A45'],
                                   className=['51_self_pmd', '51_self_pmd']) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/PC_system_source_className_I_MEAS_I_REF.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    # PM: signal query
    def test_sq_query_parameters_rqda12_fgc_i_meas_i_ref_polymorphic(self):
        """Tests PM metadata for 2 signals, 1 systems, 1 sources, 1 classNames"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(1426220469520000000) \
            .with_query_parameters(signal=['STATUS.I_MEAS', 'STATUS.I_REF'], system='FGC',
                                   source='RPTE.UA47.RB.A45',
                                   className='51_self_pmd') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/PC_system_source_className_I_MEAS_I_REF.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    # NXCALS: signal_query
    def test_sq_query_parameters_i_meas(self):
        """Tests NXCALS metadata for 1 circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""
        # arrange
        # act
        spark = None
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_query_parameters(nxcals_system='CMW', nxcals_device='RPHE.UA23.RQD.A12',
                                   nxcals_property='SUB', signal='I_MEAS')\
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_sq/FGC_system_device_property_I_MEAS_1_circuit.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

    # NXCALS: signal_query
    def test_sq_query_parameters_i_meas_2_circuits(self):
        """Tests NXCALS metadata for 2 circuit_names, 2 systems, 2 signal, 0 sources, 0 wildcards"""
        # arrange
        # act
        spark = None
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_query_parameters(nxcals_system=['CMW', 'CMW'],
                                   nxcals_device=['RPHE.UA23.RQD.A12','RPHE.UA23.RQF.A12'],
                                   nxcals_property=['SUB', 'SUB'], signal=['I_MEAS', 'I_MEAS']) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_sq/FGC_system_device_property_I_MEAS_2_circuits.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

    # NXCALS: signal_query
    def test_sq_query_parameters_i_meas_2_circuits_single_system(self):
        """Tests NXCALS metadata for 2 circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""
        # arrange
        # act
        spark = None

        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_query_parameters(nxcals_system='CMW',
                                   nxcals_device=['RPHE.UA23.RQD.A12', 'RPHE.UA23.RQF.A12'],
                                   nxcals_property='SUB', signal='I_MEAS') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_sq/FGC_system_device_property_I_MEAS_2_circuits.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

    # with_db().with_circuit_type().with_duration()/with_timestamp().with_metadata()

    # PM: event_query
    # 1 circuit_name
    # 1 system
    # 1 source (optional)
    def test_eq_metadata_rqda12_pc_fgc_1_source(self):
        """Tests event query PM metadata for 1 circuit_name, 1 systems, 0 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_duration(t_start=1544622149620000000, duration=[(10, 's'), (400, 's')]) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A12', system='PC') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_eq/RQD.A12_PC_1_SOURCE.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_eq_metadata_rba12_ee(self):
        """Tests event query PM metadata for 1 circuit_name, 1 systems, 2 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_duration(t_start=1544622149620000000, duration=[(10, 's'), (400, 's')]) \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A12', system=['EE_ODD', 'EE_EVEN']) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_eq/RB.A12_EE.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_eq_metadata_rqda12_qps_qh_wildcard_source(self):
        """Tests event query PM metadata for 1 circuit_name, 1 systems, 0 sources, 0 wildcards"""
        # arrange
        # act
        # assert
        with self.assertRaises(KeyError) as context:
            QueryBuilder().with_pm() \
                .with_duration(timestamp=1544622149620000000, duration=[(10, 's'), (400, 's')]) \
                .with_circuit_type('RQ') \
                .with_metadata(circuit_name='RQD.A12', system='QH')

        self.assertTrue('source' in context.exception.args[0])

        # assert time definition

    def test_eq_metadata_rqda12_qps_qh_16l2_source(self):
        """Tests event query PM metadata for 1 circuit_name, 1 systems, 1 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_duration(t_start=1544622149620000000, duration=[(10, 's'), (400, 's')]) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A12', system='QH', source='16L2') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_eq/RQD.A12_QH_16L2_SOURCE.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition
    def test_eq_metadata_rqda12_qps_qh_no_source(self):
        """Tests event query PM metadata for 1 circuit_name, 1 systems, 0 sources, 0 wildcards"""
        # arrange
        # act
        with self.assertRaises(KeyError) as context:
            QueryBuilder().with_pm() \
                .with_duration(t_start=1544622149620000000, duration=[(10, 's'), (400, 's')]) \
                .with_circuit_type('RQ') \
                .with_metadata(circuit_name='RQD.A12', system='QH')

        # assert
        self.assertTrue('source' in context.exception.args[0])

    def test_eq_metadata_rqda12_rqfa12_pc_fgc_1_source(self):
        """Tests event query PM metadata for 2 circuit_name, 1 systems, 0 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_duration(timestamp=1544622149620000000, duration=[(10, 's'), (400, 's')]) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name=['RQD.A12', 'RQF.A12'], system='PC') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_eq/RQD.A12_RQF.A12_PC_1_SOURCE.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_eq_metadata_rqda12_rqfa12_qps_qh_wildcard_source(self):
        """Tests event query PM metadata for 2 circuit_name, 1 systems, 0 sources, 0 wildcards"""
        # arrange
        # act
        # assert
        with self.assertRaises(KeyError) as context:
            QueryBuilder().with_pm() \
                .with_duration(t_start=1544622149620000000, duration=[(10, 's'), (400, 's')]) \
                .with_circuit_type('RQ') \
                .with_metadata(circuit_name=['RQD.A12', 'RQF.A12'], system='QH')

        self.assertTrue('source' in context.exception.args[0])

    # signal_query
    # 1..N circuit_names
    # 1..N systems
    # 1..N signals
    # 1..N sources
    # 1..N wildcards

    # PM: signal_query
    # 1..N circuit names
    # 1 system
    # 1..N signals
    # 1..N sources
    # 1..N wildcards
    def test_sq_metadata_rqda12_qps_qh_u_hds_16l2_source(self):
        """Tests PM metadata for 1 circuit_name, 1 systems, 1 signals, 1 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(timestamp=1544622149620000000) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A12', system='QH', source='16L2',
                           signal='U_HDS', wildcard={'CELL': '16L2'}) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/RQD.A12_QH_16L2_SOURCE.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_sq_metadata_rqda12_qps_qh_i_hds_16l2_source(self):
        """Tests PM metadata for 1 circuit_name, 1 systems, 1 signals, 1 sources, 0 wildcards"""
        # arrange
        # act
        with self.assertRaises(KeyError) as context:
            QueryBuilder().with_pm() \
                .with_timestamp(timestamp=1544622149620000000) \
                .with_circuit_type('RQ') \
                .with_metadata(circuit_name='RQD.A12', system='QH', source='16L2',
                               signal='I_HDS', wildcard={'CELL': '16L2'}) \
                .signal_query()

        # assert
        # test should fail for timestamp before 2020-01-01 00:00:00
        self.assertTrue('I_HDS' in context.exception.args[0])

        # assert time definition

    def test_sq_metadata_rqda12_qps_qh_u_hds_16l2_source_run3(self):
        """Tests PM metadata for 1 circuit_name, 1 systems, 1 signals, 1 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(timestamp=1577833300000000000) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A12', system='QH', source='16L2',
                           signal='U_HDS', wildcard={'CELL': '16L2'}) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/RQD.A12_QH_16L2_SOURCE.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_sq_metadata_rqda12_qps_qh_i_hds_16l2_source_run3(self):
        """Tests PM metadata for 1 circuit_name, 1 systems, 1 signals, 1 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(timestamp=1577833300000000000) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A12', system='QH', source='16L2',
                           signal='I_HDS', wildcard={'CELL': '16L2'}) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/RQD.A12_QH_16L2_SOURCE_I_HDS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_sq_metadata_rqda12_pm_fgc_i_meas(self):
        """Tests PM metadata for 1 circuit_name, 1 systems, 1 signals, 0 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(1544622149620000000) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A12', system='PC', signal='I_MEAS') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/RQD.A12_PC_I_MEAS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_sq_metadata_rqda12_rqfa12_pm_fgc_i_meas(self):
        """Tests PM metadata for 2 circuit_name, 1 systems, 1 signals, 0 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(1544622149620000000) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name=['RQD.A12', 'RQF.A12'], system='PC', signal='I_MEAS') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/RQD.A12_RQF.A12_PC_I_MEAS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_sq_metadata_rqda12_pm_qds_u_1_ext_u_2_ext(self):
        """Tests PM metadata for 1 circuit_names, 1 systems, 2 signals, 1 sources, 1 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm()\
            .with_timestamp(1544622149598000000) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A12', system='QDS', signal=['U_1_EXT', 'U_2_EXT'],
                           source='16L2', wildcard={'CELL': '16L2'})\
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/RQD.A12_QDS_U_1_EXT_U_2_EXT.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_sq_metadata_rba12_pm_leads_even_leads_odd_u_hts(self):
        """Tests PM metadata for 1 circuit_names, 2 systems, 1 signals, 1 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm()\
            .with_timestamp(1544622149598000000) \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A12', system=['LEADS_EVEN', 'LEADS_ODD'], signal='U_HTS')\
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/RB.A12_LEADS_EVEN_LEADS_ODD_U_HTS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_sq_metadata_rba12_pm_leads_even_leads_odd_u_hts_u_res(self):
        """Tests PM metadata for 1 circuit_names, 2 systems, 2 signals, 0 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(1544622149598000000) \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A12', system=['LEADS_EVEN', 'LEADS_ODD'], signal=['U_HTS', 'U_RES']) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/RB.A12_LEADS_EVEN_LEADS_ODD_U_HTS_U_RES.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_sq_metadata_rba12_rba23_pm_leads_even_leads_odd_u_hts_u_res(self):
        """Tests PM metadata for 2 circuit_names, 2 systems, 2 signals, 0 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(1544622149598000000) \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name=['RB.A12', 'RB.A23'], system=['LEADS_EVEN', 'LEADS_ODD'],
                           signal=['U_HTS', 'U_RES'])\
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/RB.A12_RB.A23_LEADS_EVEN_LEADS_ODD_U_HTS_U_RES.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    # NXCALS: signal_query
    # 1..N circuit names
    # 1 system
    # 1..N signals
    # 1..N sources
    # 1..N wildcards

    def test_metadata_nxcals_signal_query_i_meas_single_circuit(self):
        """Tests CALS metadata for 1 circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""
        # arrange
        t_start = '2018-12-12 00:00:00+01:00'
        t_end = '2018-12-13 00:00:00+01:00'
        circuit_type = 'RQ'
        circuit_name = 'RQD.A12'
        signal = 'I_MEAS'
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=t_start, t_end=t_end) \
            .with_circuit_type(circuit_type) \
            .with_metadata(circuit_name=circuit_name, system='PC', signal=signal) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_sq/RQD.A12_I_MEAS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_metadata_nxcals_signal_query_rqd_i_meas(self):
        """Tests CALS metadata for N circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""
        # arrange
        circuit_name = [cn for cn in SignalMetadata.get_circuit_names('RQ') if 'D' in cn]
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name=circuit_name, system='PC', signal='I_MEAS') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_sq/RQD_I_MEAS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_metadata_nxcals_signal_query_i_meas_single_circuit_many_pcs(self):
        """Tests CALS metadata for 1 circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""
        # arrange
        t_start = '2017-04-11 20:38:19.627'
        t_end = '2017-04-11 22:24:50.702'
        circuit_type = 'IT'
        circuit_name = 'RQX.L1'
        signal = 'I_MEAS'
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=t_start, t_end=t_end) \
            .with_circuit_type(circuit_type) \
            .with_metadata(circuit_name=circuit_name, system='PC', signal=signal) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_sq/RQX.L1_I_MEAS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    # feature_query
    # 1..N circuit_names
    # 1..N systems
    # 1..N signals
    # 1..N sources
    # 1..N wildcards

    def test_metadata_nxcals_feature_query_rb_i_meas(self):
        """Tests NXCALS metadata for N circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""

        # arrange
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='*', system='PC', signal='I_MEAS') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_fq/RB_I_MEAS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_metadata_nxcals_feature_query_rba12_u_res(self):
        """Tests NXCALS metadata for 1 circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""

        # arrange
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A12', system='BUSBAR', signal='U_RES', wildcard={'BUSBAR': '*'}) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_fq/RB.A12_U_RES.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_metadata_nxcals_feature_query_rqda12_u_res(self):
        """Tests NXCALS metadata for 1 circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""

        # arrange
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A12', system='BUSBAR', signal='U_RES', wildcard={'BUSBAR': '*'}) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_fq/RQD.A12_U_RES.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_metadata_nxcals_feature_query_rqfa12_u_res(self):
        """Tests NXCALS metadata for 1 circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""

        # arrange
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQF.A12', system='BUSBAR', signal='U_RES', wildcard={'BUSBAR': '*'}) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_fq/RQF.A12_U_RES.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_metadata_nxcals_feature_query_rqda12_rqfa12_u_res(self):
        """Tests NXCALS metadata for 1 circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""

        # arrange
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name=['RQD.A12', 'RQF.A12'], system='BUSBAR',
                           signal='U_RES', wildcard={'BUSBAR': '*'}) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_fq/RQD.A12_RQF.A12_U_RES.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_metadata_nxcals_feature_query_rb_u_res(self):
        """Tests NXCALS metadata for N circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""

        # arrange
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='*', system='BUSBAR', signal='U_RES', wildcard={'BUSBAR': '*'}) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_fq/RB_U_RES.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_query_parameters_nxcals_feature_query_rqd_u_res(self):
        """Query NXCALS query parameters for 1 circuit_names, 1 systems, 3 signal, 0 sources, 0 wildcards"""

        # arrange
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2017-04-24 19:29:45.679', t_end='2017-04-24 19:51:14.809') \
            .with_query_parameters(nxcals_system='WINCCOA',
                                   signal=['DCQFD.7L2.R:U_RES', 'DCQFB.C21R1.R:U_RES', 'DCQFQ.12L2.L:U_RES']) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_fq/RQ_U_RES.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def explode(query_params: pd.DataFrame, column: str):
        import numpy as np
        if any([row_type == list for row_type in query_params[column].apply(type).values]):
            return query_params.reindex(query_params.index.repeat(query_params[column].str.len())) \
                .assign(**{column: np.concatenate(query_params[column].values)}) \
                .reset_index(drop=True)
        else:
            return query_params

    # ======================= MOCK
    # ======================================================
    def test_metadata_nxcals_feature_query_rb_i_meas_fill_missing_rows(self):
        """Tests pyeDSL feature query with fill_missing_rows function,
            NXCALS metadata for N circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards
        """
        # arrange
        spark = None
        t_start='2018-12-12 00:00:00+01:00'
        t_end='2018-12-13 00:00:00+01:00'
        circuit_type = 'RB'
        api_mock = read_csv('query_builder/nxcals_fq/RB_I_MEAS_Missing_Rows.csv')

        with patch.object(NxcalsDbQueryBuilder, 'nxcals_feature_query') as mock_read:
            mock_read.return_value = api_mock
            # act
            query_params = QueryBuilder().with_nxcals(spark) \
                .with_duration(t_start=t_start, t_end=t_end) \
                .with_circuit_type(circuit_type) \
                .with_metadata(circuit_name='*', system='PC', signal='I_MEAS') \
                .feature_query(['mean', 'std', 'max', 'min', 'count']) \
                .df

            # assert
            metadata_ref = read_csv('query_builder/nxcals_fq/RB_I_MEAS_Missing_Rows_Filled.csv')
            pd.testing.assert_frame_equal(metadata_ref, query_params)


    def test_metadata_nxcals_feature_query_rb_u_mag_fill_missing_rows_class(self):
        """Tests pyeDSL feature query with fill_missing_rows function,
            NXCALS metadata for N circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards
        """
        # arrange
        spark = None
        t_start_injs = 1526898157236000000
        t_end_sbs = 1526903352338000000
        circuit_type = 'RB'
        translate_udf = None
        api_mock = read_csv('query_builder/nxcals_fq/RB_U_MAG_Missing_Rows.csv')

        with patch.object(NxcalsDbQueryBuilder, 'nxcals_feature_query') as mock_read:
            mock_read.return_value = api_mock
            # act
            query_params = QueryBuilder().with_nxcals(spark) \
                .with_duration(t_start=t_start_injs, t_end=t_end_sbs) \
                .with_circuit_type(circuit_type) \
                .with_metadata(circuit_name=['RB.A12','RB.A23', 'RB.A45'], system='BUSBAR', signal='U_MAG',
                               wildcard={'BUSBAR': ['DCBB.8L2.R', 'DCBA.8R2.R', 'DCBA.15R4.R']}) \
                .feature_query(['mean', 'max', 'min'], function=translate_udf) \
                .correct_voltage_sign() \
                .df

            # assert
            metadata_ref = read_csv('query_builder/nxcals_fq/RB_U_MAG_Missing_Rows_Filled.csv')
            pd.testing.assert_frame_equal(metadata_ref, query_params)



    def test_metadata_nxcals_feature_query_rb_i_meas_features(self):
        """Tests pyeDSL feature query with fill_missing_rows function,
            NXCALS metadata for N circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards
        """

        # arrange
        spark = None
        t_start='2018-12-12 00:00:00+01:00'
        t_end='2018-12-13 00:00:00+01:00'
        circuit_type = 'RB'
        api_mock = read_csv('query_builder/nxcals_fq/RB_I_MEAS_Feature_RBA45.csv')

        with patch.object(NxcalsDbQueryBuilder, 'execute_feature_query') as mock_read:
            mock_read.return_value = api_mock

            # act
            query_params = QueryBuilder().with_nxcals(spark) \
                .with_duration(t_start=t_start, t_end=t_end) \
                .with_circuit_type(circuit_type) \
                .with_metadata(circuit_name='*', system='PC', signal='I_MEAS') \
                .feature_query(['mean', 'std', 'max', 'min', 'count']) \
                .df

            mean_qp_0 = query_params['mean'][0]
            mean_qp_1 = query_params['mean'][1]
            mean_qp_2 = query_params['mean'][2]

            # assert
            metadata_ref = read_csv('query_builder/nxcals_fq/RB_I_MEAS_Raw_No_Feature_RBA45.csv')
            i_meas_df = metadata_ref

            #i_meas_df.index[0]
            list_0 = []
            list_1 = []
            list_2 = []

            t_start_injs = 1526898157236000000
            t_end_injs = 1526899957236000000
            t_start_sbs = 1526901552338000000
            t_end_sbs = 1526903352338000000

            for i in range(len(i_meas_df)):
                timestamp = i_meas_df.index[i]

                if (timestamp >= t_start_injs and timestamp < t_end_injs):
                    list_1.append(i_meas_df['I_MEAS'][timestamp])

                elif (timestamp >= t_end_injs and timestamp < t_start_sbs):
                    list_0.append(i_meas_df['I_MEAS'][timestamp])

                elif (timestamp >= t_start_sbs and timestamp <= t_end_sbs):
                    list_2.append(i_meas_df['I_MEAS'][timestamp])
                else:
                    print(i)

            mean_ref_0 = sum(list_0) / len(list_0)
            mean_ref_1 = sum(list_1) / len(list_1)
            mean_ref_2 = sum(list_2) / len(list_2)

            self.assertAlmostEqual(mean_qp_0, mean_ref_0)
            self.assertAlmostEqual(mean_qp_1, mean_ref_1)
            self.assertAlmostEqual(mean_qp_2, mean_ref_2)


    def test_metadata_nxcals_feature_query_rb_u_mag_correct_voltage_sign(self):
        """Tests pyeDSL feature query with fill_missing_rows function,
            NXCALS metadata for N circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards
        """

        # arrange
        spark = None
        t_start_injs = 1526898157236000000
        t_end_sbs = 1526903352338000000
        circuit_type = 'RB'
        translate_udf = None

        api_mock = read_csv('query_builder/nxcals_fq/RB_U_MAG_RAW_Voltage_Sign_RB12_RB45.csv')

        with patch.object(NxcalsDbQueryBuilder, 'execute_feature_query') as mock_read:
            mock_read.return_value = api_mock
                # act
            query_params = QueryBuilder().with_nxcals(spark) \
                .with_duration(t_start=t_start_injs, t_end=t_end_sbs) \
                .with_circuit_type(circuit_type) \
                .with_metadata(circuit_name=['RB.A12', 'RB.A45'], system='BUSBAR', signal='U_MAG', wildcard={'BUSBAR': ['DCBB.8L2.R', 'DCBA.15R4.R']}) \
                .feature_query(['mean', 'max', 'min'], function=translate_udf) \
                .correct_voltage_sign() \
                .df

            # assert
            metadata_ref = read_csv('query_builder/nxcals_fq/RB_U_MAG_Corrected_Voltage_Sign_RB12_RB45.csv')
            pd.testing.assert_frame_equal(metadata_ref, query_params)


    def test_metadata_nxcals_feature_query_rb_u_mag_correct_voltage_sign_All(self):
        """Tests pyeDSL feature query with fill_missing_rows function,
            NXCALS metadata for N circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards
        """

        # arrange
        spark = None
        t_start_injs = 1526898157236000000
        t_end_sbs = 1526903352338000000
        circuit_type = 'RB'
        translate_udf = None

        api_mock = read_csv('query_builder/nxcals_fq/RB_U_MAG_RAW_Voltage_Sign_All.csv')

        with patch.object(NxcalsDbQueryBuilder, 'execute_feature_query') as mock_read:
            mock_read.return_value = api_mock
            # act
            query_params = QueryBuilder().with_nxcals(spark) \
                .with_duration(t_start=t_start_injs, t_end=t_end_sbs) \
                .with_circuit_type(circuit_type) \
                .with_metadata(circuit_name='*', system='BUSBAR', signal='U_MAG', wildcard={'BUSBAR': '*'}) \
                .feature_query(['mean', 'max', 'min'], function=translate_udf) \
                .correct_voltage_sign() \
                .df

        # assert
        metadata_ref = read_csv('query_builder/nxcals_fq/RB_U_MAG_Corrected_Voltage_Sign_All.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

    # FeatureProcessing
    def test_feature_processing_sort_busbar_location_warning(self):
        # arrange
        busbar_df = pd.DataFrame(columns=['R_MAG'])

        # act
        # assert
        with warnings.catch_warnings(record=True) as w:
            FeatureProcessing(busbar_df).sort_busbar_location(circuit_type='RB', circuit_name='RB.A12')
            if w:
                self.assertEquals('This operation is only supported to feature DataFrame with nxcals_variable_name '
                                  'column, got Index([\'R_MAG\'], dtype=\'object\')', str(w[0].message))

    def test_feature_processing_sort_busbar_calculate_max_abs(self):
        # arrange
        busbar_df = pd.DataFrame(list(zip([-1, 0, 1, 2], [0, 0, 1, 2])), columns=['min', 'max'])

        # act
        max_abs_act_df = FeatureProcessing(busbar_df).calculate_max_abs().df

        # assert
        max_abs_ref_df = pd.DataFrame(list(zip([-1, 0, 1, 2], [0, 0, 1, 2], [1, 0, 1, 2])),
                                      columns=['min', 'max', 'max_abs'])
        pd.testing.assert_frame_equal(max_abs_act_df, max_abs_ref_df)

    def test_feature_processing_sort_busbar_calculate_max_abs_warning(self):
        # arrange
        busbar_df = pd.DataFrame(columns=['R_MAG'], data=[-1, 0, 1, 2])

        # act
        with warnings.catch_warnings(record=True) as w:
            FeatureProcessing(busbar_df).calculate_max_abs()
            if w:
                self.assertEquals('This operation is only supported to feature DataFrame with min, max columns'
                                  ', got Index([\'R_MAG\'], dtype=\'object\')', str(w[0].message))

    def test_feature_processing_sort_values(self):
        # arrange
        busbar_df = pd.DataFrame(list(zip([4, 0, 1, 2], [0, 0, 1, 2])), columns=['min', 'max'])

        # act
        max_abs_act_df = FeatureProcessing(busbar_df).sort_values(by='min').df

        # assert
        max_abs_ref_df = pd.DataFrame(list(zip([0, 1, 2, 4], [0, 1, 2, 0])),
                                      columns=['min', 'max'])
        pd.testing.assert_frame_equal(max_abs_act_df, max_abs_ref_df)

    def test_feature_processing_convert_to_row(self):
        # arrange
        busbar_df = pd.DataFrame(list(zip(['R_1', 'R_2', 'R_3', 'R_4'], [1,2,3,4])),
                                 columns=['nxcals_variable_name', 'R_MAG'])

        # act
        max_abs_act_df = FeatureProcessing(busbar_df).convert_into_row().df

        # assert
        max_abs_ref_df = pd.DataFrame({'R_1_R_MAG': {0: 1}, 'R_2_R_MAG': {0: 2},
                                       'R_3_R_MAG': {0: 3}, 'R_4_R_MAG': {0: 4}})
        pd.testing.assert_frame_equal(max_abs_act_df, max_abs_ref_df)

    def test_feature_processing_convert_to_row_warning(self):
        # arrange
        busbar_df = pd.DataFrame(columns=['R_MAG'], data=[-1, 0, 1, 2])

        # act
        # assert
        with warnings.catch_warnings(record=True) as w:
            FeatureProcessing(busbar_df).convert_into_row()
            if w:
                self.assertEquals('This operation is only supported to feature DataFrame with nxcals_variable_name '
                                  'column, got Index([\'R_MAG\'], dtype=\'object\')', str(w[0].message))


    def test_signal_processing_map_values(self):
        # arrange
        bm_df = pd.DataFrame(columns=['BEAMMODE'], data=[1, 2, 1, 3])

        # act
        bm_mapped_df = SignalProcessing([bm_df], 'NXCALS') \
                        .map_values({1: 'INJECTION', 2: 'RAMPUP', 3: 'STABLEBEAMS'}).dfs[0]

        # assert
        bm_mapped_ref_df = pd.DataFrame(columns=['BEAMMODE'], data=['INJECTION', 'RAMPUP', 'INJECTION', 'STABLEBEAMS'])
        pd.testing.assert_frame_equal(bm_mapped_ref_df, bm_mapped_df)

    def test_signal_processing_filter_values(self):
        # arrange
        bm_df = pd.DataFrame(columns=['BEAMMODE'], data=[1, 2, 1, 3])

        # act
        bm_mapped_df = SignalProcessing([bm_df], 'NXCALS') \
                        .filter_values([1, 2]).dfs[0]

        # assert
        bm_mapped_ref_df = pd.DataFrame(columns=['BEAMMODE'], data=[1,2,1])
        pd.testing.assert_frame_equal(bm_mapped_ref_df, bm_mapped_df)

    def test_signal_processing_create_col_from_index(self):
        # arrange
        bm_df = pd.DataFrame(columns=['BEAMMODE'], data=[1, 2, 1, 3])

        # act
        bm_new_col_df = SignalProcessing([bm_df], 'NXCALS') \
                        .create_col_from_index('INDEX').dfs[0]

        # assert
        bm_new_col_ref_df = pd.DataFrame({'BEAMMODE': {0: 1, 1: 2, 2: 1, 3: 3},
                                         'INDEX': {0: 0, 1: 1, 2: 2, 3: 3}})
        pd.testing.assert_frame_equal(bm_new_col_ref_df, bm_new_col_df)


    def test_signal_processing_overwrite_sampling_time(self):
        # arrange
        bm_df = pd.DataFrame(columns=['BEAMMODE'], data=[1, 2, 1, 3])

        # act
        bm_new_index_df = SignalProcessing([bm_df], 'NXCALS') \
                        .overwrite_sampling_time(0.25, 1).dfs[0]

        # assert
        bm_new_index_ref_df = pd.DataFrame({'BEAMMODE': {0.0: 1, 0.25: 2, 0.5: 1, 0.75: 3}})
        pd.testing.assert_frame_equal(bm_new_index_ref_df, bm_new_index_df)

    # EventProcessing
    def test_event_processing_filter_source(self):
        # arrange
        ep = EventProcessing(pd.DataFrame())

        # act
        ep_sorted = ep.filter_source(circuit_name='RB.A12', system='QH')

        # assert
        self.assertEqual(ep, ep_sorted)

    def test_event_processing_filter_source_it(self):
        # arrange
        ep = EventProcessing(pd.DataFrame(list(zip(['RQX.R1'], [1])), columns=['source', 'timestamp']))

        # act
        ep_sorted = ep.filter_source(circuit_name='RQX.R1', system='QH')

        # assert
        self.assertEqual(ep, ep_sorted)

    def test_event_processing_drop_duplicate_source(self):
        # arrange
        ep = EventProcessing(pd.DataFrame())

        # act
        ep_dropped = ep.drop_duplicate_source()

        # assert
        self.assertEqual(ep, ep_dropped)

    def test_event_processing_sort_values(self):
        # arrange
        ep = EventProcessing(pd.DataFrame())

        # act
        ep_sorted = ep.sort_values(by='')

        # assert
        self.assertEqual(ep, ep_sorted)
