import json
import os
import unittest
from pathlib import Path
from unittest import mock
from unittest.mock import patch
import requests

from lhcsmapi.pyedsl.dbsignal.post_mortem.PmDbSignal import PmDbSignal as signal_pm, PmDbSignal
import pandas as pd
from datetime import datetime


def mocked_pmdbsignal_requests_get(*args, **kwargs):
    path = Path(os.path.dirname(__file__))

    class MockedRequest(object):
        def __init__(self, text):
            self.text = text

    if args and 'STATUS.I_MEAS' in args[0]:
        json_path = os.path.join(path.parent.parent.parent,
                                 'resources/dbsignal/post_mortem/'
                                 '1426220469520000000_FGC_51_self_pmd_RPTE.UA47.RB.A45_STATUS.I_MEAS.json')
    elif args and 'IEARTH.IEARTH' in args[0]:
        json_path = os.path.join(path.parent.parent.parent,
                                 'resources/dbsignal/post_mortem/'
                                 '1426220469520000000_FGC_51_self_pmd_RPTE.UA47.RB.A45_IEARTH.IEARTH.json')
    elif args and 'IEARTH.IEARTH_ERROR' in args[0]:
        json_path = os.path.join(path.parent.parent.parent,
                                 'resources/dbsignal/post_mortem/'
                                 '1426220469520000000_FGC_51_self_pmd_RPTE.UA47.RB.A45_IEARTH.IEARTH_ERROR.json')
    elif args and 'I_HDS_4' in args[0]:
        json_path = os.path.join(path.parent.parent.parent,
                                 'resources/dbsignal/post_mortem/'
                                 '1426220469491000000_QPS_DQAMCNMB_PMHSU_B20L5_I_HDS_4.json')
    else:
        json_path = ''
    with open(json_path) as json_file:
        return MockedRequest(json.dumps(json.load(json_file)))


class TestPmDbSignal(unittest.TestCase):

    def setUp(self):
        self.SYSTEM = "QPS"
        self.CLASS_NAME = "DQAMCNMB_PMHSU"
        self.SOURCE = "B20L5"
        self.SIGNAL_NAME = "I_HDS_4"
        self.TIMESTAMP = 1426220469491000000
        self.defaultSignalPM = signal_pm()

    def tearDown(self):
        pass

    def assertEqualSignals(self, firstSignalPM, secondSignalPM):
        """Assert each property of firstSignalPM against of the same property of secondSignalPM"""
        self.assertEqual(firstSignalPM.system, secondSignalPM.system)
        self.assertEqual(firstSignalPM.className, secondSignalPM.className)
        self.assertEqual(firstSignalPM.source, secondSignalPM.source)
        self.assertEqual(firstSignalPM.signal, secondSignalPM.signal)
        self.assertEqual(firstSignalPM.eventTime, secondSignalPM.eventTime)

    def test_toString(self):
        # arrange
        pms = signal_pm().build_signal(system=self.SYSTEM, className=self.CLASS_NAME, source=self.SOURCE,
                                       signal=self.SIGNAL_NAME, eventTime=self.TIMESTAMP)
        # act
        pms_string_act = str(pms)

        # assert
        pms_string_exp = 'Db: PM, signal: I_HDS_4, eventTime = 1426220469491000000 (2015-03-13 05:21:09.491000+01:00), ' \
                       'system=QPS, source=B20L5, className=DQAMCNMB_PMHSU'
        self.assertEqual(pms_string_exp, pms_string_act)

    def test_createSignalPM_noParameters_defaultSignalCreated(self):
        """Tests that the default CalsSignal is created if no arguments are passed to the createSignalPM method"""

        # act:
        pm_signal = signal_pm.build_signal()

        # assert:
        self.assertEqualSignals(pm_signal, self.defaultSignalPM)

    def test_createSignalPM_initParametersProvided_CorrectSignalCreated(self):
        """Tests that the correct CalsSignal is created if arbitrary arguments are passed to the createSignalPM method"""

        # arrange:
        now = datetime.now()
        arbitrary_pm_signal = signal_pm("System", "className", "source", "test", pd.Timestamp(now))

        # act:
        pm_signal = signal_pm.build_signal(system="System", className="className", source="source", signal="test",
                                           eventTime=pd.Timestamp(now))

        # assert:
        self.assertEqualSignals(pm_signal, arbitrary_pm_signal)

    def test_createSignalPM_parameterIsInvalidType_TypeErrorIsThrown(self):
        """Tests that when kwargs contain a username keyword that isn't a string, a TypeError
        is thrown"""

        now = datetime.now()

        with self.assertRaises(TypeError):
            signal_pm.build_signal(system=1000, className="className", source="source", signal="test",
                                   eventTime=pd.Timestamp(now))

    def test_getNamesOfValues_correctDBAccessWriteCalled(self):
        """Test that write correctly call the write method of the correct DbSignal type dependent on the api parameter"""

        # arrange:
        uut = signal_pm()

        test_cases = [("zero", {'namesAndValues': []}, []),
                     ('one', {'namesAndValues': [{'name': 'name1'}]}, ['name1']),
                     ('many', {'namesAndValues': [{'name': 'name1'}, {'name': 'name2'}]}, ['name1', 'name2'])]

        for test_name, content, valueNames in test_cases:
            with self.subTest(name=test_name):

                # act:
                result = uut._get_names_of_value(content)

                # assert:
                self.assertEqual(result, valueNames)

    def test_parseSignal_queryingFailed_RaisesError(self):

        pm_signal = signal_pm("System", "className", "source", "test", pd.Timestamp(datetime.now()))

        with patch.object(requests, 'get') as mock_get:
            mock_get.side_effect = Exception()

            with self.assertRaises(Exception):
                pm_signal._parse_signal()

    @mock.patch('requests.get', side_effect=mocked_pmdbsignal_requests_get)
    def test_read_fgc_i_meas(self, mock_get):
        # arrange

        # act
        i_meas_df = PmDbSignal(system='FGC', className='51_self_pmd', source='RPTE.UA47.RB.A45', signal='STATUS.I_MEAS',
                               eventTime=1426220469520000000).read()

        # assert
        i_meas_df_ref = pd.DataFrame({'STATUS.I_MEAS': {1426220409500000000: 9571.501,
                                                        1426220409520000000: 9571.699,
                                                        1426220409540000000: 9571.899,
                                                        1426220409560000000: 9572.1,
                                                        1426220409580000000: 9572.3}})
        self.assertEqual(len(mock_get.call_args_list), 0)
        pd.testing.assert_frame_equal(i_meas_df_ref, i_meas_df.head())

    @mock.patch('requests.get', side_effect=mocked_pmdbsignal_requests_get)
    def test_read_fgc_i_meas_allow_duplicate_index(self, mock_get):
        # arrange

        # act
        i_meas_df = PmDbSignal(system='FGC', className='51_self_pmd', source='RPTE.UA47.RB.A45', signal='STATUS.I_MEAS',
                               eventTime=1426220469520000000).read(allow_duplicate_index=True)

        # assert
        i_meas_df_ref = pd.DataFrame({'STATUS.I_MEAS': {1426220409500000000: 9571.501,
                                                        1426220409520000000: 9571.699,
                                                        1426220409540000000: 9571.899,
                                                        1426220409560000000: 9572.1,
                                                        1426220409580000000: 9572.3}})
        self.assertEqual(len(mock_get.call_args_list), 0)
        pd.testing.assert_frame_equal(i_meas_df_ref, i_meas_df.head())

    @mock.patch('requests.get', side_effect=mocked_pmdbsignal_requests_get)
    def test_read_fgc_iearth(self, mock_get):
        # arrange

        # act
        i_meas_df = PmDbSignal(system='FGC', className='51_self_pmd', source='RPTE.UA47.RB.A45', signal='IEARTH.IEARTH',
                               eventTime=1426220469520000000).read()

        # assert
        i_meas_df_ref = pd.DataFrame({'IEARTH.IEARTH': {1426220459260000000: -0.009606788,
                                                        1426220459280000000: -0.009633326,
                                                        1426220459300000000: -0.009633326,
                                                        1426220459320000000: -0.009633326,
                                                        1426220459340000000: -0.00958025}})
        self.assertEqual(len(mock_get.call_args_list), 1)
        pd.testing.assert_frame_equal(i_meas_df_ref, i_meas_df.head())

    @mock.patch('requests.get', side_effect=mocked_pmdbsignal_requests_get)
    def test_read_fgc_iearth_error(self, mock_get):
        # arrange

        # act
        i_meas_df = PmDbSignal(system='FGC', className='51_self_pmd', source='RPTE.UA47.RB.A45',
                               signal='IEARTH.IEARTH_ERROR', eventTime=1426220469520000000).read()

        # assert
        i_meas_df_ref = pd.DataFrame()
        pd.testing.assert_frame_equal(i_meas_df_ref, i_meas_df.head())

    @mock.patch('requests.get', side_effect=mocked_pmdbsignal_requests_get)
    def test_read_qps(self, mock_get):
        # arrange

        # act
        i_meas_df = PmDbSignal(system='QPS', className='DQAMCNMB_PMHSU', source='B20L5', signal='I_HDS_4',
                               eventTime=1426220469491000000).read()

        # assert
        i_meas_df_ref = pd.DataFrame({'I_HDS_4': {1426220469491020832: 0.0896478,
                                                  1426220469491026040: 0.0915552,
                                                  1426220469491031248: 0.0915552,
                                                  1426220469491036456: 0.0934626,
                                                  1426220469491041664: 0.0915552}})

        pd.testing.assert_frame_equal(i_meas_df_ref, i_meas_df.head())

    @mock.patch('requests.get', side_effect=mocked_pmdbsignal_requests_get)
    def test_read_qps_missing_json(self, mock_get):
        # arrange

        # act
        i_meas_df = PmDbSignal(system='QPS', className='DQAMCNMB_PMHSU', source='B20L5', signal='I_HDS_3',
                               eventTime=1426220469491000000).read()

        # assert
        i_meas_df_ref = pd.DataFrame()
        self.assertEqual(len(mock_get.call_args_list), 1)

        pd.testing.assert_frame_equal(i_meas_df_ref, i_meas_df.head())

