import unittest

import pandas as pd

from lhcsmapi.reference.Reference import Reference


class TestReference(unittest.TestCase):

    def test_get_quench_heater_reference_discharges_RB(self):
        # arrange
        cell = "A8R1"
        circuit_type = 'RB'

        # act
        ref_discharge_df = Reference.read_quench_heater_reference_discharges(circuit_type)
        timestamp_act = ref_discharge_df[ref_discharge_df["Cell"] == cell]["Timestamp"].values[0]

        # assert
        timestamp_exp = 1417014253613000000
        self.assertEqual(timestamp_exp, timestamp_act)

    def test_get_quench_heater_reference_discharges_RQ(self):
        # arrange
        cell = "16L2"
        circuit_type = 'RQ'

        # act
        ref_discharge_df = Reference.read_quench_heater_reference_discharges(circuit_type)
        timestamp_act = ref_discharge_df[ref_discharge_df["Cell"] == cell]["Timestamp"].values[0]

        # assert
        timestamp_exp = 1417079205409000000
        self.assertEqual(timestamp_exp, timestamp_act)

    def test_get_quench_heater_reference_discharge_RQ(self):
        # arrange
        cell = "16L2"
        circuit_type = 'RQ'

        # act
        timestamp_act = Reference.get_quench_heater_reference_discharge(circuit_type, cell)

        # assert
        timestamp_exp = 1417079205409000000
        self.assertEqual(timestamp_exp, timestamp_act)

    def test_get_quench_heater_reference_discharge_error(self):
        # arrange
        cell = ""
        circuit_type = '600A'

        # act
        with self.assertRaises(ValueError) as context:
            Reference.get_quench_heater_reference_discharge(circuit_type, cell)

        # assert
        self.assertEqual('Circuit type 600A not supported!', str(context.exception))

    def test_read_energy_extraction_reference_features_rb(self):
        # arrange
        circuit_type = 'RB'

        # act
        features_df_act = Reference.read_energy_extraction_reference_features(circuit_type)

        # assert
        features_df_ref = pd.DataFrame({'min': {'R_odd': 0.0675, 'R_even': 0.0675, 't_delay_ee_odd': 0.05,
                                                't_delay_ee_even': 0.55, 'tau_u_dump_res_odd': 110.0,
                                                'tau_u_dump_res_even': 110.0},
                                        'max': {'R_odd': 0.0825, 'R_even': 0.0825, 't_delay_ee_odd': 0.15,
                                                't_delay_ee_even': 0.65, 'tau_u_dump_res_odd': 130.0,
                                                'tau_u_dump_res_even': 130.0}})
        pd.testing.assert_frame_equal(features_df_ref, features_df_act)

    def test_read_energy_extraction_reference_features_error(self):
        # arrange
        circuit_type = '600A'

        # act
        with self.assertRaises(KeyError) as context:
            Reference.read_energy_extraction_reference_features(circuit_type)

        # assert
        self.assertEqual('\'Circuit type 600A not supported!\'', str(context.exception))

    def test_read_power_converter_reference_features_rb(self):
        # arrange
        circuit_type = 'RB'

        # act
        features_df_act = Reference.read_power_converter_reference_features(circuit_type)

        # assert
        features_df_ref = pd.DataFrame({'min': {'tau_i_meas': 90}, 'max': {'tau_i_meas': 110}})
        pd.testing.assert_frame_equal(features_df_ref, features_df_act)

    def test_read_power_converter_reference_features_error(self):
        # arrange
        circuit_type = 'IT'

        # act
        with self.assertRaises(ValueError) as context:
            Reference.read_power_converter_reference_features(circuit_type)

        # assert
        self.assertEqual('Reference features for IT circuit type are not defined.', str(context.exception))

    def test_get_power_converter_reference_fpa_rb(self):
        # arrange
        circuit_type = 'RB'
        circuit_name = 'RB.A12'

        # act
        fgc_pm_act = Reference.get_power_converter_reference_fpa(circuit_type, circuit_name, col='fgcPm')

        # assert
        fgc_pm_ref = '2018-03-17 16:00:28.680'
        self.assertEqual(fgc_pm_ref, fgc_pm_act)

    def test_get_power_converter_reference_fpa_rq(self):
        # arrange
        circuit_type = 'RQ'
        circuit_name = 'RQD.A12'

        # act
        fgc_pm_act = Reference.get_power_converter_reference_fpa(circuit_type, circuit_name, col='fgcPm')

        # assert
        fgc_pm_ref = '2018-03-18 11:58:33.600'
        self.assertEqual(fgc_pm_ref, fgc_pm_act)

    def test_get_power_converter_reference_fpa_it_error(self):
        # arrange
        circuit_type = 'IT'
        circuit_name = 'RQX.L1'

        # act
        with self.assertRaises(ValueError) as context:
            Reference.get_power_converter_reference_fpa(circuit_type, circuit_name, col='fgcPm')

        # assert
        self.assertEqual('Circuit type IT not supported!', str(context.exception))

    def test_compare_features_to_reference_error(self):
        # arrange
        features_df = pd.DataFrame()
        circuit_type = 'RB'
        system = 'VF'

        # act
        with self.assertRaises(KeyError) as context:
            Reference.compare_features_to_reference(features_df, circuit_type, system)

        # assert
        self.assertEqual('\'System VF is not supported!\'', str(context.exception))

    def test_compare_difference_of_features_to_reference_error(self):
        # arrange
        features_df = pd.DataFrame()
        reference_df = pd.DataFrame()
        circuit_type = 'RB'
        system = 'EE'
        wildcard = ''

        # act
        with self.assertRaises(KeyError) as context:
            Reference.compare_difference_of_features_to_reference(features_df, reference_df, circuit_type,
                                                                  system, wildcard)

        # assert
        self.assertEqual('\'System EE is not supported!\'', str(context.exception))

    def test_get_mp3_period(self):
        # arrange
        date_time = '2018-01-01 00:00:00.000+01:00'

        # act
        period_act = Reference.get_mp3_period(date_time)

        # assert
        period_exp = 'eYETS 2017/2018'
        self.assertEqual(period_exp, period_act)

